@@ -58,6 +58,8 @@ struct usb_driver wrap_usb_driver;
 
 extern int debug;
 
+int load_pe_images(struct pe_image[], int n);
+
 int wrap_device_type(int data1)
 {
 	int i;
@@ -114,8 +116,7 @@ struct wrap_driver *load_wrap_driver(struct wrap_device *wd)
 
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
 		/* wait for the driver to load and initialize */
-		set_current_state(TASK_INTERRUPTIBLE);
-		schedule_timeout(HZ);
+		sleep_hz(HZ);
 #endif
 		found = 0;
 		DBGTRACE1("%s", wd->driver_name);
@@ -270,8 +271,7 @@ struct wrap_bin_file *get_bin_file(char *bin_file_name)
 		}
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
 		/* wait for the driver to load and initialize */
-		set_current_state(TASK_INTERRUPTIBLE);
-		schedule_timeout(HZ);
+		sleep_hz(HZ);
 #endif
 		DBGTRACE2("bin file: %s/%s",
 			  wrap_bin_file.driver_name, wrap_bin_file.name);
@@ -544,8 +544,7 @@ static int load_user_space_driver(struct load_driver *load_driver)
 	struct wrap_driver *wrap_driver = NULL;
 
 	TRACEENTER1("");
-	drv_obj = allocate_object(sizeof(*drv_obj), OBJECT_TYPE_DRIVER,
-				  load_driver->name);
+	drv_obj = allocate_object(sizeof(*drv_obj), OBJECT_TYPE_DRIVER, NULL);
 	if (!drv_obj) {
 		ERROR("couldn't allocate memory");
 		TRACEEXIT1(return -ENOMEM);
@@ -561,7 +560,7 @@ static int load_user_space_driver(struct load_driver *load_driver)
 	InitializeListHead(&drv_obj->drv_ext->custom_ext);
 	DBGTRACE1("");
 	if (IoAllocateDriverObjectExtension(drv_obj,
-					    (void *)CE_WRAP_DRIVER_CLIENT_ID,
+					    (void *)WRAP_DRIVER_CLIENT_ID,
 					    sizeof(*wrap_driver),
 					    (void **)&wrap_driver) !=
 	    STATUS_SUCCESS)
