@@ -34,7 +34,7 @@ void misc_funcs_exit(void)
 	TRACEEXIT4(return);
 }
 
-noregparm INT WRAP_EXPORT(_win_sprintf)
+noregparm INT WIN_FUNC(_win_sprintf,12)
 	(char *buf, const char *format, ...)
 {
 	va_list args;
@@ -46,13 +46,13 @@ noregparm INT WRAP_EXPORT(_win_sprintf)
 	return res;
 }
 
-noregparm INT WRAP_EXPORT(swprintf)
+noregparm INT WIN_FUNC(swprintf,12)
 	(wchar_t *buf, const wchar_t *format, ...)
 {
-	UNIMPL();
+	TODO();
 	TRACEEXIT2(return 0);
 }
-noregparm INT WRAP_EXPORT(_win_vsprintf)
+noregparm INT WIN_FUNC(_win_vsprintf,3)
 	(char *str, const char *format, va_list ap)
 {
 	INT i;
@@ -61,7 +61,7 @@ noregparm INT WRAP_EXPORT(_win_vsprintf)
 	TRACEEXIT2(return i);
 }
 
-noregparm INT WRAP_EXPORT(_win_snprintf)
+noregparm INT WIN_FUNC(_win_snprintf,12)
 	(char *buf, SIZE_T count, const char *format, ...)
 {
 	va_list args;
@@ -74,7 +74,7 @@ noregparm INT WRAP_EXPORT(_win_snprintf)
 	return res;
 }
 
-noregparm INT WRAP_EXPORT(_win__snprintf)
+noregparm INT WIN_FUNC(_win__snprintf,12)
 	(char *buf, SIZE_T count, const char *format, ...)
 {
 	va_list args;
@@ -87,7 +87,7 @@ noregparm INT WRAP_EXPORT(_win__snprintf)
 	return res;
 }
 
-noregparm INT WRAP_EXPORT(_win_vsnprintf)
+noregparm INT WIN_FUNC(_win_vsnprintf,4)
 	(char *str, SIZE_T size, const char *format, va_list ap)
 {
 	INT i;
@@ -96,7 +96,7 @@ noregparm INT WRAP_EXPORT(_win_vsnprintf)
 	TRACEEXIT2(return i);
 }
 
-noregparm INT WRAP_EXPORT(_win__vsnprintf)
+noregparm INT WIN_FUNC(_win__vsnprintf,4)
 	(char *str, SIZE_T size, const char *format, va_list ap)
 {
 	INT i;
@@ -105,259 +105,259 @@ noregparm INT WRAP_EXPORT(_win__vsnprintf)
 	TRACEEXIT2(return i);
 }
 
-noregparm char *WRAP_EXPORT(_win_strncpy)
+noregparm char *WIN_FUNC(_win_strncpy,3)
 	(char *dst, char *src, SIZE_T n)
 {
 	return strncpy(dst, src, n);
 }
 
-noregparm SIZE_T WRAP_EXPORT(_win_strlen)
+noregparm SIZE_T WIN_FUNC(_win_strlen,1)
 	(const char *s)
 {
 	return strlen(s);
 }
 
-noregparm INT WRAP_EXPORT(_win_strncmp)
+noregparm INT WIN_FUNC(_win_strncmp,3)
 	(const char *s1, const char *s2, SIZE_T n)
 {
 	return strncmp(s1, s2, n);
 }
 
-noregparm INT WRAP_EXPORT(_win_strcmp)
+noregparm INT WIN_FUNC(_win_strcmp,2)
 	(const char *s1, const char *s2)
 {
 	return strcmp(s1, s2);
 }
 
-noregparm INT WRAP_EXPORT(_win_stricmp)
+noregparm INT WIN_FUNC(_win_stricmp,2)
 	(const char *s1, const char *s2)
 {
 	return stricmp(s1, s2);
 }
 
-noregparm char *WRAP_EXPORT(_win_strncat)
+noregparm char *WIN_FUNC(_win_strncat,3)
 	(char *dest, const char *src, SIZE_T n)
 {
 	return strncat(dest, src, n);
 }
 
-noregparm INT WRAP_EXPORT(_win_wcscmp)
+noregparm INT WIN_FUNC(_win_wcscmp,2)
 	(const wchar_t *s1, const wchar_t *s2)
 {
-	while (*s1 != L'\0' && *s1 == *s2) {
+	while (*s1 != '\0' && *s1 == *s2) {
 		s1++;
 		s2++;
 	}
 	return *s1 - *s2;
 }
 
-noregparm INT WRAP_EXPORT(_win_wcsicmp)
+noregparm INT WIN_FUNC(_win_wcsicmp,2)
 	(const wchar_t *s1, const wchar_t *s2)
 {
-	while (*s1 != L'\0' && tolower((char)*s1) == tolower((char)*s2)) {
+	while (*s1 != '\0' && tolower((char)*s1) == tolower((char)*s2)) {
 		s1++;
 		s2++;
 	}
 	return *s1 - *s2;
 }
 
-noregparm SIZE_T WRAP_EXPORT(_win_wcslen)
+noregparm SIZE_T WIN_FUNC(_win_wcslen,1)
 	(const wchar_t *s)
 {
 	SIZE_T i = 0;
-	while (s[i] != L'\0')
+	while (s[i] != '\0')
 		i++;
 	return i;
 }
 
-noregparm wchar_t *WRAP_EXPORT(_win_wcsncpy)
+noregparm wchar_t *WIN_FUNC(_win_wcsncpy,3)
 	(wchar_t *dest, const wchar_t *src, SIZE_T n)
 {
 	SIZE_T i;
-	for (i = 0; i < n && src[i] != L'\0'; i++)
+	for (i = 0; i < n && src[i] != '\0'; i++)
 		dest[i] = src[i];
 	if (i < n)
 		dest[i] = 0;
 	return dest;
 }
 
-noregparm wchar_t *WRAP_EXPORT(_win_wcscpy)
+noregparm wchar_t *WIN_FUNC(_win_wcscpy,2)
 	(wchar_t *dest, const wchar_t *src)
 {
 	SIZE_T i;
-	for (i = 0; src[i] != L'\0'; i++)
+	for (i = 0; src[i] != '\0'; i++)
 		dest[i] = src[i];
 	dest[i] = src[i];
 	return dest;
 }
 
-noregparm wchar_t *WRAP_EXPORT(_win_wcscat)
+noregparm wchar_t *WIN_FUNC(_win_wcscat,2)
 	(wchar_t *dest, const wchar_t *src)
 {
 	SIZE_T i;
 	wchar_t *destend;
-	for (i = 0; dest[i] != L'\0'; i++)
+	for (i = 0; dest[i] != '\0'; i++)
 		;
 	destend = &dest[i];
-	for (i = 0; src[i] != L'\0'; i++)
+	for (i = 0; src[i] != '\0'; i++)
 		destend[i] = src[i];
-	destend[i] = L'\0';
+	destend[i] = '\0';
 	return dest;
 }
 
-noregparm INT WRAP_EXPORT(_win_towupper)
+noregparm INT WIN_FUNC(_win_towupper,1)
 	(wchar_t c)
 {
 	return toupper(c);
 }
 
-noregparm INT WRAP_EXPORT(_win_towlower)
+noregparm INT WIN_FUNC(_win_towlower,1)
 	(wchar_t c)
 {
 	return tolower(c);
 }
 
-noregparm INT WRAP_EXPORT(_win_tolower)
+noregparm INT WIN_FUNC(_win_tolower,1)
 	(INT c)
 {
 	return tolower(c);
 }
 
-noregparm INT WRAP_EXPORT(_win_toupper)
+noregparm INT WIN_FUNC(_win_toupper,1)
 	(INT c)
 {
 	return toupper(c);
 }
 
-noregparm void *WRAP_EXPORT(_win_strcpy)
+noregparm void *WIN_FUNC(_win_strcpy,2)
 	(void *to, const void *from)
 {
 	return strcpy(to, from);
 }
 
-noregparm char *WRAP_EXPORT(_win_strstr)
+noregparm char *WIN_FUNC(_win_strstr,2)
 	(const char *s1, const char *s2)
 {
 	return strstr(s1, s2);
 }
 
-noregparm char *WRAP_EXPORT(_win_strchr)
+noregparm char *WIN_FUNC(_win_strchr,2)
 	(const char *s, int c)
 {
 	return strchr(s, c);
 }
 
-noregparm char *WRAP_EXPORT(_win_strrchr)
+noregparm char *WIN_FUNC(_win_strrchr,2)
 	(const char *s, int c)
 {
 	return strrchr(s, c);
 }
 
-noregparm void *WRAP_EXPORT(_win_memmove)
+noregparm void *WIN_FUNC(_win_memmove,3)
 	(void *to, void *from, SIZE_T count)
 {
 	return memmove(to, from, count);
 }
 
-noregparm void *WRAP_EXPORT(_win_memchr)
+noregparm void *WIN_FUNC(_win_memchr,3)
 	(const void *s, INT c, SIZE_T n)
 {
 	return memchr(s, c, n);
 }
 
 /* memcpy and memset are macros so we can't map them */
-noregparm void *WRAP_EXPORT(_win_memcpy)
+noregparm void *WIN_FUNC(_win_memcpy,3)
 	(void *to, const void *from, SIZE_T n)
 {
 	return memcpy(to, from, n);
 }
 
-noregparm void *WRAP_EXPORT(_win_memset)
+noregparm void *WIN_FUNC(_win_memset,3)
 	(void *s, char c, SIZE_T count)
 {
 	return memset(s, c, count);
 }
 
-noregparm void WRAP_EXPORT(_win_srand)
+noregparm void WIN_FUNC(_win_srand,1)
 	(UINT seed)
 {
 	net_srandom(seed);
 }
 
-noregparm int WRAP_EXPORT(_win_atoi)
+noregparm int WIN_FUNC(_win_atoi,1)
 	(const char *ptr)
 {
 	int i = simple_strtol(ptr, NULL, 10);
 	return i;
 }
 
-wstdcall int WRAP_EXPORT(_win_isprint)
+wstdcall int WIN_FUNC(_win_isprint,1)
 	(int c)
 {
 	return isprint(c);
 }
 
-wstdcall s64 WRAP_EXPORT(_alldiv)
+wstdcall s64 WIN_FUNC(_alldiv,2)
 	(s64 a, s64 b)
 {
 	return (a / b);
 }
 
-wstdcall u64 WRAP_EXPORT(_aulldiv)
+wstdcall u64 WIN_FUNC(_aulldiv,2)
 	(u64 a, u64 b)
 {
 	return (a / b);
 }
 
-wstdcall s64 WRAP_EXPORT(_allmul)
+wstdcall s64 WIN_FUNC(_allmul,2)
 	(s64 a, s64 b)
 {
 	return (a * b);
 }
 
-wstdcall u64 WRAP_EXPORT(_aullmul)
+wstdcall u64 WIN_FUNC(_aullmul,2)
 	(u64 a, u64 b)
 {
 	return (a * b);
 }
 
-wstdcall s64 WRAP_EXPORT(_allrem)
+wstdcall s64 WIN_FUNC(_allrem,2)
 	(s64 a, s64 b)
 {
 	return (a % b);
 }
 
-wstdcall u64 WRAP_EXPORT(_aullrem)
+wstdcall u64 WIN_FUNC(_aullrem,2)
 	(u64 a, u64 b)
 {
 	return (a % b);
 }
 
-__attribute__ ((regparm(3))) s64 WRAP_EXPORT(_allshl)
+__attribute__ ((regparm(3))) s64 WIN_FUNC(_allshl,2)
 	(s64 a, u8 b)
 {
 	return (a << b);
 }
 
-__attribute__ ((regparm(3))) u64 WRAP_EXPORT(_aullshl)
+__attribute__ ((regparm(3))) u64 WIN_FUNC(_aullshl,2)
 	(u64 a, u8 b)
 {
 	return (a << b);
 }
 
-__attribute__ ((regparm(3))) s64 WRAP_EXPORT(_allshr)
+__attribute__ ((regparm(3))) s64 WIN_FUNC(_allshr,2)
 	(s64 a, u8 b)
 {
 	return (a >> b);
 }
 
-__attribute__ ((regparm(3))) u64 WRAP_EXPORT(_aullshr)
+__attribute__ ((regparm(3))) u64 WIN_FUNC(_aullshr,2)
 	(u64 a, u8 b)
 {
 	return (a >> b);
 }
 
-wstdcall SIZE_T WRAP_EXPORT(RtlCompareMemory)
+wstdcall SIZE_T WIN_FUNC(RtlCompareMemory,3)
 	(const void *a, const void *b, SIZE_T len)
 {
 	size_t i;
@@ -382,25 +382,37 @@ wstdcall SIZE_T WRAP_EXPORT(RtlCompareMemory)
 	return i;
 }
 
-wstdcall void WRAP_EXPORT(RtlCopyMemory)
+wstdcall void WIN_FUNC(RtlCopyMemory,3)
 	(void *dst, const void *src, SIZE_T length)
 {
 	memcpy(dst, src, length);
 }
 
-wstdcall void WRAP_EXPORT(RtlZeroMemory)
+wstdcall void WIN_FUNC(RtlZeroMemory,2)
 	(void *dst, SIZE_T length)
 {
 	memset(dst, 0, length);
 }
 
-wstdcall void WRAP_EXPORT(RtlSecureZeroMemory)
+wstdcall void WIN_FUNC(RtlSecureZeroMemory,2)
 	(void *dst, SIZE_T length)
 {
 	memset(dst, 0, length);
 }
 
-wstdcall LONG WRAP_EXPORT(RtlCompareString)
+wstdcall void WIN_FUNC(RtlFillMemory,3)
+	(void *dest, SIZE_T length, UCHAR fill)
+{
+	memset(dest, fill, length);
+}
+
+wstdcall void WIN_FUNC(RtlMoveMemory,3)
+	(void *dest, const void *src, SIZE_T length)
+{
+	memmove(dest, src, length);
+}
+
+wstdcall LONG WIN_FUNC(RtlCompareString,3)
 	(const struct ansi_string *s1, const struct ansi_string *s2,
 	 BOOLEAN case_insensitive)
 {
@@ -423,7 +435,7 @@ wstdcall LONG WRAP_EXPORT(RtlCompareString)
 	TRACEEXIT2(return ret);
 }
 
-wstdcall LONG WRAP_EXPORT(RtlCompareUnicodeString)
+wstdcall LONG WIN_FUNC(RtlCompareUnicodeString,3)
 	(const struct unicode_string *s1, const struct unicode_string *s2,
 	 BOOLEAN case_insensitive)
 {
@@ -448,7 +460,7 @@ wstdcall LONG WRAP_EXPORT(RtlCompareUnicodeString)
 	TRACEEXIT2(return ret);
 }
 
-wstdcall BOOLEAN WRAP_EXPORT(RtlEqualString)
+wstdcall BOOLEAN WIN_FUNC(RtlEqualString,3)
 	(const struct ansi_string *s1, const struct ansi_string *s2,
 	 BOOLEAN case_insensitive)
 {
@@ -458,7 +470,7 @@ wstdcall BOOLEAN WRAP_EXPORT(RtlEqualString)
 	return !RtlCompareString(s1, s2, case_insensitive);
 }
 
-wstdcall BOOLEAN WRAP_EXPORT(RtlEqualUnicodeString)
+wstdcall BOOLEAN WIN_FUNC(RtlEqualUnicodeString,3)
 	(const struct unicode_string *s1, const struct unicode_string *s2,
 	 BOOLEAN case_insensitive)
 {
@@ -467,38 +479,52 @@ wstdcall BOOLEAN WRAP_EXPORT(RtlEqualUnicodeString)
 	return !RtlCompareUnicodeString(s1, s2, case_insensitive);
 }
 
-wstdcall void WRAP_EXPORT(RtlCopyUnicodeString)
+wstdcall void WIN_FUNC(RtlCopyUnicodeString,2)
 	(struct unicode_string *dst, struct unicode_string *src)
 {
 	TRACEENTER1("%p, %p", dst, src);
-	if (src) {
+	if (src && src->buf && dst->buf) {
 		dst->length = min(src->length, dst->max_length);
 		memcpy(dst->buf, src->buf, dst->length);
 		if (dst->length < dst->max_length)
-			dst->buf[dst->length / sizeof(wchar_t)] = 0;
+			dst->buf[dst->length / sizeof(dst->buf[0])] = 0;
 	} else
-		dst->max_length = 0;
+		dst->length = 0;
 	TRACEEXIT1(return);
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(RtlAppendUnicodeToString)
+wstdcall void WIN_FUNC(RtlCopyString,2)
+	(struct ansi_string *dst, struct ansi_string *src)
+{
+	TRACEENTER1("%p, %p", dst, src);
+	if (src && src->buf && dst->buf) {
+		dst->length = min(src->length, dst->max_length);
+		memcpy(dst->buf, src->buf, dst->length);
+		if (dst->length < dst->max_length)
+			dst->buf[dst->length] = 0;
+	} else
+		dst->length = 0;
+	TRACEEXIT1(return);
+}
+
+wstdcall NTSTATUS WIN_FUNC(RtlAppendUnicodeToString,2)
 	(struct unicode_string *dst, wchar_t *src)
 {
 	if (src) {
 		int len;
 		for (len = 0; src[len]; len++)
 			;
-		if (dst->length + (len * sizeof(wchar_t)) > dst->max_length)
+		if (dst->length + (len * sizeof(dst->buf[0])) > dst->max_length)
 			return STATUS_BUFFER_TOO_SMALL;
-		memcpy(&dst->buf[dst->length], src, len * sizeof(wchar_t));
-		dst->length += len * sizeof(wchar_t);
+		memcpy(&dst->buf[dst->length], src, len * sizeof(dst->buf[0]));
+		dst->length += len * sizeof(dst->buf[0]);
 		if (dst->max_length > dst->length)
-			dst->buf[dst->length / sizeof(wchar_t)] = 0;
+			dst->buf[dst->length / sizeof(dst->buf[0])] = 0;
 	}
 	return STATUS_SUCCESS;
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(RtlAppendUnicodeStringToString)
+wstdcall NTSTATUS WIN_FUNC(RtlAppendUnicodeStringToString,2)
 	(struct unicode_string *dst, struct unicode_string *src)
 {
 	if (dst->max_length < src->length + dst->length)
@@ -507,12 +533,12 @@ wstdcall NTSTATUS WRAP_EXPORT(RtlAppendUnicodeStringToString)
 		memcpy(&dst->buf[dst->length], src->buf, src->length);
 		dst->length += src->length;
 		if (dst->max_length > dst->length)
-			dst->buf[dst->length / sizeof(wchar_t)] = 0;
+			dst->buf[dst->length / sizeof(dst->buf[0])] = 0;
 	}
 	TRACEEXIT2(return STATUS_SUCCESS);
 }
 
-wstdcall ULONG WRAP_EXPORT(RtlxAnsiStringToUnicodeSize)
+wstdcall ULONG WIN_FUNC(RtlxAnsiStringToUnicodeSize,1)
 	(const struct ansi_string *string)
 {
 	int i;
@@ -522,7 +548,7 @@ wstdcall ULONG WRAP_EXPORT(RtlxAnsiStringToUnicodeSize)
 	return (i * sizeof(wchar_t));
 }
 
-wstdcall ULONG WRAP_EXPORT(RtlxUnicodeStringToAnsiSize)
+wstdcall ULONG WIN_FUNC(RtlxUnicodeStringToAnsiSize,1)
 	(const struct unicode_string *string)
 {
 	int i;
@@ -532,7 +558,7 @@ wstdcall ULONG WRAP_EXPORT(RtlxUnicodeStringToAnsiSize)
 	return i;
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(RtlAnsiStringToUnicodeString)
+wstdcall NTSTATUS WIN_FUNC(RtlAnsiStringToUnicodeString,3)
 	(struct unicode_string *dst, const struct ansi_string *src,
 	 BOOLEAN alloc)
 {
@@ -560,17 +586,17 @@ wstdcall NTSTATUS WRAP_EXPORT(RtlAnsiStringToUnicodeString)
 		TRACEEXIT2(return STATUS_BUFFER_TOO_SMALL);
 
 	dst->length = n;
-	n /= sizeof(wchar_t);
+	n /= sizeof(dst->buf[0]);
 	for (i = 0; i < n; i++)
 		dst->buf[i] = src->buf[i];
-	if (i * sizeof(wchar_t) < dst->max_length)
+	if (i * sizeof(dst->buf[0]) < dst->max_length)
 		dst->buf[i] = 0;
 	DBGTRACE2("dst: length: %d, max_length: %d, string: %p",
 		  dst->length, dst->max_length, src->buf);
 	TRACEEXIT2(return STATUS_SUCCESS);
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(RtlUnicodeStringToAnsiString)
+wstdcall NTSTATUS WIN_FUNC(RtlUnicodeStringToAnsiString,3)
 	(struct ansi_string *dst, const struct unicode_string *src,
 	 BOOLEAN alloc)
 {
@@ -607,7 +633,7 @@ wstdcall NTSTATUS WRAP_EXPORT(RtlUnicodeStringToAnsiString)
 	TRACEEXIT2(return STATUS_SUCCESS);
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(RtlUnicodeStringToInteger)
+wstdcall NTSTATUS WIN_FUNC(RtlUnicodeStringToInteger,3)
 	(struct unicode_string *ustring, ULONG base, ULONG *value)
 {
 	int i, negsign;
@@ -671,17 +697,17 @@ wstdcall NTSTATUS WRAP_EXPORT(RtlUnicodeStringToInteger)
 	return STATUS_SUCCESS;
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(RtlIntegerToUnicodeString)
+wstdcall NTSTATUS WIN_FUNC(RtlIntegerToUnicodeString,3)
 	(ULONG value, ULONG base, struct unicode_string *ustring)
 {
-	wchar_t *buf = ustring->buf;
+	typeof(ustring->buf) buf = ustring->buf;
 	int i;
 
 	if (base == 0)
 		base = 10;
 	if (!(base == 2 || base == 8 || base == 10 || base == 16))
 		return STATUS_INVALID_PARAMETER;
-	for (i = 0; value && i * sizeof(wchar_t) < ustring->max_length; i++) {
+	for (i = 0; value && i * sizeof(buf[0]) < ustring->max_length; i++) {
 		int r;
 		r = value % base;
 		value /= base;
@@ -692,31 +718,31 @@ wstdcall NTSTATUS WRAP_EXPORT(RtlIntegerToUnicodeString)
 	}
 	if (value)
 		return STATUS_BUFFER_OVERFLOW;
-	ustring->length = i * sizeof(wchar_t);
+	ustring->length = i * sizeof(buf[0]);
 	return STATUS_SUCCESS;
 }
 
-wstdcall LARGE_INTEGER WRAP_EXPORT(RtlConvertUlongToLargeInteger)
+wstdcall LARGE_INTEGER WIN_FUNC(RtlConvertUlongToLargeInteger,1)
 	(ULONG ul)
 {
 	LARGE_INTEGER li = ul;
 	return li;
 }
 
-wfastcall USHORT WRAP_EXPORT(RtlUShortByteSwap)
+wfastcall USHORT WIN_FUNC(RtlUShortByteSwap,1)
 	(USHORT src)
 {
 	return __swab16(src);
 }
 
-wfastcall ULONG WRAP_EXPORT(RtlUlongByteSwap)
+wfastcall ULONG WIN_FUNC(RtlUlongByteSwap,1)
 	(ULONG src)
 {
 	/* ULONG is 32 bits for both 32-bit and 64-bit architectures */
 	return __swab32(src);
 }
 
-wstdcall void WRAP_EXPORT(RtlInitUnicodeString)
+wstdcall void WIN_FUNC(RtlInitUnicodeString,2)
 	(struct unicode_string *dst, const wchar_t *src)
 {
 	TRACEENTER2("%p", dst);
@@ -727,16 +753,17 @@ wstdcall void WRAP_EXPORT(RtlInitUnicodeString)
 		dst->buf = NULL;
 	} else {
 		int i = 0;
-		while (src[i])
-			i++;
+		char c;
+		while ((c = src[i++]))
+			;
 		dst->buf = (wchar_t *)src;
-		dst->length = i * sizeof(wchar_t);
-		dst->max_length = (i + 1) * sizeof(wchar_t);
+		dst->length = i * sizeof(dst->buf[0]);
+		dst->max_length = (i + 1) * sizeof(dst->buf[0]);
 	}
 	TRACEEXIT1(return);
 }
 
-wstdcall void WRAP_EXPORT(RtlInitAnsiString)
+wstdcall void WIN_FUNC(RtlInitAnsiString,2)
 	(struct ansi_string *dst, const char *src)
 {
 	TRACEENTER2("%p", dst);
@@ -747,8 +774,8 @@ wstdcall void WRAP_EXPORT(RtlInitAnsiString)
 		dst->buf = NULL;
 	} else {
 		int i = 0;
-		while (src[i])
-			i++;
+		while (src[i++])
+			;
 		dst->buf = (char *)src;
 		dst->length = i;
 		dst->max_length = i + 1;
@@ -757,7 +784,7 @@ wstdcall void WRAP_EXPORT(RtlInitAnsiString)
 	TRACEEXIT2(return);
 }
 
-wstdcall void WRAP_EXPORT(RtlInitString)
+wstdcall void WIN_FUNC(RtlInitString,2)
 	(struct ansi_string *dst, const char *src)
 {
 	TRACEENTER2("%p", dst);
@@ -765,7 +792,7 @@ wstdcall void WRAP_EXPORT(RtlInitString)
 	TRACEEXIT2(return);
 }
 
-wstdcall void WRAP_EXPORT(RtlFreeUnicodeString)
+wstdcall void WIN_FUNC(RtlFreeUnicodeString,1)
 	(struct unicode_string *string)
 {
 	TRACEENTER2("%p", string);
@@ -778,7 +805,7 @@ wstdcall void WRAP_EXPORT(RtlFreeUnicodeString)
 	return;
 }
 
-wstdcall void WRAP_EXPORT(RtlFreeAnsiString)
+wstdcall void WIN_FUNC(RtlFreeAnsiString,1)
 	(struct ansi_string *string)
 {
 	TRACEENTER2("%p", string);
@@ -791,7 +818,40 @@ wstdcall void WRAP_EXPORT(RtlFreeAnsiString)
 	return;
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(RtlQueryRegistryValues)
+/* guid string is of the form: {XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX} */
+wstdcall NTSTATUS WIN_FUNC(RtlGUIDFromString,2)
+	(struct unicode_string *guid_string, struct guid *guid)
+{
+	struct ansi_string ansi;
+	NTSTATUS ret;
+	int i, j, k, l, m;
+	if (guid_string->length != 37 || guid_string->buf[0] != '{' ||
+	    guid_string->buf[36] != '}' || guid_string->buf[9] != '-' ||
+	    guid_string->buf[14] != '-' || guid_string->buf[19] != '-' ||
+	    guid_string->buf[24] != '-')
+		TRACEEXIT2(return STATUS_INVALID_PARAMETER);
+	ret = RtlUnicodeStringToAnsiString(&ansi, guid_string, FALSE);
+	if (ret != STATUS_SUCCESS)
+		return ret;
+	memcpy(&guid->data4, &ansi.buf[29], sizeof(guid->data3));
+	/* set end of data3 for scanf */
+	ansi.buf[29] = 0;
+	if (sscanf(&ansi.buf[1], "%x", &i) == 1 &&
+	    sscanf(&ansi.buf[10], "%x", &j) == 1 &&
+	    sscanf(&ansi.buf[15], "%x", &k) == 1 &&
+	    sscanf(&ansi.buf[20], "%x", &l) == 1 &&
+	    sscanf(&ansi.buf[25], "%x", &m) == 1) {
+		guid->data1 = (i << 16) | (j < 8) | k;
+		guid->data2 = l;
+		guid->data3 = m;
+		ret = STATUS_SUCCESS;
+	} else
+		ret = STATUS_INVALID_PARAMETER;
+	RtlFreeAnsiString(&ansi);
+	return ret;
+}
+
+wstdcall NTSTATUS WIN_FUNC(RtlQueryRegistryValues,5)
 	(ULONG relative, wchar_t *path, struct rtl_query_registry_table *tbl,
 	 void *context, void *env)
 {
@@ -801,7 +861,7 @@ wstdcall NTSTATUS WRAP_EXPORT(RtlQueryRegistryValues)
 	static int i = 0;
 
 	TRACEENTER3("%x, %p", relative, tbl);
-//	UNIMPL();
+//	TODO();
 
 	RtlInitUnicodeString(&unicode, path);
 	if (RtlUnicodeStringToAnsiString(&ansi, &unicode, TRUE) ==
@@ -865,7 +925,7 @@ wstdcall NTSTATUS WRAP_EXPORT(RtlQueryRegistryValues)
 	TRACEEXIT3(return ret);
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(RtlWriteRegistryValue)
+wstdcall NTSTATUS WIN_FUNC(RtlWriteRegistryValue,6)
 	(ULONG relative, wchar_t *path, wchar_t *name, ULONG type,
 	 void *data, ULONG length)
 {
@@ -873,7 +933,7 @@ wstdcall NTSTATUS WRAP_EXPORT(RtlWriteRegistryValue)
 	struct unicode_string unicode;
 
 	TRACEENTER3("%d", relative);
-	UNIMPL();
+	TODO();
 
 	RtlInitUnicodeString(&unicode, path);
 	if (RtlUnicodeStringToAnsiString(&ansi, &unicode, TRUE) ==
@@ -890,22 +950,21 @@ wstdcall NTSTATUS WRAP_EXPORT(RtlWriteRegistryValue)
 	TRACEEXIT5(return STATUS_SUCCESS);
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(RtlDeleteRegistryValue)
+wstdcall NTSTATUS WIN_FUNC(RtlDeleteRegistryValue,3)
 	(ULONG relative, wchar_t *path, wchar_t *name)
 {
 	return STATUS_SUCCESS;
 }
 
-wstdcall void WRAP_EXPORT(RtlAssert)
-	(char *failed_assertion, char *file_name, ULONG line_num,
-	 char *message)
+wstdcall void WIN_FUNC(RtlAssert,4)
+	(char *failed_assertion, char *file_name, ULONG line_num, char *message)
 {
 	ERROR("assertion '%s' failed at %s line %d%s",
 	      failed_assertion, file_name, line_num, message ? message : "");
 	return;
 }
 
-wstdcall int WRAP_EXPORT(rand)
+wstdcall int WIN_FUNC(rand,0)
 	(void)
 {
 	char buf[6];
@@ -917,7 +976,11 @@ wstdcall int WRAP_EXPORT(rand)
 	return r;
 }
 
-void WRAP_EXPORT(RtlUnwind)(void){UNIMPL();}
+void WIN_FUNC(RtlUnwind,0)
+	(void)
+{
+	TODO();
+}
 
 int stricmp(const char *s1, const char *s2)
 {
@@ -928,22 +991,11 @@ int stricmp(const char *s1, const char *s2)
 	return (int)*s1 - (int)*s2;
 }
 
-void *get_sp(void)
-{
-	ULONG_PTR i;
-
-#ifdef CONFIG_X86_64
-	__asm__ __volatile__("movq %%rsp, %0\n" : "=g"(i));
-#else
-	__asm__ __volatile__("movl %%esp, %0\n" : "=g"(i));
-#endif
-	return (void *)i;
-}
-
 void dump_stack(void)
 {
-	ULONG_PTR *sp = get_sp();
+	ULONG_PTR *sp;
 	int i;
+	get_sp(sp);
 	for (i = 0; i < 20; i++)
 		printk(KERN_DEBUG "sp[%d] = %p\n", i, (void *)sp[i]);
 }
@@ -953,8 +1005,7 @@ void dump_bytes(const char *ctx, const u8 *from, int len)
 	int i, j;
 	u8 *buf;
 
-	buf = kmalloc(len * 3 + 1, current_irql() < DISPATCH_LEVEL ?
-		      GFP_KERNEL : GFP_ATOMIC);
+	buf = kmalloc(len * 3 + 1, gfp_irql());
 	if (!buf) {
 		ERROR("couldn't allocate memory");
 		return;
