@@ -351,8 +351,8 @@ struct miniport_char {
 	USHORT filler;
 	UINT reserved;
 	BOOLEAN (*hangcheck)(void *ctx) wstdcall;
-	void (*disable_interrupts)(void *ctx) wstdcall;
-	void (*enable_interrupts)(void *ctx) wstdcall;
+	void (*disable_interrupt)(void *ctx) wstdcall;
+	void (*enable_interrupt)(void *ctx) wstdcall;
 	void (*miniport_halt)(void *ctx) wstdcall;
 	ndis_interrupt_handler handle_interrupt;
 	NDIS_STATUS (*init)(NDIS_STATUS *error_status, UINT *medium_index,
@@ -425,7 +425,7 @@ struct ndis_work_item {
 	void *ctx;
 	/* this should be NDIS_PROC, but we masquerade it as
 	 * NTOS_WORK_FUNC so we can use wrap_worker */
-	NTOS_WORK_FUNC func;
+	NTOS_WORK_FUNC func wstdcall;
 	UCHAR reserved[8 * sizeof(void *)];
 };
 
@@ -486,7 +486,7 @@ struct wrap_ndis_driver {
 };
 
 /* IDs used to store extensions in driver_object's custom extension */
-#define CE_NDIS_DRIVER_CLIENT_ID 10
+#define NDIS_DRIVER_CLIENT_ID 10
 
 struct ndis_wireless_stats {
 	ULONG length;
@@ -542,7 +542,7 @@ enum ndis_media_stream_mode {
 };
 
 enum wrapper_work {
-	LINK_STATUS_CHANGED, SET_MULTICAST_LIST, COLLECT_STATS, HANGCHECK,
+	LINK_STATUS_CHANGED, SET_MULTICAST_LIST, COLLECT_STATS, MINIPORT_RESET,
 	/* do not work when this is set */
 	SHUTDOWN
 };
@@ -617,8 +617,6 @@ struct ndis_timer {
 struct ndis_miniport_timer {
 	struct nt_timer nt_timer;
 	struct kdpc kdpc;
-	/* since kdpc already can store func, ctx, I don't know what
-	 * the following two fields are for */
 	DPC func;
 	void *ctx;
 	void *nmb;
@@ -801,16 +799,12 @@ struct wrap_ndis_device {
 
 	struct work_struct tx_work;
 	struct ndis_packet *tx_ring[TX_RING_SIZE];
-	struct ndis_packet **tx_array;
 	unsigned int tx_ring_start;
 	unsigned int tx_ring_end;
 	unsigned char is_tx_ring_full;
 	struct semaphore tx_ring_mutex;
 	unsigned int max_tx_packets;
-
-	unsigned char tx_ok;
-	NT_SPIN_LOCK tx_stats_lock;
-
+	u8 tx_ok;
 	struct semaphore ndis_comm_mutex;
 	wait_queue_head_t ndis_comm_wq;
 	int ndis_comm_done;
@@ -825,7 +819,6 @@ struct wrap_ndis_device {
 	struct timer_list hangcheck_timer;
 	struct timer_list stats_timer;
 	unsigned long scan_timestamp;
-	unsigned char link_status;
 	struct encr_info encr_info;
 	char nick[IW_ESSID_MAX_SIZE+1];
 	struct ndis_essid essid;
@@ -854,6 +847,8 @@ struct wrap_ndis_device {
 	enum ndis_physical_medium physical_medium;
 	u32 ndis_wolopts;
 	BOOLEAN pm_capa;
+	struct nt_list timer_list;
+	char netdev_name[IFNAMSIZ];
 };
 
 struct ndis_pmkid_candidate {
@@ -867,74 +862,70 @@ struct ndis_pmkid_candidate_list {
 	struct ndis_pmkid_candidate candidates[1];
 };
 
-wstdcall void NdisAllocatePacketPoolEx
-	(NDIS_STATUS *status, struct ndis_packet_pool **pool_handle,
-	 UINT num_descr, UINT overflowsize, UINT proto_rsvd_length);
-wstdcall void NdisFreePacketPool(struct ndis_packet_pool *pool);
-wstdcall void NdisAllocatePacket(NDIS_STATUS *status,
-				struct ndis_packet **packet,
-				struct ndis_packet_pool *pool);
-wstdcall void NdisFreePacket(struct ndis_packet *descr);
-wstdcall void NdisAllocateBufferPool
-	(NDIS_STATUS *status, struct ndis_buffer_pool **pool_handle,
-	 UINT num_descr);
-wstdcall void NdisFreeBufferPool(struct ndis_buffer_pool *pool);
-wstdcall void NdisAllocateBuffer
-	(NDIS_STATUS *status, ndis_buffer **buffer,
-	 struct ndis_buffer_pool *pool, void *virt, UINT length);
-wstdcall void NdisFreeBuffer(ndis_buffer *descr);
-wstdcall void NdisMIndicateReceivePacket(struct ndis_miniport_block *nmb,
-					struct ndis_packet **packets,
-					UINT nr_packets);
-wstdcall void NdisMSendComplete(struct ndis_miniport_block *nmb,
-			       struct ndis_packet *packet, NDIS_STATUS status);
-wstdcall void NdisMSendResourcesAvailable(struct ndis_miniport_block *nmb);
-wstdcall void NdisMIndicateStatus(struct ndis_miniport_block *nmb,
-				 NDIS_STATUS status, void *buf, UINT len);
-wstdcall void NdisMIndicateStatusComplete(struct ndis_miniport_block *nmb);
-wstdcall void NdisMQueryInformationComplete(struct ndis_miniport_block *nmb,
-					   NDIS_STATUS status);
-wstdcall void NdisMSetInformationComplete(struct ndis_miniport_block *nmb,
-					 NDIS_STATUS status);
-wstdcall void NdisMResetComplete(struct ndis_miniport_block *nmb,
-				NDIS_STATUS status, BOOLEAN address_reset);
-wstdcall ULONG NDIS_BUFFER_TO_SPAN_PAGES(ndis_buffer *buffer);
-wstdcall BOOLEAN NdisWaitEvent(struct ndis_event *event, UINT timeout);
-wstdcall void NdisSetEvent(struct ndis_event *event);
-irqreturn_t ndis_isr_shared(int irq, void *data, struct pt_regs *pt_regs);
-irqreturn_t ndis_isr_dynamic(int irq, void *data, struct pt_regs *pt_regs);
-wstdcall void NdisMDeregisterInterrupt(struct ndis_irq *ndis_irq);
-wstdcall void EthRxIndicateHandler
-	(struct ndis_miniport_block *nmb, void *rx_ctx, char *header1,
-	 char *header, UINT header_size, void *look_ahead,
-	 UINT look_ahead_size, UINT packet_size);
-wstdcall void EthRxComplete(struct ndis_miniport_block *nmb);
-wstdcall void NdisMTransferDataComplete
-	(struct ndis_miniport_block *nmb,
-	 struct ndis_packet *packet, NDIS_STATUS status, UINT bytes_txed);
-wstdcall void NdisWriteConfiguration
-	(NDIS_STATUS *status, struct ndis_miniport_block *nmb,
-	 struct unicode_string *key,
-	 struct ndis_configuration_parameter *param);
-wstdcall void NdisReadConfiguration
-	(NDIS_STATUS *status, struct ndis_configuration_parameter **param,
-	 struct ndis_miniport_block *nmb, struct unicode_string *key,
-	 enum ndis_parameter_type type);
-
+irqreturn_t ndis_isr(int irq, void *data, struct pt_regs *pt_regs);
 void init_nmb_functions(struct ndis_miniport_block *nmb);
 
-void *get_sp(void);
 int ndis_init(void);
 void ndis_exit_device(struct wrap_ndis_device *wnd);
 void ndis_exit(void);
 void insert_ndis_kdpc_work(struct kdpc *kdpc);
 BOOLEAN remove_ndis_kdpc_work(struct kdpc *kdpc);
 
-int load_pe_images(struct pe_image[], int n);
-
 int wrap_procfs_add_ndis_device(struct wrap_ndis_device *wnd);
 void wrap_procfs_remove_ndis_device(struct wrap_ndis_device *wnd);
 
+void NdisAllocatePacketPoolEx(NDIS_STATUS *status,
+			      struct ndis_packet_pool **pool_handle,
+			      UINT num_descr, UINT overflowsize,
+			      UINT proto_rsvd_length) wstdcall;
+void NdisFreePacketPool(struct ndis_packet_pool *pool) wstdcall;
+void NdisAllocatePacket(NDIS_STATUS *status, struct ndis_packet **packet,
+			struct ndis_packet_pool *pool) wstdcall;
+void NdisFreePacket(struct ndis_packet *descr) wstdcall;
+void NdisAllocateBufferPool(NDIS_STATUS *status,
+			    struct ndis_buffer_pool **pool_handle,
+			    UINT num_descr) wstdcall;
+void NdisFreeBufferPool(struct ndis_buffer_pool *pool) wstdcall;
+void NdisAllocateBuffer(NDIS_STATUS *status, ndis_buffer **buffer,
+			struct ndis_buffer_pool *pool, void *virt,
+			UINT length) wstdcall;
+void NdisFreeBuffer(ndis_buffer *descr) wstdcall;
+void NdisMIndicateReceivePacket(struct ndis_miniport_block *nmb,
+				struct ndis_packet **packets,
+				UINT nr_packets) wstdcall;
+void NdisMSendComplete(struct ndis_miniport_block *nmb,
+		       struct ndis_packet *packet, NDIS_STATUS status) wstdcall;
+void NdisMSendResourcesAvailable(struct ndis_miniport_block *nmb) wstdcall;
+void NdisMIndicateStatus(struct ndis_miniport_block *nmb,
+			 NDIS_STATUS status, void *buf, UINT len) wstdcall;
+void NdisMIndicateStatusComplete(struct ndis_miniport_block *nmb) wstdcall;
+void NdisMQueryInformationComplete(struct ndis_miniport_block *nmb,
+				   NDIS_STATUS status) wstdcall;
+void NdisMSetInformationComplete(struct ndis_miniport_block *nmb,
+				 NDIS_STATUS status) wstdcall;
+void NdisMResetComplete(struct ndis_miniport_block *nmb,
+			NDIS_STATUS status, BOOLEAN address_reset) wstdcall;
+ULONG NDIS_BUFFER_TO_SPAN_PAGES(ndis_buffer *buffer) wstdcall;
+BOOLEAN NdisWaitEvent(struct ndis_event *event, UINT timeout) wstdcall;
+void NdisSetEvent(struct ndis_event *event) wstdcall;
+void NdisMDeregisterInterrupt(struct ndis_irq *ndis_irq) wstdcall;
+void EthRxIndicateHandler(struct ndis_miniport_block *nmb, void *rx_ctx,
+			  char *header1, char *header, UINT header_size,
+			  void *look_ahead, UINT look_ahead_size,
+			  UINT packet_size) wstdcall;
+void EthRxComplete(struct ndis_miniport_block *nmb) wstdcall;
+void NdisMTransferDataComplete(struct ndis_miniport_block *nmb,
+			       struct ndis_packet *packet, NDIS_STATUS status,
+			       UINT bytes_txed) wstdcall;
+void NdisWriteConfiguration(NDIS_STATUS *status, struct ndis_miniport_block *nmb,
+			    struct unicode_string *key,
+			    struct ndis_configuration_parameter *param) wstdcall;
+void NdisReadConfiguration(NDIS_STATUS *status,
+			   struct ndis_configuration_parameter **param,
+			   struct ndis_miniport_block *nmb,
+			   struct unicode_string *key,
+			   enum ndis_parameter_type type) wstdcall;
+
 /* Required OIDs */
 #define OID_GEN_SUPPORTED_LIST			0x00010101
 #define OID_GEN_HARDWARE_STATUS			0x00010102
@@ -1221,4 +1212,45 @@ void wrap_procfs_remove_ndis_device(struct wrap_ndis_device *wnd);
 
 #define OID_TCP_TASK_OFFLOAD			0xFC010201
 
+#define deserialized_driver(wnd) (wnd->attributes & NDIS_ATTRIBUTE_DESERIALIZE)
+
+static inline void serialize_lock(struct wrap_ndis_device *wnd)
+{
+	nt_spin_lock(&wnd->nmb->lock);
+}
+
+static inline void serialize_unlock(struct wrap_ndis_device *wnd)
+{
+	nt_spin_unlock(&wnd->nmb->lock);
+}
+
+static inline KIRQL serialize_lock_irql(struct wrap_ndis_device *wnd)
+{
+	if (deserialized_driver(wnd))
+		return raise_irql(DISPATCH_LEVEL);
+	else
+		return nt_spin_lock_irql(&wnd->nmb->lock, DISPATCH_LEVEL);
+}
+
+static inline void serialize_unlock_irql(struct wrap_ndis_device *wnd,
+					 KIRQL irql)
+{
+	if (deserialized_driver(wnd))
+		lower_irql(irql);
+	else
+		nt_spin_unlock_irql(&wnd->nmb->lock, irql);
+}
+
+static inline void if_serialize_lock(struct wrap_ndis_device *wnd)
+{
+	if (!deserialized_driver(wnd))
+		nt_spin_lock(&wnd->nmb->lock);
+}
+
+static inline void if_serialize_unlock(struct wrap_ndis_device *wnd)
+{
+	if (!deserialized_driver(wnd))
+		nt_spin_unlock(&wnd->nmb->lock);
+}
+
 #endif /* NDIS_H */
