@@ -46,6 +46,8 @@
 #include <linux/if_arp.h>
 #include <linux/rtnetlink.h>
 
+#include "winnt_types.h"
+
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,7)
 #include <linux/kthread.h>
 #endif
@@ -67,7 +69,8 @@ typedef struct workqueue_struct *workqueue;
 #include <asm/dma-mapping.h>
 
 #define PCI_DMA_ALLOC_COHERENT(pci_dev,size,dma_handle)			\
-	dma_alloc_coherent(&pci_dev->dev,size,dma_handle, GFP_KERNEL)
+	dma_alloc_coherent(&pci_dev->dev,size,dma_handle,		\
+			   GFP_KERNEL | __GFP_REPEAT)
 #define PCI_DMA_FREE_COHERENT(pci_dev,size,cpu_addr,dma_handle)		\
 	dma_free_coherent(&pci_dev->dev,size,cpu_addr,dma_handle)
 #define PCI_DMA_MAP_SINGLE(pci_dev,addr,size,direction)		\
@@ -112,7 +115,7 @@ struct work_struct {
 	void *data;
 	/* whether/on which workqueue scheduled */
 	struct workqueue_struct *workq;
-};	
+};
 
 #define INIT_WORK(work_struct, worker_func, worker_data)	\
 	do {							\
@@ -124,7 +127,7 @@ struct work_struct {
 struct workqueue_struct *create_singlethread_workqueue(const char *name);
 void destroy_workqueue(struct workqueue_struct *workq);
 void queue_work(struct workqueue_struct *workq,
-		struct work_struct *work_struct);
+		struct work_struct *work_struct) fastcall;
 void cancel_delayed_work(struct work_struct *work_struct);
 
 #include <linux/smp_lock.h>
@@ -141,8 +144,8 @@ void cancel_delayed_work(struct work_struct *work_struct);
 
 #ifndef in_atomic
 #ifdef CONFIG_PREEMPT
-#define in_atomic()							\
-	 ((preempt_get_count() & ~PREEMPT_ACTIVE) != 0)
+#define in_atomic()					\
+	((preempt_get_count() & ~PREEMPT_ACTIVE) != 0)
 #else
 #define in_atomic() (in_interrupt())
 #endif // CONFIG_PREEMPT
@@ -243,6 +246,10 @@ do {									\
 #define NETDEV_TX_BUSY 1
 #endif
 
+#ifndef CHECKSUM_HW
+#define CHECKSUM_HW CHECKSUM_PARTIAL
+#endif
+
 /* this ugly hack is to handle RH kernels; I don't know any better,
  * but this has to be fixed soon */
 #ifndef rt_task
@@ -282,9 +289,6 @@ do {									\
 
 #endif // LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
 
-#define addr_offset(drvr) (__builtin_return_address(0) -	\
-			   (drvr)->drv_obj->driver_start)
-
 #ifndef offset_in_page
 #define offset_in_page(p) ((unsigned long)(p) & ~PAGE_MASK)
 #endif
@@ -330,115 +334,33 @@ typedef u32 pm_message_t;
 #define netdev_priv(dev)  ((dev)->priv)
 #endif
 
-#include "winnt_types.h"
 #include "ndiswrapper.h"
 #include "pe_linker.h"
 #include "wrapmem.h"
+#include "lin2win.h"
+
+#ifdef CONFIG_X86_64
+#define get_sp(sp) __asm__ __volatile__("mov %%rsp, %0\n\t" : "=m"(sp))
+#else
+#define get_sp(sp) __asm__ __volatile__("mov %%esp, %0\n\t" : "=m"(sp))
+#endif
+
+#define print_sp() do {				\
+		void *sp;			\
+		get_sp(sp);			\
+		DBGTRACE1("sp: %p", sp);	\
+	} while (0)
 
 //#define DEBUG_IRQL 1
 
-#if defined(DISABLE_USB)
-#undef CONFIG_USB
-#undef CONFIG_USB_MODULE
-#endif
-
 #if !defined(CONFIG_USB) && defined(CONFIG_USB_MODULE)
 #define CONFIG_USB 1
 #endif
 
-#ifdef CONFIG_X86_64
-#define LIN2WIN1(func, arg1)						\
-({									\
-	DBGTRACE6("calling %p", func);					\
-	lin_to_win1(func, (unsigned long)arg1);				\
-})
-#define LIN2WIN2(func, arg1, arg2)					\
-({									\
-	DBGTRACE6("calling %p", func);					\
-	lin_to_win2(func, (unsigned long)arg1, (unsigned long)arg2);	\
-})
-#define LIN2WIN3(func, arg1, arg2, arg3)				\
-({									\
-	DBGTRACE6("calling %p", func);					\
-	lin_to_win3(func, (unsigned long)arg1, (unsigned long)arg2,	\
-		    (unsigned long)arg3);				\
-})
-#define LIN2WIN4(func, arg1, arg2, arg3, arg4)				\
-({									\
-	DBGTRACE6("calling %p", func);					\
-	lin_to_win4(func, (unsigned long)arg1, (unsigned long)arg2,	\
-		    (unsigned long)arg3, (unsigned long)arg4);		\
-})
-#define LIN2WIN5(func, arg1, arg2, arg3, arg4, arg5)			\
-({									\
-	DBGTRACE6("calling %p", func);					\
-	lin_to_win5(func, (unsigned long)arg1, (unsigned long)arg2,	\
-		    (unsigned long)arg3, (unsigned long)arg4,		\
-		    (unsigned long)arg5);				\
-})
-#define LIN2WIN6(func, arg1, arg2, arg3, arg4, arg5, arg6)		\
-({									\
-	DBGTRACE6("calling %p", func);					\
-	lin_to_win6(func, (unsigned long)arg1, (unsigned long)arg2,	\
-		    (unsigned long)arg3, (unsigned long)arg4,		\
-		    (unsigned long)arg5, (unsigned long)arg6);		\
-})
-
-/* NOTE: these macros assume function arguments are quads and
- * arguments are not touched in any way before calling these macros */
-#define WIN2LIN2(arg1, arg2)						\
-do {									\
-	__asm__ __volatile__("mov %%rcx, %0\n\t"			\
-			     "mov %%rdx, %1\n\t"			\
-			     : "=m" (arg1), "=m" (arg2));		\
-} while (0)
-
-#define WIN2LIN3(arg1, arg2, arg3)					\
-do {									\
-	__asm__ __volatile__("mov %%rcx, %0\n\t"			\
-			     "mov %%rdx, %1\n\t"			\
-			     "mov %%r8, %2\n\t"				\
-			     : "=m" (arg1), "=m" (arg2), "=m" (arg3));	\
-} while (0)
-
-#else // CONFIG_X86_64
-
-#define LIN2WIN1(func, arg1)						\
-({									\
-	DBGTRACE6("calling %p", func);					\
-	func(arg1);							\
-})
-#define LIN2WIN2(func, arg1, arg2)					\
-({									\
-	DBGTRACE6("calling %p", func);					\
-	func(arg1, arg2);						\
-})
-#define LIN2WIN3(func, arg1, arg2, arg3)				\
-({									\
-	DBGTRACE6("calling %p", func);					\
-	func(arg1, arg2, arg3);						\
-})
-#define LIN2WIN4(func, arg1, arg2, arg3, arg4)				\
-({									\
-	DBGTRACE6("calling %p", func);					\
-	func(arg1, arg2, arg3, arg4);					\
-})
-#define LIN2WIN5(func, arg1, arg2, arg3, arg4, arg5)			\
-({									\
-	DBGTRACE6("calling %p", func);					\
-	func(arg1, arg2, arg3, arg4, arg5);				\
-})
-#define LIN2WIN6(func, arg1, arg2, arg3, arg4, arg5, arg6)		\
-({									\
-	DBGTRACE6("calling %p", func);					\
-	func(arg1, arg2, arg3, arg4, arg5, arg6);			\
-})
-
-#define WIN2LIN2(arg1, arg2) do { } while (0)
-
-#define WIN2LIN3(arg1, arg2, arg3) do { } while (0)
-
-#endif // CONFIG_X86_64
+#if defined(DISABLE_USB)
+#undef CONFIG_USB
+#undef CONFIG_USB_MODULE
+#endif
 
 #define KMALLOC_THRESHOLD 130000
 
@@ -446,9 +368,6 @@ do {									\
 #define TICKSPERSEC		10000000LL
 #define TICKSPERMSEC		10000
 #define SECSPERDAY		86400
-#define SECSPERHOUR		3600
-#define SECSPERMIN		60
-#define DAYSPERWEEK		7
 
 /* 1601 to 1970 is 369 years plus 89 leap days */
 #define SECS_1601_TO_1970	((369 * 365 + 89) * (u64)SECSPERDAY)
@@ -470,27 +389,35 @@ static inline u64 ticks_1601(void)
 	return wrap_ticks_to_boot + (u64)jiffies * TICKSPERSEC / HZ;
 }
 
-typedef void (*WRAP_EXPORT_FUNC)(void);
+typedef void (*generic_func)(void);
 
 struct wrap_export {
 	const char *name;
-	WRAP_EXPORT_FUNC func;
+	generic_func func;
 };
 
 #ifdef CONFIG_X86_64
-#define WRAP_EXPORT_SYMBOL(f) {#f, (WRAP_EXPORT_FUNC)x86_64_ ## f}
-#define WRAP_EXPORT_WIN_FUNC(f) {#f, (WRAP_EXPORT_FUNC)x86_64__win_ ## f}
-#define WRAP_FUNC_PTR(f) &x86_64_ ## f
-#define WRAP_FUNC_PTR_DECL(f) void x86_64_ ## f(void);
+
+#define WIN_SYMBOL(name, argc)					\
+	{#name, (generic_func) win2lin_ ## name ## _ ## argc}
+#define WIN_WIN_SYMBOL(name, argc)					\
+	{#name, (generic_func) win2lin__win_ ## name ## _ ## argc}
+#define WIN_FUNC_DECL(name, argc)			\
+	typeof(name) win2lin_ ## name ## _ ## argc ;
+#define WIN_FUNC_PTR(name, argc) win2lin_ ## name ## _ ## argc
+
 #else
-#define WRAP_EXPORT_SYMBOL(f) {#f, (WRAP_EXPORT_FUNC)f}
-#define WRAP_EXPORT_WIN_FUNC(f) {#f, (WRAP_EXPORT_FUNC)_win_ ## f}
-#define WRAP_FUNC_PTR(f) &f
-#define WRAP_FUNC_PTR_DECL(f)
+
+#define WIN_SYMBOL(name, argc) {#name, (generic_func)name}
+#define WIN_WIN_SYMBOL(name, argc) {#name, (generic_func)_win_ ## name}
+#define WIN_FUNC_DECL(name, argc)
+#define WIN_FUNC_PTR(name, argc) name
+
 #endif
-/* map name s to function f - if f is different from s */
-#define WRAP_EXPORT_MAP(s,f)
-#define WRAP_EXPORT(x) x
+
+#define WIN_FUNC(name, argc) name
+/* map name s to f - if f is different from s */
+#define WIN_SYMBOL_MAP(s, f)
 
 #define POOL_TAG(A, B, C, D)					\
 	((ULONG)((A) + ((B) << 8) + ((C) << 16) + ((D) << 24)))
@@ -506,24 +433,23 @@ struct pe_image {
 	IMAGE_OPTIONAL_HEADER *opt_hdr;
 };
 
+struct ndis_miniport_block;
+
 struct wrap_timer {
 	long repeat;
 	struct nt_list list;
 	struct timer_list timer;
 	struct nt_timer *nt_timer;
-#ifdef DEBUG_TIMER
+#ifdef TIMER_DEBUG
 	unsigned long wrap_timer_magic;
 #endif
 };
 
-enum worker_func_type { WORKER_FUNC_LINUX, WORKER_FUNC_WIN };
-
 struct ntos_work_item {
 	struct nt_list list;
 	void *arg1;
 	void *arg2;
 	void (*func)(void *arg1, void *arg2) wstdcall;
-	enum worker_func_type func_type;
 };
 
 struct wrap_device_setting {
@@ -540,7 +466,7 @@ struct wrap_bin_file {
 	void *data;
 };
 
-#define CE_WRAP_DRIVER_CLIENT_ID 1
+#define WRAP_DRIVER_CLIENT_ID 1
 
 struct wrap_driver {
 	struct nt_list list;
@@ -555,6 +481,8 @@ struct wrap_driver {
 	struct wrap_ndis_driver *ndis_driver;
 };
 
+struct usbd_pipe_information;
+
 struct wrap_device {
 	/* first part is (de)initialized once by loader */
 	int dev_bus_type;
@@ -585,36 +513,48 @@ struct wrap_device {
 			struct nt_list wrap_urb_list;
 		} usb;
 	};
-	struct wrap_ndis_device *wnd;
-	struct nt_list timer_list;
-	NT_SPIN_LOCK timer_lock;
+	union {
+		struct wrap_ndis_device *wnd;
+	};
 	struct cm_resource_list *resource_list;
 	BOOLEAN surprise_removed;
 };
 
+#define wrap_is_pci_bus(dev_bus_type)				\
+	(WRAP_BUS_TYPE(dev_bus_type) == WRAP_PCI_BUS)
+#ifdef CONFIG_USB
+/* earlier versions of ndiswrapper used 0 as USB_BUS */
+#define wrap_is_usb_bus(dev_bus_type)				\
+	(WRAP_BUS_TYPE(dev_bus_type) == WRAP_USB_BUS ||		\
+	 WRAP_BUS_TYPE(dev_bus_type) == 0)
+#else
+#define wrap_is_usb_bus(dev_bus_type) 0
+#endif
+#define wrap_is_bluetooth_device(dev_bus_type)				\
+	(WRAP_DEVICE_TYPE(dev_bus_type) == WRAP_BLUETOOTH_DEVICE1 ||	\
+	 WRAP_DEVICE_TYPE(dev_bus_type) == WRAP_BLUETOOTH_DEVICE2)
+
 /* Some drivers use worker entries to complete functions called from
- * within worker threads. So we should have separate workqueues to
- * make sure worker entries run properly */
+ * within other worker threads. So we should have separate workqueues
+ * to make sure worker entries run properly */
 
-#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
 #define USE_OWN_WORKQUEUE 1
+
+#ifdef USE_OWN_WORKQUEUE
+
 extern struct workqueue_struct *wrap_wq;
 #define schedule_ndis_work(work_struct) queue_work(ndis_wq, (work_struct))
 #define schedule_wrap_work(work_struct) queue_work(wrap_wq, (work_struct))
-#else
-#define schedule_ndis_work(work_struct) schedule_work(work_struct)
-#define schedule_wrap_work(work_struct) schedule_work(work_struct)
-#endif
 
 /* Normally workqueue for ntos is not required, as worker entries in
  * it are not supposed to wait; however, it helps to have separate
  * workqueue so keyboard etc. work when kernel crashes */
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,7)
-/* workqueues appeared in 2.6.7 or thereabouts */
+#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,41)
 #define USE_OWN_NTOS_WORKQUEUE 1
 #endif
 
+//#define USE_OWN_NTOS_WORKQUEUE 1
 #ifdef USE_OWN_NTOS_WORKQUEUE
 extern struct workqueue_struct *ntos_wq;
 #define schedule_ntos_work(work_struct) queue_work(ntos_wq, (work_struct))
@@ -622,12 +562,18 @@ extern struct workqueue_struct *ntos_wq;
 #define schedule_ntos_work(work_struct) schedule_work(work_struct)
 #endif
 
+#else // USE_OWN_WORKQUEUE
+#define schedule_ndis_work(work_struct) schedule_work(work_struct)
+#define schedule_wrap_work(work_struct) schedule_work(work_struct)
+#endif // USE_OWN_WORKQUEUE
+
 int ntoskernel_init(void);
 void ntoskernel_exit(void);
 int ntoskernel_init_device(struct wrap_device *wd);
 void ntoskernel_exit_device(struct wrap_device *wd);
-void *allocate_object(ULONG size, enum common_object_type type, char *name);
-void  free_object(void *object);
+void *allocate_object(ULONG size, enum common_object_type type,
+		      struct unicode_string *name);
+void free_object(void *object);
 
 int usb_init(void);
 void usb_exit(void);
@@ -637,194 +583,165 @@ void usb_cancel_pending_urbs(void);
 
 int misc_funcs_init(void);
 void misc_funcs_exit(void);
-
 int wrap_procfs_init(void);
 void wrap_procfs_remove(void);
-
 int stricmp(const char *s1, const char *s2);
 void dump_bytes(const char *name, const u8 *from, int len);
-
+struct mdl *allocate_init_mdl(void *virt, ULONG length);
+void free_mdl(struct mdl *mdl);
 struct driver_object *find_bus_driver(const char *name);
+void free_custom_extensions(struct driver_extension *drv_obj_ext);
+struct nt_thread *get_current_nt_thread(void);
+u64 ticks_1601(void);
+int schedule_ntos_work_item(NTOS_WORK_FUNC func, void *arg1, void *arg2);
+void wrap_init_timer(struct nt_timer *nt_timer, enum timer_type type,
+		     struct kdpc *kdpc, struct ndis_miniport_block *nmb);
+BOOLEAN wrap_set_timer(struct nt_timer *nt_timer, unsigned long expires_hz,
+		       unsigned long repeat_hz, struct kdpc *kdpc);
 
-wstdcall void WRITE_PORT_UCHAR(ULONG_PTR port, UCHAR value);
-wstdcall UCHAR READ_PORT_UCHAR(ULONG_PTR port);
+LONG InterlockedDecrement(LONG volatile *val) wfastcall;
+LONG InterlockedIncrement(LONG volatile *val) wfastcall;
+struct nt_list *ExInterlockedInsertHeadList
+	(struct nt_list *head, struct nt_list *entry,
+	 NT_SPIN_LOCK *lock) wfastcall;
+struct nt_list *ExInterlockedInsertTailList
+	(struct nt_list *head, struct nt_list *entry,
+	 NT_SPIN_LOCK *lock) wfastcall;
+struct nt_list *ExInterlockedRemoveHeadList
+	(struct nt_list *head, NT_SPIN_LOCK *lock) wfastcall;
+NTSTATUS IofCallDriver(struct device_object *dev_obj, struct irp *irp) wfastcall;
+KIRQL KfRaiseIrql(KIRQL newirql) wfastcall;
+void KfLowerIrql(KIRQL oldirql) wfastcall;
+KIRQL KfAcquireSpinLock(NT_SPIN_LOCK *lock) wfastcall;
+void KfReleaseSpinLock(NT_SPIN_LOCK *lock, KIRQL oldirql) wfastcall;
+void IofCompleteRequest(struct irp *irp, CHAR prio_boost) wfastcall;
+void KefReleaseSpinLockFromDpcLevel(NT_SPIN_LOCK *lock) wfastcall;
+
+LONG ObfReferenceObject(void *object) wfastcall;
+void ObfDereferenceObject(void *object) wfastcall;
+int dereference_object(void *object);
+
+#define ObReferenceObject(object) ObfReferenceObject(object)
+#define ObDereferenceObject(object) ObfDereferenceObject(object)
+
+void WRITE_PORT_UCHAR(ULONG_PTR port, UCHAR value) wstdcall;
+UCHAR READ_PORT_UCHAR(ULONG_PTR port) wstdcall;
 
 #undef ExAllocatePoolWithTag
-wstdcall void *ExAllocatePoolWithTag(enum pool_type pool_type, SIZE_T size,
-				    ULONG tag);
+void *ExAllocatePoolWithTag(enum pool_type pool_type, SIZE_T size,
+			    ULONG tag) wstdcall;
 #if defined(ALLOC_DEBUG) && ALLOC_DEBUG > 1
 #define ExAllocatePoolWithTag(pool_type, size, tag)			\
 	wrap_ExAllocatePoolWithTag(pool_type, size, tag, __FILE__, __LINE__)
 #endif
 
-wstdcall void ExFreePool(void *p);
-wstdcall ULONG MmSizeOfMdl(void *base, ULONG length);
-wstdcall void *MmMapIoSpace(PHYSICAL_ADDRESS phys_addr, SIZE_T size,
-			   enum memory_caching_type cache);
-wstdcall void MmUnmapIoSpace(void *addr, SIZE_T size);
-wstdcall void MmProbeAndLockPages(struct mdl *mdl, KPROCESSOR_MODE access_mode,
-	 enum lock_operation operation);
-wstdcall void MmUnlockPages(struct mdl *mdl);
-wstdcall void KeInitializeEvent(struct nt_event *nt_event,
-			       enum event_type type, BOOLEAN state);
-wstdcall LONG KeSetEvent(struct nt_event *nt_event, KPRIORITY incr,
-			BOOLEAN wait);
-wstdcall LONG KeResetEvent(struct nt_event *nt_event);
-wstdcall void KeClearEvent(struct nt_event *nt_event);
-wstdcall void KeInitializeDpc(struct kdpc *kdpc, void *func, void *ctx);
-wstdcall BOOLEAN KeInsertQueueDpc(struct kdpc *kdpc, void *arg1, void *arg2);
-wstdcall BOOLEAN KeRemoveQueueDpc(struct kdpc *kdpc);
-wstdcall void KeFlushQueuedDpcs(void);
-wstdcall NTSTATUS KeWaitForSingleObject
-	(void *object, KWAIT_REASON reason, KPROCESSOR_MODE waitmode,
-	 BOOLEAN alertable, LARGE_INTEGER *timeout);
-struct mdl *allocate_init_mdl(void *virt, ULONG length);
-void free_mdl(struct mdl *mdl);
-wstdcall struct mdl *IoAllocateMdl
-	(void *virt, ULONG length, BOOLEAN second_buf, BOOLEAN charge_quota,
-	 struct irp *irp);
-wstdcall void MmBuildMdlForNonPagedPool(struct mdl *mdl);
-wstdcall void IoFreeMdl(struct mdl *mdl);
-wfastcall LONG InterlockedDecrement(LONG volatile *val);
-wfastcall LONG InterlockedIncrement(LONG volatile *val);
-wfastcall struct nt_list *ExInterlockedInsertHeadList
-	(struct nt_list *head, struct nt_list *entry, NT_SPIN_LOCK *lock);
-wfastcall struct nt_list *ExInterlockedInsertTailList
-	(struct nt_list *head, struct nt_list *entry, NT_SPIN_LOCK *lock);
-wfastcall struct nt_list *ExInterlockedRemoveHeadList
-	(struct nt_list *head, NT_SPIN_LOCK *lock);
-wstdcall NTSTATUS IoCreateDevice
-	(struct driver_object *driver, ULONG dev_ext_length,
-	 struct unicode_string *dev_name, DEVICE_TYPE dev_type,
-	 ULONG dev_chars, BOOLEAN exclusive, struct device_object **dev_obj);
-wstdcall NTSTATUS IoCreateSymbolicLink
-	(struct unicode_string *link, struct unicode_string *dev_name);
-wstdcall void IoDeleteDevice(struct device_object *dev);
-wstdcall void IoDetachDevice(struct device_object *topdev);
-wstdcall struct device_object *IoGetAttachedDevice(struct device_object *dev);
-wstdcall struct device_object *IoGetAttachedDeviceReference
-	(struct device_object *dev);
-wstdcall NTSTATUS IoAllocateDriverObjectExtension
+void ExFreePool(void *p) wstdcall;
+ULONG MmSizeOfMdl(void *base, ULONG length) wstdcall;
+void *MmMapIoSpace(PHYSICAL_ADDRESS phys_addr, SIZE_T size,
+		   enum memory_caching_type cache) wstdcall;
+void MmUnmapIoSpace(void *addr, SIZE_T size) wstdcall;
+void MmProbeAndLockPages(struct mdl *mdl, KPROCESSOR_MODE access_mode,
+			 enum lock_operation operation) wstdcall;
+void MmUnlockPages(struct mdl *mdl) wstdcall;
+void KeInitializeEvent(struct nt_event *nt_event,
+		       enum event_type type, BOOLEAN state) wstdcall;
+LONG KeSetEvent(struct nt_event *nt_event, KPRIORITY incr,
+		BOOLEAN wait) wstdcall;
+LONG KeResetEvent(struct nt_event *nt_event) wstdcall;
+void KeClearEvent(struct nt_event *nt_event) wstdcall;
+void KeInitializeDpc(struct kdpc *kdpc, void *func, void *ctx) wstdcall;
+BOOLEAN KeInsertQueueDpc(struct kdpc *kdpc, void *arg1, void *arg2) wstdcall;
+BOOLEAN KeRemoveQueueDpc(struct kdpc *kdpc) wstdcall;
+void KeFlushQueuedDpcs(void) wstdcall;
+NTSTATUS KeWaitForSingleObject(void *object, KWAIT_REASON reason,
+			       KPROCESSOR_MODE waitmode, BOOLEAN alertable,
+			       LARGE_INTEGER *timeout) wstdcall;
+struct mdl *IoAllocateMdl(void *virt, ULONG length, BOOLEAN second_buf,
+			  BOOLEAN charge_quota, struct irp *irp) wstdcall;
+void MmBuildMdlForNonPagedPool(struct mdl *mdl) wstdcall;
+void IoFreeMdl(struct mdl *mdl) wstdcall;
+NTSTATUS IoCreateDevice(struct driver_object *driver, ULONG dev_ext_length,
+			struct unicode_string *dev_name, DEVICE_TYPE dev_type,
+			ULONG dev_chars, BOOLEAN exclusive,
+			struct device_object **dev_obj) wstdcall;
+NTSTATUS IoCreateSymbolicLink(struct unicode_string *link,
+			      struct unicode_string *dev_name) wstdcall;
+void IoDeleteDevice(struct device_object *dev) wstdcall;
+void IoDetachDevice(struct device_object *topdev) wstdcall;
+struct device_object *IoGetAttachedDevice(struct device_object *dev) wstdcall;
+struct device_object *IoGetAttachedDeviceReference
+	(struct device_object *dev) wstdcall;
+NTSTATUS IoAllocateDriverObjectExtension
 	(struct driver_object *drv_obj, void *client_id, ULONG extlen,
-	 void **ext);
-wstdcall void *IoGetDriverObjectExtension
-	(struct driver_object *drv, void *client_id);
-wstdcall struct device_object *IoAttachDeviceToDeviceStack
-	(struct device_object *src, struct device_object *dst);
-wstdcall void KeInitializeEvent
-	(struct nt_event *nt_event, enum event_type type, BOOLEAN state);
-void free_custom_extensions(struct driver_extension *drv_obj_ext);
-wstdcall struct irp *IoAllocateIrp(char stack_count, BOOLEAN charge_quota);
-wstdcall void IoFreeIrp(struct irp *irp);
-wstdcall BOOLEAN IoCancelIrp(struct irp *irp);
-wfastcall NTSTATUS IofCallDriver
-	(struct device_object *dev_obj, struct irp *irp);
-wstdcall struct irp *IoBuildSynchronousFsdRequest
+	 void **ext) wstdcall;
+void *IoGetDriverObjectExtension(struct driver_object *drv,
+				 void *client_id) wstdcall;
+struct device_object *IoAttachDeviceToDeviceStack
+	(struct device_object *src, struct device_object *dst) wstdcall;
+void KeInitializeEvent(struct nt_event *nt_event, enum event_type type,
+		       BOOLEAN state) wstdcall;
+struct irp *IoAllocateIrp(char stack_count, BOOLEAN charge_quota) wstdcall;
+void IoFreeIrp(struct irp *irp) wstdcall;
+BOOLEAN IoCancelIrp(struct irp *irp) wstdcall;
+struct irp *IoBuildSynchronousFsdRequest
 	(ULONG major_func, struct device_object *dev_obj, void *buf,
 	 ULONG length, LARGE_INTEGER *offset, struct nt_event *event,
-	 struct io_status_block *status);
-wstdcall struct irp *IoBuildAsynchronousFsdRequest
-	(ULONG major_func,struct device_object *dev_obj, void *buf,
-	 ULONG length, LARGE_INTEGER *offset, struct io_status_block *status);
-wstdcall NTSTATUS PoCallDriver(struct device_object *dev_obj, struct irp *irp);
+	 struct io_status_block *status) wstdcall;
+struct irp *IoBuildAsynchronousFsdRequest
+	(ULONG major_func, struct device_object *dev_obj, void *buf,
+	 ULONG length, LARGE_INTEGER *offset,
+	 struct io_status_block *status) wstdcall;
+NTSTATUS PoCallDriver(struct device_object *dev_obj, struct irp *irp) wstdcall;
 
-wstdcall NTSTATUS IoPassIrpDown
-	(struct device_object *dev_obj, struct irp *irp);
-wstdcall NTSTATUS IoSyncForwardIrp
-	(struct device_object *dev_obj, struct irp *irp);
-wstdcall NTSTATUS IoAsyncForwardIrp
-	(struct device_object *dev_obj, struct irp *irp);
-wstdcall NTSTATUS IoInvalidDeviceRequest
-	(struct device_object *dev_obj, struct irp *irp);
+NTSTATUS IoPassIrpDown(struct device_object *dev_obj, struct irp *irp) wstdcall;
+WIN_FUNC_DECL(IoPassIrpDown,2)
+NTSTATUS IoSyncForwardIrp(struct device_object *dev_obj,
+			  struct irp *irp) wstdcall;
+NTSTATUS IoAsyncForwardIrp (struct device_object *dev_obj,
+			    struct irp *irp) wstdcall;
+NTSTATUS IoInvalidDeviceRequest(struct device_object *dev_obj,
+				struct irp *irp) wstdcall;
 
-struct nt_thread *get_current_nt_thread(void);
-u64 ticks_1601(void);
+KIRQL KeGetCurrentIrql(void) wstdcall;
+void KeInitializeSpinLock(NT_SPIN_LOCK *lock) wstdcall;
+void KeAcquireSpinLock(NT_SPIN_LOCK *lock, KIRQL *irql) wstdcall;
+void KeReleaseSpinLock(NT_SPIN_LOCK *lock, KIRQL oldirql) wstdcall;
+KIRQL KeAcquireSpinLockRaiseToDpc(NT_SPIN_LOCK *lock) wstdcall;
 
-int schedule_ntos_work_item(NTOS_WORK_FUNC func, void *arg1, void *arg2,
-			    enum worker_func_type func_type);
+void IoAcquireCancelSpinLock(KIRQL *irql) wstdcall;
+void IoReleaseCancelSpinLock(KIRQL irql) wstdcall;
 
-wstdcall KIRQL KeGetCurrentIrql(void);
-wstdcall void KeInitializeSpinLock(NT_SPIN_LOCK *lock);
-wstdcall void KeAcquireSpinLock(NT_SPIN_LOCK *lock, KIRQL *irql);
-wstdcall void KeReleaseSpinLock(NT_SPIN_LOCK *lock, KIRQL oldirql);
-wstdcall KIRQL KeAcquireSpinLockRaiseToDpc(NT_SPIN_LOCK *lock);
-
-wstdcall void IoAcquireCancelSpinLock(KIRQL *irql);
-wstdcall void IoReleaseCancelSpinLock(KIRQL irql);
-
-wfastcall KIRQL KfRaiseIrql(KIRQL newirql);
-wfastcall void KfLowerIrql(KIRQL oldirql);
-wfastcall KIRQL KfAcquireSpinLock(NT_SPIN_LOCK *lock);
-wfastcall void KfReleaseSpinLock(NT_SPIN_LOCK *lock, KIRQL oldirql);
-wfastcall void IofCompleteRequest(struct irp *irp, CHAR prio_boost);
-wfastcall void KefReleaseSpinLockFromDpcLevel(NT_SPIN_LOCK *lock);
-wstdcall void RtlCopyMemory(void *dst, const void *src, SIZE_T length);
-wstdcall NTSTATUS RtlUnicodeStringToAnsiString
-	(struct ansi_string *dst, const struct unicode_string *src, BOOLEAN dup);
-wstdcall NTSTATUS RtlAnsiStringToUnicodeString
-	(struct unicode_string *dst, const struct ansi_string *src, BOOLEAN dup);
-wstdcall void RtlInitAnsiString(struct ansi_string *dst, const char *src);
-wstdcall void RtlInitString(struct ansi_string *dst, const char *src);
-wstdcall void RtlInitUnicodeString
-	(struct unicode_string *dest, const wchar_t *src);
-wstdcall void RtlFreeUnicodeString(struct unicode_string *string);
-wstdcall void RtlFreeAnsiString(struct ansi_string *string);
-wstdcall LONG RtlCompareUnicodeString
-	(const struct unicode_string *s1, const struct unicode_string *s2,
-	 BOOLEAN case_insensitive);
-wstdcall void RtlCopyUnicodeString
-	(struct unicode_string *dst, struct unicode_string *src);
-noregparm SIZE_T _win_wcslen(const wchar_t *s);
-
-void wrap_init_timer(struct nt_timer *nt_timer, enum timer_type type,
-		     struct wrap_device *wd);
-BOOLEAN wrap_set_timer(struct nt_timer *nt_timer, unsigned long expires_hz,
-		       unsigned long repeat_hz, struct kdpc *kdpc);
-
-wstdcall void KeInitializeTimer(struct nt_timer *nt_timer);
-wstdcall void KeInitializeTimerEx
-	(struct nt_timer *nt_timer, enum timer_type type);
-wstdcall BOOLEAN KeSetTimerEx
-	(struct nt_timer *nt_timer, LARGE_INTEGER duetime_ticks, LONG period_ms,
-	 struct kdpc *kdpc);
-wstdcall BOOLEAN KeSetTimer
-	(struct nt_timer *nt_timer, LARGE_INTEGER duetime_ticks,
-	 struct kdpc *kdpc);
-wstdcall BOOLEAN KeCancelTimer(struct nt_timer *nt_timer);
-wstdcall void KeInitializeDpc(struct kdpc *kdpc, void *func, void *ctx);
-
-unsigned long lin_to_win1(void *func, unsigned long);
-unsigned long lin_to_win2(void *func, unsigned long, unsigned long);
-unsigned long lin_to_win3(void *func, unsigned long, unsigned long,
-			  unsigned long);
-unsigned long lin_to_win4(void *func, unsigned long, unsigned long,
-			  unsigned long, unsigned long);
-unsigned long lin_to_win5(void *func, unsigned long, unsigned long,
-			  unsigned long, unsigned long, unsigned long);
-unsigned long lin_to_win6(void *func, unsigned long, unsigned long,
-			  unsigned long, unsigned long, unsigned long,
-			  unsigned long);
-
-wstdcall struct task_struct *KeGetCurrentThread(void);
-wstdcall NTSTATUS ObReferenceObjectByHandle
-	(void *handle, ACCESS_MASK desired_access, void *obj_type,
-	 KPROCESSOR_MODE access_mode, void **object, void *handle_info);
-wfastcall LONG ObfReferenceObject(void *object);
-wfastcall void ObfDereferenceObject(void *object);
-
-#define ObReferenceObject(object) ObfReferenceObject(object)
-#define ObDereferenceObject(object) ObfDereferenceObject(object)
-
-#define MSG(level, fmt, ...)				\
-	printk(level "ndiswrapper (%s:%d): " fmt "\n",	\
-	       __FUNCTION__, __LINE__ , ## __VA_ARGS__)
-#define WARNING(fmt, ...) MSG(KERN_WARNING, fmt, ## __VA_ARGS__)
-#define ERROR(fmt, ...) MSG(KERN_ERR, fmt , ## __VA_ARGS__)
-#define INFO(fmt, ...) MSG(KERN_INFO, fmt , ## __VA_ARGS__)
-
-#define INFOEXIT(stmt) do { INFO("Exit"); stmt; } while(0)
-
-#define UNIMPL() ERROR("--UNIMPLEMENTED--")
+void RtlCopyMemory(void *dst, const void *src, SIZE_T length) wstdcall;
+NTSTATUS RtlUnicodeStringToAnsiString
+	(struct ansi_string *dst, const struct unicode_string *src,
+	 BOOLEAN dup) wstdcall;
+NTSTATUS RtlAnsiStringToUnicodeString
+	(struct unicode_string *dst, const struct ansi_string *src,
+	 BOOLEAN dup) wstdcall;
+void RtlInitAnsiString(struct ansi_string *dst, const char *src) wstdcall;
+void RtlInitString(struct ansi_string *dst, const char *src) wstdcall;
+void RtlInitUnicodeString(struct unicode_string *dest,
+			  const wchar_t *src) wstdcall;
+void RtlFreeUnicodeString(struct unicode_string *string) wstdcall;
+void RtlFreeAnsiString(struct ansi_string *string) wstdcall;
+LONG RtlCompareUnicodeString(const struct unicode_string *s1,
+			     const struct unicode_string *s2,
+			     BOOLEAN case_insensitive) wstdcall;
+void RtlCopyUnicodeString(struct unicode_string *dst,
+			  struct unicode_string *src) wstdcall;
+void KeInitializeTimer(struct nt_timer *nt_timer) wstdcall;
+void KeInitializeTimerEx(struct nt_timer *nt_timer,
+			 enum timer_type type) wstdcall;
+BOOLEAN KeSetTimerEx(struct nt_timer *nt_timer, LARGE_INTEGER duetime_ticks,
+		     LONG period_ms, struct kdpc *kdpc) wstdcall;
+BOOLEAN KeSetTimer(struct nt_timer *nt_timer, LARGE_INTEGER duetime_ticks,
+		   struct kdpc *kdpc) wstdcall;
+BOOLEAN KeCancelTimer(struct nt_timer *nt_timer) wstdcall;
+void KeInitializeDpc(struct kdpc *kdpc, void *func, void *ctx) wstdcall;
+struct task_struct *KeGetCurrentThread(void) wstdcall;
+NTSTATUS ObReferenceObjectByHandle(void *handle, ACCESS_MASK desired_access,
+				   void *obj_type, KPROCESSOR_MODE access_mode,
+				   void **object, void *handle_info) wstdcall;
 
 void adjust_user_shared_data_addr(char *driver, unsigned long length);
 
@@ -834,36 +751,38 @@ void adjust_user_shared_data_addr(char *driver, unsigned long length);
 static inline KIRQL current_irql(void)
 {
 	if (in_irq() || irqs_disabled())
-		return DEVICE_LEVEL;
+		TRACEEXIT5(return DEVICE_LEVEL);
 	if (in_atomic())
-		return DISPATCH_LEVEL;
-	return PASSIVE_LEVEL;
+		TRACEEXIT5(return DISPATCH_LEVEL);
+	TRACEEXIT5(return PASSIVE_LEVEL);
 }
 
 static inline KIRQL raise_irql(KIRQL newirql)
 {
 	KIRQL irql = current_irql();
+//	assert (newirql == DISPATCH_LEVEL);
 	if (irql < DISPATCH_LEVEL && newirql == DISPATCH_LEVEL) {
 		local_bh_disable();
 		preempt_disable();
 	}
+	DBGTRACE5("%d, %d", irql, newirql);
 	return irql;
 }
 
 static inline void lower_irql(KIRQL oldirql)
 {
-#ifdef DEBUG_IRQL
 	KIRQL irql = current_irql();
+	DBGTRACE5("%d, %d", irql, oldirql);
 	if (irql < oldirql)
 		ERROR("invalid irql: %d < %d", irql, oldirql);
-#endif
-//	if (irql == DISPATCH_LEVEL && oldirql < DISPATCH_LEVEL) {
-	if (oldirql < DISPATCH_LEVEL) {
-		preempt_enable_no_resched();
+	if (oldirql < DISPATCH_LEVEL && irql == DISPATCH_LEVEL) {
+		preempt_enable();
 		local_bh_enable();
 	}
 }
 
+#define gfp_irql() (current_irql() < DISPATCH_LEVEL ? GFP_KERNEL : GFP_ATOMIC)
+
 /* Windows spinlocks are of type ULONG_PTR which is not big enough to
  * store Linux spinlocks; so we implement Windows spinlocks using
  * ULONG_PTR space with our own functions/macros */
@@ -897,8 +816,8 @@ static inline void nt_spin_lock(volatile NT_SPIN_LOCK *lock)
 		"  jmp 1b\n"
 		"3:\n\t"
 		: "=m" (*lock)
-		: "r" (NT_SPIN_LOCK_LOCKED), "i" (NT_SPIN_LOCK_UNLOCKED)
-		: "memory");
+		: "r" (NT_SPIN_LOCK_LOCKED), "i" (NT_SPIN_LOCK_UNLOCKED),
+		  "m" (*lock));
 }
 
 static inline void nt_spin_unlock(volatile NT_SPIN_LOCK *lock)
@@ -951,6 +870,48 @@ do {									\
 	preempt_enable();						\
 } while (0)
 
+#define atomic_unary_op(var, size, oper)			\
+	do {							\
+		if (size == 1)					\
+			__asm__ __volatile__(			\
+				LOCK_PREFIX oper "b %b0\n\t"	\
+				: "=m" (var) : "m" (var));	\
+		else if (size == 2)				\
+			__asm__ __volatile__(			\
+				LOCK_PREFIX oper "w %w0\n\t"	\
+				: "=m" (var) : "m" (var));	\
+		else if (size == 4)				\
+			__asm__ __volatile__(			\
+				LOCK_PREFIX oper "l %0\n\t"	\
+				: "=m" (var) : "m" (var));	\
+		else if (size == 8)				\
+			__asm__ __volatile__(			\
+				LOCK_PREFIX oper "q %q0\n\t"	\
+				: "=m" (var) : "m" (var));	\
+		else						\
+			ERROR("invalid size %d", (int)size);	\
+	} while (0)
+
+#define atomic_inc_var_size(var, size) atomic_unary_op(var, size, "inc")
+
+#define atomic_inc_var(var) atomic_inc_var_size(var, sizeof(var))
+
+#define atomic_dec_var_size(var, size) atomic_unary_op(var, size, "dec")
+
+#define atomic_dec_var(var) atomic_dec_var_size(var, sizeof(var))
+
+#define pre_atomic_add(var, i)					\
+({								\
+	typeof(var) pre;					\
+	__asm__ __volatile__(					\
+		LOCK_PREFIX "xadd %0, %1\n\t"			\
+		: "=r"(pre), "=m"(var)				\
+		: "0"(i), "m" (var));				\
+	pre;							\
+})
+
+#define post_atomic_add(var, i) (pre_atomic_add(var, i) + i)
+
 static inline ULONG SPAN_PAGES(void *ptr, SIZE_T length)
 {
 	ULONG n;
@@ -962,9 +923,6 @@ static inline ULONG SPAN_PAGES(void *ptr, SIZE_T length)
 }
 
 #ifdef CONFIG_X86_64
-/* it is not clear how nt_slist_header is used to store pointer to
- * slists and depth; here we assume 'align' field is used to store
- * depth and 'region' field is used to store slist pointers */
 
 /* TODO: can these be implemented without using spinlock? */
 
@@ -973,11 +931,11 @@ static inline struct nt_slist *PushEntrySList(nt_slist_header *head,
 					      NT_SPIN_LOCK *lock)
 {
 	KIRQL irql = nt_spin_lock_irql(lock, DISPATCH_LEVEL);
-	entry->next = (struct nt_slist *)head->region;
-	head->region = (ULONGLONG)entry;
-	head->align = (head->align & 0x0000ffff) |
-		((head->align & 0xffff) + 1);
+	entry->next = head->next;
+	head->next = entry;
+	head->depth++;
 	nt_spin_unlock_irql(lock, irql);
+	DBGTRACE4("%p, %p, %p", head, entry, entry->next);
 	return entry->next;
 }
 
@@ -986,13 +944,13 @@ static inline struct nt_slist *PopEntrySList(nt_slist_header *head,
 {
 	struct nt_slist *entry;
 	KIRQL irql = nt_spin_lock_irql(lock, DISPATCH_LEVEL);
-	entry = (struct nt_slist *)head->region;
+	entry = head->next;
 	if (entry) {
-		head->region = (ULONGLONG)entry->next;
-		head->align = (head->align & 0x0000ffff) |
-			((head->align & 0xffff) - 1);
+		head->next = entry->next;
+		head->depth--;
 	}
 	nt_spin_unlock_irql(lock, irql);
+	DBGTRACE4("%p, %p", head, entry);
 	return entry;
 }
 
@@ -1014,7 +972,7 @@ static inline u64 cmpxchg8b(volatile void *ptr, u64 old, u64 new)
 		  "=a" (ll_low(prev)), "=d" (ll_high(prev))
 		: "a" (ll_low(old)), "d" (ll_high(old)),
 		  "b" (ll_low(new)), "c" (ll_high(new))
-		: "cc", "memory");
+		: "cc");
 	return prev;
 }
 
@@ -1032,6 +990,7 @@ static inline struct nt_slist *PushEntrySList(nt_slist_header *head,
 		new.next = entry;
 		new.depth = old.depth + 1;
 	} while (cmpxchg8b(&head->align, old.align, new.align) != old.align);
+	DBGTRACE4("%p, %p, %p", head, entry, old.next);
 	return old.next;
 }
 
@@ -1048,148 +1007,12 @@ static inline struct nt_slist *PopEntrySList(nt_slist_header *head,
 		new.next = entry->next;
 		new.depth = old.depth - 1;
 	} while (cmpxchg8b(&head->align, old.align, new.align) != old.align);
+	DBGTRACE4("%p, %p", head, entry);
 	return entry;
 }
 
 #endif
 
-/* DEBUG macros */
-
-#define DBGTRACE(fmt, ...) do { } while (0)
-#define DBGTRACE1(fmt, ...) do { } while (0)
-#define DBGTRACE2(fmt, ...) do { } while (0)
-#define DBGTRACE3(fmt, ...) do { }  while (0)
-#define DBGTRACE4(fmt, ...) do { } while (0)
-#define DBGTRACE5(fmt, ...) do { } while (0)
-#define DBGTRACE6(fmt, ...) do { } while (0)
-
-/* for a block of code */
-#define DBG_BLOCK(level) while (0)
-
-extern int debug;
-
-#if defined DEBUG
-#undef DBGTRACE
-#define DBGTRACE(level, fmt, ...)				       \
-do {								       \
-	if (debug >= level)					       \
-		printk(KERN_INFO "%s (%s:%d): " fmt "\n", DRIVER_NAME, \
-		       __FUNCTION__, __LINE__ , ## __VA_ARGS__);       \
-} while (0)
-#undef DBG_BLOCK
-#define DBG_BLOCK(level) if (debug >= level)
-#endif
-
-#if defined(DEBUG) && DEBUG >= 1
-#undef DBGTRACE1
-#define DBGTRACE1(fmt, ...) DBGTRACE(1, fmt , ## __VA_ARGS__)
-#endif
-
-#if defined(DEBUG) && DEBUG >= 2
-#undef DBGTRACE2
-#define DBGTRACE2(fmt, ...) DBGTRACE(2, fmt , ## __VA_ARGS__)
-#endif
-
-#if defined(DEBUG) && DEBUG >= 3
-#undef DBGTRACE3
-#define DBGTRACE3(fmt, ...) DBGTRACE(3, fmt , ## __VA_ARGS__)
-#endif
-
-#if defined(DEBUG) && DEBUG >= 4
-#undef DBGTRACE4
-#define DBGTRACE4(fmt, ...) DBGTRACE(4, fmt , ## __VA_ARGS__)
-#endif
-
-#if defined(DEBUG) && DEBUG >= 5
-#undef DBGTRACE5
-#define DBGTRACE5(fmt, ...) DBGTRACE(5, fmt , ## __VA_ARGS__)
-#endif
-
-#if defined(DEBUG) && DEBUG >= 6
-#undef DBGTRACE6
-#define DBGTRACE6(fmt, ...) DBGTRACE(6, fmt , ## __VA_ARGS__)
-#endif
-
-#define TRACEENTER1(fmt, ...) DBGTRACE1("Enter " fmt , ## __VA_ARGS__)
-#define TRACEENTER2(fmt, ...) DBGTRACE2("Enter " fmt , ## __VA_ARGS__)
-#define TRACEENTER3(fmt, ...) DBGTRACE3("Enter " fmt , ## __VA_ARGS__)
-#define TRACEENTER4(fmt, ...) DBGTRACE4("Enter " fmt , ## __VA_ARGS__)
-#define TRACEENTER5(fmt, ...) DBGTRACE5("Enter " fmt , ## __VA_ARGS__)
-#define TRACEENTER6(fmt, ...) DBGTRACE6("Enter " fmt , ## __VA_ARGS__)
-
-#define TRACEEXIT1(stmt) do { DBGTRACE1("Exit"); stmt; } while(0)
-#define TRACEEXIT2(stmt) do { DBGTRACE2("Exit"); stmt; } while(0)
-#define TRACEEXIT3(stmt) do { DBGTRACE3("Exit"); stmt; } while(0)
-#define TRACEEXIT4(stmt) do { DBGTRACE4("Exit"); stmt; } while(0)
-#define TRACEEXIT5(stmt) do { DBGTRACE5("Exit"); stmt; } while(0)
-#define TRACEEXIT6(stmt) do { DBGTRACE6("Exit"); stmt; } while(0)
-
-#if defined(USB_DEBUG)
-#define USBTRACE DBGTRACE1
-#define USBENTER TRACEENTER1
-#define USBEXIT TRACEEXIT1
-#else
-#define USBTRACE(fmt, ...)
-#define USBENTER(fmt, ...)
-#define USBEXIT(stmt) stmt
-#endif
-
-#if defined(EVENT_DEBUG)
-#define EVENTTRACE DBGTRACE1
-#define EVENTENTER TRACEENTER1
-#define EVENTEXIT TRACEEXIT1
-#else
-#define EVENTTRACE(fmt, ...)
-#define EVENTENTER(fmt, ...)
-#define EVENTEXIT(stmt) stmt
-#endif
-
-#if defined(IO_DEBUG)
-#define IOTRACE DBGTRACE1
-#define IOENTER TRACEENTER1
-#define IOEXIT TRACEEXIT1
-#else
-#define IOTRACE(fmt, ...)
-#define IOENTER(fmt, ...)
-#define IOEXIT(stmt) stmt
-#endif
-
-#if defined(WORK_DEBUG)
-#define WORKTRACE DBGTRACE1
-#define WORKENTER TRACEENTER1
-#define WORKEXIT TRACEEXIT1
-#else
-#define WORKTRACE(fmt, ...)
-#define WORKENTER(fmt, ...)
-#define WORKEXIT(stmt) stmt
-#endif
-
-#if defined DEBUG
-#define assert(expr)							\
-do {									\
-	if (!(expr))							\
-		ERROR("assertion failed: %s", (#expr));			\
-} while (0)
-#else
-#define assert(expr) do { } while (0)
-#endif
-
-#if defined(IO_DEBUG)
-#define DUMP_IRP(__irp)							\
-do {									\
-	struct io_stack_location *_irp_sl;				\
-	_irp_sl = IoGetCurrentIrpStackLocation(__irp);			\
-	IOTRACE("irp: %p, stack size: %d, cl: %d, sl: %p, dev_obj: %p, " \
-		"mj_fn: %d, minor_fn: %d, nt_urb: %p, event: %p",	\
-		__irp, __irp->stack_count, (__irp)->current_location,	\
-		_irp_sl, _irp_sl->dev_obj, _irp_sl->major_fn,		\
-		_irp_sl->minor_fn, URB_FROM_IRP(__irp),			\
-		(__irp)->user_event);					\
-} while (0)
-#else
-#define DUMP_IRP(__irp) do { } while (0)
-#endif
-
 #define sleep_hz(n)					\
 do {							\
 	set_current_state(TASK_INTERRUPTIBLE);		\
