@@ -21,62 +21,21 @@
 
 extern NT_SPIN_LOCK ntoskernel_lock;
 extern NT_SPIN_LOCK irp_cancel_lock;
+extern struct nt_list object_list;
 
-wstdcall void WRAP_EXPORT(IoAcquireCancelSpinLock)
+wstdcall void WIN_FUNC(IoAcquireCancelSpinLock,1)
 	(KIRQL *irql)
 {
 	*irql = nt_spin_lock_irql(&irp_cancel_lock, DISPATCH_LEVEL);
 }
 
-wstdcall void WRAP_EXPORT(IoReleaseCancelSpinLock)
+wstdcall void WIN_FUNC(IoReleaseCancelSpinLock,1)
 	(KIRQL irql)
 {
 	nt_spin_unlock_irql(&irp_cancel_lock, irql);
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(IoGetDeviceProperty)
-	(struct device_object *pdo,
-	 enum device_registry_property dev_property,
-	 ULONG buffer_len, void *buffer, ULONG *result_len)
-{
-	struct ansi_string ansi;
-	struct unicode_string unicode;
-	struct wrap_device *wd;
-	ULONG need;
-
-	IOENTER("dev_obj = %p, dev_property = %d, buffer_len = %u, "
-		"buffer = %p, result_len = %p", pdo, dev_property,
-		buffer_len, buffer, result_len);
-
-	wd = pdo->reserved;
-	switch (dev_property) {
-	case DevicePropertyDeviceDescription:
-	case DevicePropertyFriendlyName:
-	case DevicePropertyDriverKeyName:
-		if (wrap_is_pci_bus(wd->dev_bus_type))
-			RtlInitAnsiString(&ansi, "PCI");
-		else // if (wrap_is_usb_bus(wd->dev_bus_type))
-			RtlInitAnsiString(&ansi, "USB");
-		need = sizeof(wchar_t) * (ansi.max_length + 1);
-		if (buffer_len < need) {
-			*result_len = need;
-			IOEXIT(return STATUS_BUFFER_TOO_SMALL);
-		}
-		unicode.max_length = buffer_len;
-		unicode.buf = buffer;
-		if (RtlAnsiStringToUnicodeString(&unicode, &ansi,
-						 FALSE) != STATUS_SUCCESS) {
-			*result_len = unicode.length;
-			IOEXIT(return STATUS_BUFFER_TOO_SMALL);
-		}
-		IOEXIT(return STATUS_SUCCESS);
-	default:
-		WARNING("%d not implemented", dev_property);
-		IOEXIT(return STATUS_INVALID_PARAMETER_2);
-	}
-}
-
-wstdcall int WRAP_EXPORT(IoIsWdmVersionAvailable)
+wstdcall int WIN_FUNC(IoIsWdmVersionAvailable,2)
 	(UCHAR major, UCHAR minor)
 {
 	IOENTER("%d, %x", major, minor);
@@ -88,7 +47,7 @@ wstdcall int WRAP_EXPORT(IoIsWdmVersionAvailable)
 	IOEXIT(return 0);
 }
 
-wstdcall BOOLEAN WRAP_EXPORT(IoIs32bitProcess)
+wstdcall BOOLEAN WIN_FUNC(IoIs32bitProcess,1)
 	(struct irp *irp)
 {
 #ifdef CONFIG_X86_64
@@ -98,7 +57,7 @@ wstdcall BOOLEAN WRAP_EXPORT(IoIs32bitProcess)
 #endif
 }
 
-wstdcall void WRAP_EXPORT(IoInitializeIrp)
+wstdcall void WIN_FUNC(IoInitializeIrp,3)
 	(struct irp *irp, USHORT size, CCHAR stack_count)
 {
 	IOENTER("irp: %p, count: %d", irp, stack_count);
@@ -113,7 +72,7 @@ wstdcall void WRAP_EXPORT(IoInitializeIrp)
 	IOEXIT(return);
 }
 
-wstdcall void WRAP_EXPORT(IoReuseIrp)
+wstdcall void WIN_FUNC(IoReuseIrp,2)
 	(struct irp *irp, NTSTATUS status)
 {
 	IOENTER("%p, %d", irp, status);
@@ -128,7 +87,7 @@ wstdcall void WRAP_EXPORT(IoReuseIrp)
 	IOEXIT(return);
 }
 
-wstdcall struct irp *WRAP_EXPORT(IoAllocateIrp)
+wstdcall struct irp *WIN_FUNC(IoAllocateIrp,2)
 	(char stack_count, BOOLEAN charge_quota)
 {
 	struct irp *irp;
@@ -139,7 +98,7 @@ wstdcall struct irp *WRAP_EXPORT(IoAllocateIrp)
 	 * need to allocate space for it so that driver can set major
 	 * function etc. even if stack_count is 0 */
 	irp_size = IoSizeOfIrp(stack_count + 1);
-	irp = kmalloc(irp_size, GFP_ATOMIC);
+	irp = kmalloc(irp_size, gfp_irql());
 	if (irp) {
 		IOTRACE("allocated irp %p", irp);
 		IoInitializeIrp(irp, irp_size, stack_count);
@@ -147,14 +106,13 @@ wstdcall struct irp *WRAP_EXPORT(IoAllocateIrp)
 	IOEXIT(return irp);
 }
 
-wstdcall BOOLEAN WRAP_EXPORT(IoCancelIrp)
+wstdcall BOOLEAN WIN_FUNC(IoCancelIrp,1)
 	(struct irp *irp)
 {
 	typeof(irp->cancel_routine) cancel_routine;
 
 	/* NB: this function may be called at DISPATCH_LEVEL */
-	IOENTER("irp = %p", irp);
-
+	IOTRACE("irp: %p", irp);
 	if (!irp)
 		return FALSE;
 	DUMP_IRP(irp);
@@ -170,11 +128,7 @@ wstdcall BOOLEAN WRAP_EXPORT(IoCancelIrp)
 		LIN2WIN2(cancel_routine, irp_sl->dev_obj, irp);
 		/* in usb's cancel, irp->cancel is set to indicate
 		 * status of cancel */
-		if (irp->cancel == FALSE) {
-			irp->cancel = TRUE;
-			IOEXIT(return FALSE);
-		} else
-			IOEXIT(return TRUE);
+		IOEXIT(return xchg(&irp->cancel, TRUE));
 	} else {
 		IOTRACE("irp %p already canceled", irp);
 		IoReleaseCancelSpinLock(irp->cancel_irql);
@@ -212,7 +166,7 @@ wstdcall void IoDequeueThreadIrp(struct irp *irp)
 	}
 }
 
-wstdcall void WRAP_EXPORT(IoFreeIrp)
+wstdcall void WIN_FUNC(IoFreeIrp,1)
 	(struct irp *irp)
 {
 	IOENTER("irp = %p", irp);
@@ -223,7 +177,7 @@ wstdcall void WRAP_EXPORT(IoFreeIrp)
 	IOEXIT(return);
 }
 
-wstdcall struct irp *WRAP_EXPORT(IoBuildAsynchronousFsdRequest)
+wstdcall struct irp *WIN_FUNC(IoBuildAsynchronousFsdRequest,6)
 	(ULONG major_fn, struct device_object *dev_obj, void *buffer,
 	 ULONG length, LARGE_INTEGER *offset,
 	 struct io_status_block *user_status)
@@ -278,7 +232,7 @@ wstdcall struct irp *WRAP_EXPORT(IoBuildAsynchronousFsdRequest)
 	return irp;
 }
 
-wstdcall struct irp *WRAP_EXPORT(IoBuildSynchronousFsdRequest)
+wstdcall struct irp *WIN_FUNC(IoBuildSynchronousFsdRequest,7)
 	(ULONG major_fn, struct device_object *dev_obj, void *buf,
 	 ULONG length, LARGE_INTEGER *offset, struct nt_event *event,
 	 struct io_status_block *user_status)
@@ -294,7 +248,7 @@ wstdcall struct irp *WRAP_EXPORT(IoBuildSynchronousFsdRequest)
 	return irp;
 }
 
-wstdcall struct irp *WRAP_EXPORT(IoBuildDeviceIoControlRequest)
+wstdcall struct irp *WIN_FUNC(IoBuildDeviceIoControlRequest,9)
 	(ULONG ioctl, struct device_object *dev_obj,
 	 void *input_buf, ULONG input_buf_len, void *output_buf,
 	 ULONG output_buf_len, BOOLEAN internal_ioctl,
@@ -302,8 +256,9 @@ wstdcall struct irp *WRAP_EXPORT(IoBuildDeviceIoControlRequest)
 {
 	struct irp *irp;
 	struct io_stack_location *irp_sl;
+	ULONG buf_len;
 
-	IOENTER("%p", dev_obj);
+	IOENTER("%p, 0x%08x, %d", dev_obj, ioctl, internal_ioctl);
 	if (!dev_obj)
 		IOEXIT(return NULL);
 	irp = IoAllocateIrp(dev_obj->stack_count, FALSE);
@@ -311,29 +266,83 @@ wstdcall struct irp *WRAP_EXPORT(IoBuildDeviceIoControlRequest)
 		WARNING("couldn't allocate irp");
 		return NULL;
 	}
-	irp->user_status = io_status;
-	irp->user_event = event;
-	irp->user_buf = output_buf;
-	irp->associated_irp.system_buffer = input_buf;
-
 	irp_sl = IoGetNextIrpStackLocation(irp);
 	irp_sl->params.dev_ioctl.code = ioctl;
 	irp_sl->params.dev_ioctl.input_buf_len = input_buf_len;
 	irp_sl->params.dev_ioctl.output_buf_len = output_buf_len;
-	irp_sl->dev_obj = dev_obj;
 	irp_sl->major_fn = (internal_ioctl) ?
 		IRP_MJ_INTERNAL_DEVICE_CONTROL : IRP_MJ_DEVICE_CONTROL;
-	irp_sl->minor_fn = 0;
-	irp_sl->flags = 0;
-	irp_sl->file_obj = NULL;
-	irp_sl->completion_routine = NULL;
+	IOTRACE("%d", IO_METHOD_FROM_CTL_CODE(ioctl));
+
+	switch (IO_METHOD_FROM_CTL_CODE(ioctl)) {
+	case METHOD_BUFFERED:
+		buf_len = max(input_buf_len, output_buf_len);
+		if (buf_len) {
+			irp->associated_irp.system_buffer =
+				ExAllocatePoolWithTag(NonPagedPool, buf_len, 0);
+			if (!irp->associated_irp.system_buffer) {
+				IoFreeIrp(irp);
+				IOEXIT(return NULL);
+			}
+			irp->associated_irp.system_buffer = input_buf;
+			if (input_buf)
+				memcpy(irp->associated_irp.system_buffer,
+				       input_buf, input_buf_len);
+			irp->flags = IRP_BUFFERED_IO | IRP_DEALLOCATE_BUFFER;
+			if (output_buf)
+				irp->flags |= IRP_INPUT_OPERATION;
+			irp->user_buf = output_buf;
+		} else { // buf_len = 0
+			irp->flags = 0;
+			irp->user_buf = NULL;
+		}
+		break;
+	case METHOD_IN_DIRECT:
+	case METHOD_OUT_DIRECT:
+		if (input_buf) {
+			irp->associated_irp.system_buffer =
+				ExAllocatePoolWithTag(NonPagedPool,
+						      input_buf_len, 0);
+			if (!irp->associated_irp.system_buffer) {
+				IoFreeIrp(irp);
+				IOEXIT(return NULL);
+			}
+			memcpy(irp->associated_irp.system_buffer,
+			       input_buf, input_buf_len);
+			irp->flags = IRP_BUFFERED_IO | IRP_DEALLOCATE_BUFFER;
+		} else
+			irp->flags = 0;
+		/* TODO: we are supposed to setup MDL, but USB layer
+		 * doesn't use MDLs. Moreover, USB layer mirrors
+		 * non-DMAable buffers, so no need to allocate
+		 * DMAable buffer here */
+		if (output_buf) {
+			irp->associated_irp.system_buffer =
+				ExAllocatePoolWithTag(NonPagedPool,
+						      output_buf_len, 0);
+			if (!irp->associated_irp.system_buffer) {
+				IoFreeIrp(irp);
+				IOEXIT(return NULL);
+			}
+			irp->flags = IRP_BUFFERED_IO | IRP_DEALLOCATE_BUFFER;
+		} else
+			irp->flags = 0;
+		break;
+	case METHOD_NEITHER:
+		irp->user_buf = output_buf;
+		irp_sl->params.dev_ioctl.type3_input_buf = input_buf;
+		break;
+	}
+
+	irp->user_status = io_status;
+	irp->user_event = event;
 	IoQueueThreadIrp(irp);
 
 	IOTRACE("irp: %p", irp);
 	IOEXIT(return irp);
 }
 
-wfastcall NTSTATUS WRAP_EXPORT(IofCallDriver)
+wfastcall NTSTATUS WIN_FUNC(IofCallDriver,2)
 	(struct device_object *dev_obj, struct irp *irp)
 {
 	struct io_stack_location *irp_sl;
@@ -360,7 +369,7 @@ wfastcall NTSTATUS WRAP_EXPORT(IofCallDriver)
 	IOEXIT(return status);
 }
 
-wfastcall void WRAP_EXPORT(IofCompleteRequest)
+wfastcall void WIN_FUNC(IofCompleteRequest,2)
 	(struct irp *irp, CHAR prio_boost)
 {
 	NTSTATUS status;
@@ -445,47 +454,32 @@ wfastcall void WRAP_EXPORT(IofCompleteRequest)
 	IOEXIT(return);
 }
 
-/* IRP functions are called as Windows functions. For 64-bit, this
- * means arguments are in rcx, rdx etc., but Linux functions expect
- * them in rdi, rsi etc. So we need to put arguments back correctly
- * before touching arguments. We also assume that all arguments are
- * pointers (or register width). */
-
 wstdcall NTSTATUS IoPassIrpDown(struct device_object *dev_obj,
 			       struct irp *irp)
 {
-	WIN2LIN2(dev_obj, irp);
-
 	IoSkipCurrentIrpStackLocation(irp);
 	IOEXIT(return IoCallDriver(dev_obj, irp));
 }
 
-
-/* called as Windows function, so call WIN2LIN3 before accessing
- * arguments */
 wstdcall NTSTATUS IoIrpSyncComplete(struct device_object *dev_obj,
-				   struct irp *irp, void *context)
+				    struct irp *irp, void *context)
 {
-	WIN2LIN3(dev_obj, irp, context);
-
 	if (irp->pending_returned == TRUE)
 		KeSetEvent(context, IO_NO_INCREMENT, FALSE);
 	IOEXIT(return STATUS_MORE_PROCESSING_REQUIRED);
 }
+WIN_FUNC_DECL(IoIrpSyncComplete,3)
 
-/* called as Windows function, so call WIN2LIN2 before accessing
- * arguments */
 wstdcall NTSTATUS IoSyncForwardIrp(struct device_object *dev_obj,
 				  struct irp *irp)
 {
 	struct nt_event event;
 	NTSTATUS status;
 
-	WIN2LIN2(dev_obj, irp);
-
 	IoCopyCurrentIrpStackLocationToNext(irp);
 	KeInitializeEvent(&event, SynchronizationEvent, FALSE);
-	IoSetCompletionRoutine(irp, IoIrpSyncComplete, &event,
+	/* completion function is called as Windows function */
+	IoSetCompletionRoutine(irp, WIN_FUNC_PTR(IoIrpSyncComplete,3), &event,
 			       TRUE, TRUE, TRUE);
 	status = IoCallDriver(dev_obj, irp);
 	IOTRACE("%08X", status);
@@ -497,31 +491,25 @@ wstdcall NTSTATUS IoSyncForwardIrp(struct device_object *dev_obj,
 	IOTRACE("%08X", status);
 	IOEXIT(return status);
 }
+WIN_FUNC_DECL(IoSyncForwardIrp,2)
 
-/* called as Windows function, so call WIN2LIN2 before accessing
- * arguments */
 wstdcall NTSTATUS IoAsyncForwardIrp(struct device_object *dev_obj,
 				   struct irp *irp)
 {
 	NTSTATUS status;
 
-	WIN2LIN2(dev_obj, irp);
-
 	IoCopyCurrentIrpStackLocationToNext(irp);
 	status = IoCallDriver(dev_obj, irp);
 	IOEXIT(return status);
 }
+WIN_FUNC_DECL(IoAsyncForwardIrp,2)
 
-/* called as Windows function, so call WIN2LIN2 before accessing
- * arguments */
 wstdcall NTSTATUS IoInvalidDeviceRequest(struct device_object *dev_obj,
 					struct irp *irp)
 {
 	struct io_stack_location *irp_sl;
 	NTSTATUS status;
 
-	WIN2LIN2(dev_obj, irp);
-
 	irp_sl = IoGetCurrentIrpStackLocation(irp);
 	WARNING("%d:%d not implemented", irp_sl->major_fn, irp_sl->minor_fn);
 	irp->io_status.status = STATUS_SUCCESS;
@@ -530,6 +518,7 @@ wstdcall NTSTATUS IoInvalidDeviceRequest(struct device_object *dev_obj,
 	IoCompleteRequest(irp, IO_NO_INCREMENT);
 	IOEXIT(return status);
 }
+WIN_FUNC_DECL(IoInvalidDeviceRequest,2)
 
 static irqreturn_t io_irq_th(int irq, void *data, struct pt_regs *pt_regs)
 {
@@ -552,7 +541,7 @@ static irqreturn_t io_irq_th(int irq, void *data, struct pt_regs *pt_regs)
 		return IRQ_NONE;
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(IoConnectInterrupt)
+wstdcall NTSTATUS WIN_FUNC(IoConnectInterrupt,11)
 	(struct kinterrupt *interrupt, PKSERVICE_ROUTINE service_routine,
 	 void *service_context, NT_SPIN_LOCK *lock, ULONG vector,
 	 KIRQL irql, KIRQL synch_irql, enum kinterrupt_mode interrupt_mode,
@@ -580,13 +569,13 @@ wstdcall NTSTATUS WRAP_EXPORT(IoConnectInterrupt)
 	IOEXIT(return STATUS_SUCCESS);
 }
 
-wstdcall void WRAP_EXPORT(IoDisconnectInterrupt)
+wstdcall void WIN_FUNC(IoDisconnectInterrupt,1)
 	(struct kinterrupt *interrupt)
 {
 	free_irq(interrupt->vector, interrupt);
 }
 
-wstdcall struct mdl *WRAP_EXPORT(IoAllocateMdl)
+wstdcall struct mdl *WIN_FUNC(IoAllocateMdl,5)
 	(void *virt, ULONG length, BOOLEAN second_buf, BOOLEAN charge_quota,
 	 struct irp *irp)
 {
@@ -608,55 +597,53 @@ wstdcall struct mdl *WRAP_EXPORT(IoAllocateMdl)
 	return mdl;
 }
 
-wstdcall void WRAP_EXPORT(IoFreeMdl)
+wstdcall void WIN_FUNC(IoFreeMdl,1)
 	(struct mdl *mdl)
 {
 	free_mdl(mdl);
 	IOEXIT(return);
 }
 
-wstdcall struct io_workitem *WRAP_EXPORT(IoAllocateWorkItem)
+wstdcall struct io_workitem *WIN_FUNC(IoAllocateWorkItem,1)
 	(struct device_object *dev_obj)
 {
 	struct io_workitem *io_workitem;
 
 	IOENTER("%p", dev_obj);
-	io_workitem = kmalloc(sizeof(*io_workitem), GFP_ATOMIC);
+	io_workitem = kmalloc(sizeof(*io_workitem), gfp_irql());
 	if (!io_workitem)
 		IOEXIT(return NULL);
 	io_workitem->dev_obj = dev_obj;
 	IOEXIT(return io_workitem);
 }
 
-wstdcall void WRAP_EXPORT(IoFreeWorkItem)
+wstdcall void WIN_FUNC(IoFreeWorkItem,1)
 	(struct io_workitem *io_workitem)
 {
 	kfree(io_workitem);
 	IOEXIT(return);
 }
 
-wstdcall void WRAP_EXPORT(IoQueueWorkItem)
+wstdcall void WIN_FUNC(IoQueueWorkItem,4)
 	(struct io_workitem *io_workitem, void *func,
 	 enum work_queue_type queue_type, void *context)
 {
 	IOENTER("%p, %p", io_workitem, io_workitem->dev_obj);
 	io_workitem->worker_routine = func;
 	io_workitem->context = context;
-	schedule_ntos_work_item(func, io_workitem->dev_obj, context,
-				WORKER_FUNC_WIN);
+	schedule_ntos_work_item(func, io_workitem->dev_obj, context);
 	IOEXIT(return);
 }
 
-wstdcall void WRAP_EXPORT(ExQueueWorkItem)
+wstdcall void WIN_FUNC(ExQueueWorkItem,2)
 	(struct io_workitem *io_workitem, enum work_queue_type queue_type)
 {
 	IOENTER("%p", io_workitem);
 	schedule_ntos_work_item(io_workitem->worker_routine,
-				io_workitem->dev_obj, io_workitem->context,
-				WORKER_FUNC_WIN);
+				io_workitem->dev_obj, io_workitem->context);
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(IoAllocateDriverObjectExtension)
+wstdcall NTSTATUS WIN_FUNC(IoAllocateDriverObjectExtension,4)
 	(struct driver_object *drv_obj, void *client_id, ULONG extlen,
 	 void **ext)
 {
@@ -664,7 +651,7 @@ wstdcall NTSTATUS WRAP_EXPORT(IoAllocateDriverObjectExtension)
 	KIRQL irql;
 
 	IOENTER("%p, %p", drv_obj, client_id);
-	ce = kmalloc(sizeof(*ce) + extlen, GFP_ATOMIC);
+	ce = kmalloc(sizeof(*ce) + extlen, gfp_irql());
 	if (ce == NULL)
 		return STATUS_INSUFFICIENT_RESOURCES;
 
@@ -679,7 +666,7 @@ wstdcall NTSTATUS WRAP_EXPORT(IoAllocateDriverObjectExtension)
 	IOEXIT(return STATUS_SUCCESS);
 }
 
-wstdcall void *WRAP_EXPORT(IoGetDriverObjectExtension)
+wstdcall void *WIN_FUNC(IoGetDriverObjectExtension,2)
 	(struct driver_object *drv_obj, void *client_id)
 {
 	struct custom_ext *ce;
@@ -713,7 +700,7 @@ void free_custom_extensions(struct driver_extension *drv_ext)
 	IOEXIT(return);
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(IoCreateDevice)
+wstdcall NTSTATUS WIN_FUNC(IoCreateDevice,7)
 	(struct driver_object *drv_obj, ULONG dev_ext_length,
 	 struct unicode_string *dev_name, DEVICE_TYPE dev_type,
 	 ULONG dev_chars, BOOLEAN exclusive, struct device_object **newdev)
@@ -721,17 +708,11 @@ wstdcall NTSTATUS WRAP_EXPORT(IoCreateDevice)
 	struct device_object *dev;
 	struct dev_obj_ext *dev_obj_ext;
 	int size;
-	struct ansi_string ansi;
 
 	IOENTER("%p, %u, %p", drv_obj, dev_ext_length, dev_name);
-	if (dev_name && (RtlUnicodeStringToAnsiString(&ansi, dev_name, TRUE)
-			 == STATUS_SUCCESS)) {
-		IOTRACE("dev_name: %s", ansi.buf);
-		RtlFreeAnsiString(&ansi);
-	}
 
 	size = sizeof(*dev) + dev_ext_length + sizeof(*dev_obj_ext);
-	dev = allocate_object(size, OBJECT_TYPE_DEVICE, ansi.buf);
+	dev = allocate_object(size, OBJECT_TYPE_DEVICE, dev_name);
 	if (!dev)
 		IOEXIT(return STATUS_INSUFFICIENT_RESOURCES);
 	if (dev_ext_length)
@@ -767,7 +748,7 @@ wstdcall NTSTATUS WRAP_EXPORT(IoCreateDevice)
 	IOEXIT(return STATUS_SUCCESS);
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(IoCreateUnprotectedSymbolicLink)
+wstdcall NTSTATUS WIN_FUNC(IoCreateUnprotectedSymbolicLink,2)
 	(struct unicode_string *link, struct unicode_string *dev_name)
 {
 	struct ansi_string ansi;
@@ -783,17 +764,17 @@ wstdcall NTSTATUS WRAP_EXPORT(IoCreateUnprotectedSymbolicLink)
 		IOTRACE("link: %s", ansi.buf);
 		RtlFreeAnsiString(&ansi);
 	}
-//	UNIMPL();
+//	TODO();
 	IOEXIT(return STATUS_SUCCESS);
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(IoCreateSymbolicLink)
+wstdcall NTSTATUS WIN_FUNC(IoCreateSymbolicLink,2)
 	(struct unicode_string *link, struct unicode_string *dev_name)
 {
 	IOEXIT(return IoCreateUnprotectedSymbolicLink(link, dev_name));
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(IoDeleteSymbolicLink)
+wstdcall NTSTATUS WIN_FUNC(IoDeleteSymbolicLink,1)
 	(struct unicode_string *link)
 {
 	struct ansi_string ansi;
@@ -807,7 +788,7 @@ wstdcall NTSTATUS WRAP_EXPORT(IoDeleteSymbolicLink)
 	IOEXIT(return STATUS_SUCCESS);
 }
 
-wstdcall void WRAP_EXPORT(IoDeleteDevice)
+wstdcall void WIN_FUNC(IoDeleteDevice,1)
 	(struct device_object *dev)
 {
 	IOENTER("%p", dev);
@@ -831,7 +812,7 @@ wstdcall void WRAP_EXPORT(IoDeleteDevice)
 	IOEXIT(return);
 }
 
-wstdcall void WRAP_EXPORT(IoDetachDevice)
+wstdcall void WIN_FUNC(IoDetachDevice,1)
 	(struct device_object *topdev)
 {
 	struct device_object *tail;
@@ -856,7 +837,7 @@ wstdcall void WRAP_EXPORT(IoDetachDevice)
 	IOEXIT(return);
 }
 
-wstdcall struct device_object *WRAP_EXPORT(IoGetAttachedDevice)
+wstdcall struct device_object *WIN_FUNC(IoGetAttachedDevice,1)
 	(struct device_object *dev)
 {
 	struct device_object *top_dev;
@@ -873,7 +854,7 @@ wstdcall struct device_object *WRAP_EXPORT(IoGetAttachedDevice)
 	IOEXIT(return top_dev);
 }
 
-wstdcall struct device_object *WRAP_EXPORT(IoGetAttachedDeviceReference)
+wstdcall struct device_object *WIN_FUNC(IoGetAttachedDeviceReference,1)
 	(struct device_object *dev)
 {
 	struct device_object *top_dev;
@@ -886,7 +867,7 @@ wstdcall struct device_object *WRAP_EXPORT(IoGetAttachedDeviceReference)
 	IOEXIT(return top_dev);
 }
 
-wstdcall struct device_object *WRAP_EXPORT(IoAttachDeviceToDeviceStack)
+wstdcall struct device_object *WIN_FUNC(IoAttachDeviceToDeviceStack,2)
 	(struct device_object *src, struct device_object *tgt)
 {
 	struct device_object *attached;
@@ -910,23 +891,91 @@ wstdcall struct device_object *WRAP_EXPORT(IoAttachDeviceToDeviceStack)
 	IOEXIT(return attached);
 }
 
+wstdcall NTSTATUS WIN_FUNC(IoGetDeviceProperty,5)
+	(struct device_object *pdo, enum device_registry_property dev_property,
+	 ULONG buffer_len, void *buffer, ULONG *result_len)
+{
+	struct ansi_string ansi;
+	struct unicode_string unicode;
+	struct wrap_device *wd;
+	ULONG need;
+
+	IOENTER("dev_obj = %p, dev_property = %d, buffer_len = %u, "
+		"buffer = %p, result_len = %p", pdo, dev_property,
+		buffer_len, buffer, result_len);
+
+	wd = pdo->reserved;
+	switch (dev_property) {
+	case DevicePropertyDeviceDescription:
+	case DevicePropertyFriendlyName:
+	case DevicePropertyDriverKeyName:
+		if (wrap_is_pci_bus(wd->dev_bus_type))
+			RtlInitAnsiString(&ansi, "PCI");
+		else // if (wrap_is_usb_bus(wd->dev_bus_type))
+			RtlInitAnsiString(&ansi, "USB");
+		need = sizeof(wchar_t) * (ansi.max_length + 1);
+		if (buffer_len < need) {
+			*result_len = need;
+			IOEXIT(return STATUS_BUFFER_TOO_SMALL);
+		}
+		unicode.max_length = buffer_len;
+		unicode.buf = buffer;
+		if (RtlAnsiStringToUnicodeString(&unicode, &ansi,
+						 FALSE) != STATUS_SUCCESS) {
+			*result_len = unicode.length;
+			IOEXIT(return STATUS_BUFFER_TOO_SMALL);
+		}
+		IOEXIT(return STATUS_SUCCESS);
+	default:
+		WARNING("%d not implemented", dev_property);
+		IOEXIT(return STATUS_INVALID_PARAMETER_2);
+	}
+}
+
+wstdcall NTSTATUS WIN_FUNC(IoGetDeviceObjectPointer,4)
+	(struct unicode_string *name, ACCESS_MASK desired_access,
+	 void *file_obj, struct device_object *dev_obj)
+{
+	KIRQL irql;
+	struct common_object_header *coh;
+
+	dev_obj = NULL;
+	/* TODO: access is not checked and file_obj is set to NULL */
+	file_obj = NULL;
+	irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
+	nt_list_for_each_entry(coh, &object_list, list) {
+		DBGTRACE5("header: %p, type: %d", coh, coh->type);
+		if (coh->type != OBJECT_TYPE_DEVICE)
+			continue;
+		if (!RtlCompareUnicodeString(&coh->name, name, TRUE)) {
+			dev_obj = HEADER_TO_OBJECT(coh);
+			DBGTRACE5("dev_obj: %p", dev_obj);
+			break;
+		}
+	}
+	nt_spin_unlock_irql(&ntoskernel_lock, irql);
+	if (dev_obj)
+		IOEXIT(return STATUS_SUCCESS);
+	else
+		IOEXIT(return STATUS_OBJECT_NAME_INVALID);
+}
 
 /* NOTE: Make sure to compile with -freg-struct-return, so gcc will
  * return union in register, like Windows */
-wstdcall union power_state WRAP_EXPORT(PoSetPowerState)
+wstdcall union power_state WIN_FUNC(PoSetPowerState,3)
 	(struct device_object *dev_obj, enum power_state_type type,
 	 union power_state state)
 {
 	IOEXIT(return state);
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(PoCallDriver)
+wstdcall NTSTATUS WIN_FUNC(PoCallDriver,2)
 	(struct device_object *dev_obj, struct irp *irp)
 {
 	return IoCallDriver(dev_obj, irp);
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(PoRequestPowerIrp)
+wstdcall NTSTATUS WIN_FUNC(PoRequestPowerIrp,6)
 	(struct device_object *dev_obj, UCHAR minor_fn,
 	 union power_state power_state, void *completion_func,
 	 void *context, struct irp **pirp)
@@ -953,36 +1002,63 @@ wstdcall NTSTATUS WRAP_EXPORT(PoRequestPowerIrp)
 	return PoCallDriver(dev_obj, irp);
 }
 
-wstdcall void WRAP_EXPORT(PoStartNextPowerIrp)
+wstdcall void WIN_FUNC(PoStartNextPowerIrp,1)
 	(struct irp *irp)
 {
 	IOENTER("irp = %p", irp);
 	IOEXIT(return);
 }
 
-wstdcall void WRAP_EXPORT(IoInitializeRemoveLockEx)
+wstdcall void WIN_FUNC(IoInitializeRemoveLockEx,5)
 	(struct io_remove_lock *lock, ULONG alloc_tag, ULONG max_locked_min,
 	 ULONG high_mark, ULONG lock_size)
 {
-	UNIMPL();
+	TODO();
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(IoAcquireRemoveLockEx)
+wstdcall void *WIN_FUNC(IoAllocateErrorLogEntry,2)
+	(void *io_object, UCHAR entry_size)
+{
+	/* not implemented fully */
+	void *ret = kmalloc(sizeof(struct io_error_log_packet) + entry_size,
+			    gfp_irql());
+	DBGTRACE2("%p", ret);
+	if (ret)
+		return ret + sizeof(struct io_error_log_packet);
+	else
+		return NULL;
+}
+
+wstdcall void WIN_FUNC(IoWriteErrorLogEntry,1)
+	(void *entry)
+{
+	/* TODO: log error with codes and message */
+	ERROR("");
+}
+
+wstdcall void WIN_FUNC(IoFreeErrorLogEntry,1)
+	(void *entry)
+{
+	DBGTRACE2("%p", entry);
+	kfree(entry - sizeof(struct io_error_log_packet));
+}
+
+wstdcall NTSTATUS WIN_FUNC(IoAcquireRemoveLockEx,5)
 	(struct io_remove_lock lock, void *tag, char *file, ULONG line,
 	 ULONG lock_size)
 {
-	UNIMPL();
+	TODO();
 	IOEXIT(return STATUS_SUCCESS);
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(IoReleaseRemoveLockEx)
+wstdcall NTSTATUS WIN_FUNC(IoReleaseRemoveLockEx,3)
 	(struct io_remove_lock lock, void *tag, ULONG lock_size)
 {
-	UNIMPL();
+	TODO();
 	IOEXIT(return STATUS_SUCCESS);
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(IoRegisterDeviceInterface)
+wstdcall NTSTATUS WIN_FUNC(IoRegisterDeviceInterface,4)
 	(struct device_object *pdo, struct guid *guid_class,
 	 struct unicode_string *reference, struct unicode_string *link)
 {
@@ -996,14 +1072,14 @@ wstdcall NTSTATUS WRAP_EXPORT(IoRegisterDeviceInterface)
 	return RtlAnsiStringToUnicodeString(link, &ansi, TRUE);
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(IoSetDeviceInterfaceState)
+wstdcall NTSTATUS WIN_FUNC(IoSetDeviceInterfaceState,2)
 	(struct unicode_string *link, BOOLEAN enable)
 {
 	TRACEENTER1("link: %p, enable: %d", link, enable);
 	return STATUS_SUCCESS;
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(IoOpenDeviceRegistryKey)
+wstdcall NTSTATUS WIN_FUNC(IoOpenDeviceRegistryKey,4)
 	(struct device_object *dev_obj, ULONG type, ACCESS_MASK mask,
 	 void **handle)
 {
@@ -1012,25 +1088,25 @@ wstdcall NTSTATUS WRAP_EXPORT(IoOpenDeviceRegistryKey)
 	return STATUS_SUCCESS;
 }
 
-wstdcall NTSTATUS WRAP_EXPORT(IoWMIRegistrationControl)
+wstdcall NTSTATUS WIN_FUNC(IoWMIRegistrationControl,2)
 	(struct device_object *dev_obj, ULONG action)
 {
 	TRACEENTER2("%p, %d", dev_obj, action);
 	TRACEEXIT2(return STATUS_SUCCESS);
 }
 
-wstdcall void WRAP_EXPORT(IoInvalidateDeviceRelations)
+wstdcall void WIN_FUNC(IoInvalidateDeviceRelations,2)
 	(struct device_object *dev_obj, enum device_relation_type type)
 {
 	INFO("%p, %d", dev_obj, type);
-	UNIMPL();
+	TODO();
 }
 
-wstdcall void WRAP_EXPORT(IoInvalidateDeviceState)
+wstdcall void WIN_FUNC(IoInvalidateDeviceState,1)
 	(struct device_object *pdo)
 {
 	INFO("%p", pdo);
-	UNIMPL();
+	TODO();
 }
 
 #include "ntoskernel_io_exports.h"
