@@ -219,16 +219,12 @@ wstdcall NTSTATUS IoSendIrpTopDev(struct device_object *dev_obj, ULONG major_fn,
 	return status;
 }
 
-/* called as Windows function, so call WIN2LIN2 before accessing
- * arguments */
 wstdcall NTSTATUS pdoDispatchDeviceControl(struct device_object *pdo,
 					  struct  irp *irp)
 {
 	struct io_stack_location *irp_sl;
 	NTSTATUS status;
 
-	WIN2LIN2(pdo, irp);
-
 	DUMP_IRP(irp);
 	irp_sl = IoGetCurrentIrpStackLocation(irp);
 #ifdef CONFIG_USB
@@ -242,11 +238,9 @@ wstdcall NTSTATUS pdoDispatchDeviceControl(struct device_object *pdo,
 #endif
 	IOEXIT(return status);
 }
+WIN_FUNC_DECL(pdoDispatchDeviceControl,2)
 
-/* called as Windows function, so call WIN2LIN2 before accessing
- * arguments */
-wstdcall NTSTATUS pdoDispatchPnp(struct device_object *pdo,
-				struct irp *irp)
+wstdcall NTSTATUS pdoDispatchPnp(struct device_object *pdo, struct irp *irp)
 {
 	struct io_stack_location *irp_sl;
 	struct wrap_device *wd;
@@ -255,11 +249,9 @@ wstdcall NTSTATUS pdoDispatchPnp(struct device_object *pdo,
 	struct usbd_bus_interface_usbdi *usb_intf;
 #endif
 
-	WIN2LIN2(pdo, irp);
-
 	irp_sl = IoGetCurrentIrpStackLocation(irp);
+	DBGTRACE2("%p %d:%d", pdo, irp_sl->major_fn, irp_sl->minor_fn);
 	wd = pdo->reserved;
-	DBGTRACE2("fn %d:%d, wd: %p", irp_sl->major_fn, irp_sl->minor_fn, wd);
 	switch (irp_sl->minor_fn) {
 	case IRP_MN_START_DEVICE:
 		status = start_pdo(pdo);
@@ -315,10 +307,9 @@ wstdcall NTSTATUS pdoDispatchPnp(struct device_object *pdo,
 	IoCompleteRequest(irp, IO_NO_INCREMENT);
 	IOEXIT(return status);
 }
+WIN_FUNC_DECL(pdoDispatchPnp,2)
 
-/* called as Windows function */
-wstdcall NTSTATUS pdoDispatchPower(struct device_object *pdo,
-				  struct irp *irp)
+wstdcall NTSTATUS pdoDispatchPower(struct device_object *pdo, struct irp *irp)
 {
 	struct io_stack_location *irp_sl;
 	struct wrap_device *wd;
@@ -326,8 +317,6 @@ wstdcall NTSTATUS pdoDispatchPower(struct device_object *pdo,
 	struct pci_dev *pdev;
 	NTSTATUS status;
 
-	WIN2LIN2(pdo, irp);
-
 	irp_sl = IoGetCurrentIrpStackLocation(irp);
 	wd = pdo->reserved;
 	DBGTRACE2("pdo: %p, fn: %d:%d, wd: %p",
@@ -386,6 +375,7 @@ wstdcall NTSTATUS pdoDispatchPower(struct device_object *pdo,
 	IoCompleteRequest(irp, IO_NO_INCREMENT);
 	return status;
 }
+WIN_FUNC_DECL(pdoDispatchPower,2)
 
 NTSTATUS pnp_set_device_power_state(struct wrap_device *wd,
 				    enum device_power_state state)
@@ -489,7 +479,7 @@ NTSTATUS pnp_remove_device(struct wrap_device *wd)
 		DBGTRACE1("unloading driver: %p", fdo_drv_obj);
 		wrap_driver =
 			IoGetDriverObjectExtension(fdo_drv_obj,
-					   (void *)CE_WRAP_DRIVER_CLIENT_ID);
+					   (void *)WRAP_DRIVER_CLIENT_ID);
 		if (fdo_drv_obj->unload)
 			LIN2WIN1(fdo_drv_obj->unload, fdo_drv_obj);
 		if (wrap_driver) {
@@ -504,25 +494,45 @@ NTSTATUS pnp_remove_device(struct wrap_device *wd)
 	TRACEEXIT1(return status);
 }
 
+WIN_FUNC_DECL(IoInvalidDeviceRequest,2)
+
 static struct device_object *alloc_pdo(struct driver_object *drv_obj)
 {
 	struct device_object *pdo;
 	NTSTATUS status ;
 	int i;
+	struct ansi_string ansi_name;
+	struct unicode_string unicode_name;
 
-	status = IoCreateDevice(drv_obj, 0, NULL, FILE_DEVICE_UNKNOWN,
-			     0, FALSE, &pdo);
+	RtlInitAnsiString(&ansi_name, "NDISpdo");
+	if (RtlAnsiStringToUnicodeString(&unicode_name, &ansi_name, TRUE) ==
+	    STATUS_SUCCESS) {
+		status = IoCreateDevice(drv_obj, 0, &unicode_name,
+					FILE_DEVICE_UNKNOWN,
+					FILE_AUTOGENERATED_DEVICE_NAME,
+					FALSE, &pdo);
+		RtlFreeUnicodeString(&unicode_name);
+	} else {
+		status = IoCreateDevice(drv_obj, 0, NULL,
+					FILE_DEVICE_UNKNOWN,
+					FILE_AUTOGENERATED_DEVICE_NAME,
+					FALSE, &pdo);
+	}
 	DBGTRACE1("%p, %d, %p", drv_obj, status, pdo);
 	if (status != STATUS_SUCCESS)
 		return NULL;
+	/* dispatch routines are called as Windows functions */
 	for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++)
-		drv_obj->major_func[i] = IoInvalidDeviceRequest;
+		drv_obj->major_func[i] =
+			WIN_FUNC_PTR(IoInvalidDeviceRequest,2);
 	drv_obj->major_func[IRP_MJ_INTERNAL_DEVICE_CONTROL] =
-		pdoDispatchDeviceControl;
+		WIN_FUNC_PTR(pdoDispatchDeviceControl,2);
 	drv_obj->major_func[IRP_MJ_DEVICE_CONTROL] =
-		pdoDispatchDeviceControl;
-	drv_obj->major_func[IRP_MJ_POWER] = pdoDispatchPower;
-	drv_obj->major_func[IRP_MJ_PNP] = pdoDispatchPnp;
+		WIN_FUNC_PTR(pdoDispatchDeviceControl,2);
+	drv_obj->major_func[IRP_MJ_POWER] =
+		WIN_FUNC_PTR(pdoDispatchPower,2);
+	drv_obj->major_func[IRP_MJ_PNP] =
+		WIN_FUNC_PTR(pdoDispatchPnp,2);
 	return pdo;
 }
 
