@@ -84,6 +84,7 @@ static int inline wrap_cancel_urb(struct wrap_urb *wrap_urb)
 	if (wrap_urb->state != URB_SUBMITTED)
 		USBEXIT(return -1);
 	ret = usb_unlink_urb(wrap_urb->urb);
+	USBTRACE("ret: %d", ret);
 	if (ret == -EINPROGRESS)
 		return 0;
 	else {
@@ -127,8 +128,6 @@ static void usb_kill_urb(struct urb *urb)
 
 static struct nt_list wrap_urb_complete_list;
 static NT_SPIN_LOCK wrap_urb_complete_list_lock;
-static wstdcall void wrap_cancel_irp(struct device_object *dev_obj,
-				     struct irp *irp);
 
 /* use tasklet instead worker to process completed urbs */
 #define USB_TASKLET 1
@@ -196,6 +195,7 @@ static void kill_all_urbs(struct wrap_device *wd, int complete)
 				wrap_urb->urb->complete = NULL;
 			usb_kill_urb(wrap_urb->urb);
 		}
+		USBTRACE("%p, %p", wrap_urb, wrap_urb->urb);
 		usb_free_urb(wrap_urb->urb);
 		kfree(wrap_urb);
 	}
@@ -230,6 +230,7 @@ static USBD_STATUS wrap_urb_status(int urb_status)
 		return USBD_STATUS_ERROR_SHORT_TRANSFER;;
 	case -ENODEV:
 	case -ESHUTDOWN:
+	case -ENOENT:
 		return USBD_STATUS_DEVICE_GONE;
 	case -ENOMEM:
 		return USBD_STATUS_NO_MEMORY;
@@ -304,6 +305,24 @@ void wrap_resume_urbs(struct wrap_device *wd)
 	USBTRACE("%p, %d", wd, wd->usb.num_alloc_urbs);
 }
 
+wstdcall void wrap_cancel_irp(struct device_object *dev_obj, struct irp *irp)
+{
+	struct urb *urb;
+
+	/* NB: this function is called holding Cancel spinlock */
+	USBENTER("irp: %p", irp);
+	urb = irp->wrap_urb->urb;
+	USBTRACE("canceling urb %p", urb);
+	if (wrap_cancel_urb(irp->wrap_urb)) {
+		irp->cancel = FALSE;
+		ERROR("urb %p can't be canceld: %d", urb, irp->wrap_urb->state);
+	} else
+		USBTRACE("urb %p canceled", urb);
+	IoReleaseCancelSpinLock(irp->cancel_irql);
+	return;
+}
+WIN_FUNC_DECL(wrap_cancel_irp,2)
+
 static struct urb *wrap_alloc_urb(struct irp *irp, unsigned int pipe,
 				  void *buf, unsigned int buf_len)
 {
@@ -313,11 +332,8 @@ static struct urb *wrap_alloc_urb(struct irp *irp, unsigned int pipe,
 	struct wrap_device *wd;
 
 	USBENTER("irp: %p", irp);
-	if (current_irql() < DISPATCH_LEVEL)
-		alloc_flags = GFP_KERNEL;
-	else
-		alloc_flags = GFP_ATOMIC;
 	wd = irp->wd;
+	alloc_flags = gfp_irql();
 	IoAcquireCancelSpinLock(&irp->cancel_irql);
 	urb = NULL;
 	nt_list_for_each_entry(wrap_urb, &wd->usb.wrap_urb_list, list) {
@@ -361,7 +377,8 @@ static struct urb *wrap_alloc_urb(struct irp *irp, unsigned int pipe,
 	urb->context = wrap_urb;
 	wrap_urb->irp = irp;
 	irp->wrap_urb = wrap_urb;
-	irp->cancel_routine = wrap_cancel_irp;
+	/* called as Windows function */
+	irp->cancel_routine = WIN_FUNC_PTR(wrap_cancel_irp,2);
 	IoReleaseCancelSpinLock(irp->cancel_irql);
 	USBTRACE("allocated urb: %p", urb);
 
@@ -399,14 +416,8 @@ static USBD_STATUS wrap_submit_urb(struct irp *irp)
 {
 	int ret;
 	struct urb *urb;
-	unsigned int alloc_flags;
 	union nt_urb *nt_urb;
 
-	if (current_irql() < DISPATCH_LEVEL)
-		alloc_flags = GFP_KERNEL;
-	else
-		alloc_flags = GFP_ATOMIC;
-	IoAcquireCancelSpinLock(&irp->cancel_irql);
 	urb = irp->wrap_urb->urb;
 	nt_urb = URB_FROM_IRP(irp);
 #ifdef USB_DEBUG
@@ -414,13 +425,10 @@ static USBD_STATUS wrap_submit_urb(struct irp *irp)
 		ERROR("urb %p is in wrong state: %d",
 		      urb, irp->wrap_urb->state);
 		NT_URB_STATUS(nt_urb) = USBD_STATUS_REQUEST_FAILED;
-		IoReleaseCancelSpinLock(irp->cancel_irql);
 		return NT_URB_STATUS(nt_urb);
 	}
-	irp->wrap_urb->id = urb_id++;
+	irp->wrap_urb->id = pre_atomic_add(urb_id, 1);
 #endif
-	irp->wrap_urb->state = URB_SUBMITTED;
-	IoReleaseCancelSpinLock(irp->cancel_irql);
 	DUMP_WRAP_URB(irp->wrap_urb, USB_DIR_OUT);
 	irp->io_status.status = STATUS_PENDING;
 	irp->io_status.info = 0;
@@ -428,8 +436,9 @@ static USBD_STATUS wrap_submit_urb(struct irp *irp)
 	IoMarkIrpPending(irp);
 	DUMP_URB_BUFFER(urb, USB_DIR_OUT);
 	USBTRACE("%p", urb);
+	irp->wrap_urb->state = URB_SUBMITTED;
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
-	ret = usb_submit_urb(urb, alloc_flags);
+	ret = usb_submit_urb(urb, gfp_irql());
 #else
 	ret = usb_submit_urb(urb);
 #endif
@@ -453,6 +462,7 @@ static void int_urb_unlink_complete(struct urb *urb)
 	USBTRACE("%p, %d, %d", urb, urb->status, wrap_urb->state);
 	if (xchg(&wrap_urb->state, URB_INT_UNLINKED) != URB_COMPLETED)
 		return;
+	urb->status = URB_STATUS(wrap_urb);
 	nt_spin_lock(&wrap_urb_complete_list_lock);
 	InsertTailList(&wrap_urb_complete_list, &wrap_urb->complete_list);
 	nt_spin_unlock(&wrap_urb_complete_list_lock);
@@ -541,8 +551,8 @@ static void wrap_urb_complete_worker(void *dummy)
 		DUMP_IRP(irp);
 		nt_urb = URB_FROM_IRP(irp);
 		USBTRACE("urb: %p, nt_urb: %p, status: %d",
-			 urb, nt_urb, URB_STATUS(wrap_urb));
-		switch (URB_STATUS(wrap_urb)) {
+			 urb, nt_urb, urb->status);
+		switch (urb->status) {
 		case 0:
 			/* succesfully transferred */
 			irp->io_status.info = urb->actual_length;
@@ -557,7 +567,7 @@ static void wrap_urb_complete_worker(void *dummy)
 					memcpy(bulk_int_tx->transfer_buffer,
 					       urb->transfer_buffer,
 					       urb->actual_length);
-			} else {
+			} else { // vendor or class request
 				vc_req = &nt_urb->vendor_class_request;
 				vc_req->transfer_buffer_length =
 					urb->actual_length;
@@ -568,25 +578,22 @@ static void wrap_urb_complete_worker(void *dummy)
 					       urb->transfer_buffer,
 					       urb->actual_length);
 			}
-			NT_URB_STATUS(nt_urb) =
-				wrap_urb_status(URB_STATUS(wrap_urb));
-			irp->io_status.status =
-				nt_urb_irp_status(NT_URB_STATUS(nt_urb));
+			NT_URB_STATUS(nt_urb) = USBD_STATUS_SUCCESS;
+			irp->io_status.status = STATUS_SUCCESS;
 			break;
 		case -ENOENT:
 		case -ECONNRESET:
 			/* urb canceled */
 			irp->io_status.info = 0;
-			USBTRACE("urb %p canceled", urb);
+			DBGTRACE1("urb %p canceled", urb);
 			NT_URB_STATUS(nt_urb) = USBD_STATUS_SUCCESS;
 			irp->io_status.status = STATUS_CANCELLED;
 			break;
 		default:
-			USBTRACE("irp: %p, urb: %p, status: %d",
-				 irp, urb, URB_STATUS(wrap_urb));
+			DBGTRACE1("irp: %p, urb: %p, status: %d/%d",
+				  irp, urb, urb->status, wrap_urb->state);
 			irp->io_status.info = 0;
-			NT_URB_STATUS(nt_urb) =
-				wrap_urb_status(URB_STATUS(wrap_urb));
+			NT_URB_STATUS(nt_urb) = wrap_urb_status(urb->status);
 			irp->io_status.status =
 				nt_urb_irp_status(NT_URB_STATUS(nt_urb));
 			break;
@@ -597,28 +604,6 @@ static void wrap_urb_complete_worker(void *dummy)
 	USBEXIT(return);
 }
 
-/* called as Windows function, so call WIN2LIN2 before accessing
- * arguments */
-static wstdcall void wrap_cancel_irp(struct device_object *dev_obj,
-				     struct irp *irp)
-{
-	struct urb *urb;
-
-	WIN2LIN2(dev_obj, irp);
-
-	/* NB: this function is called holding Cancel spinlock */
-	USBENTER("irp: %p", irp);
-	urb = irp->wrap_urb->urb;
-	USBTRACE("canceling urb %p", urb);
-	IoReleaseCancelSpinLock(irp->cancel_irql);
-	if (wrap_cancel_urb(irp->wrap_urb)) {
-		irp->cancel = FALSE;
-		ERROR("urb %p can't be canceld: %d", urb, irp->wrap_urb->state);
-	} else
-		USBTRACE("urb %p canceled", urb);
-	return;
-}
-
 static USBD_STATUS wrap_bulk_or_intr_trans(struct irp *irp)
 {
 	usbd_pipe_handle pipe_handle;
@@ -644,14 +629,14 @@ static USBD_STATUS wrap_bulk_or_intr_trans(struct irp *irp)
 					       pipe_handle->bEndpointAddress);
 		else
 			pipe = usb_sndbulkpipe(udev,
-					      pipe_handle->bEndpointAddress);
+					       pipe_handle->bEndpointAddress);
 	} else {
-#ifdef USB_DEBUG
-		if (!(bulk_int_tx->transfer_flags &
-		      USBD_TRANSFER_DIRECTION_IN))
-			ERROR("invalid urb: %x", bulk_int_tx->transfer_flags);
-#endif
-		pipe = usb_rcvintpipe(udev, pipe_handle->bEndpointAddress);
+		if (bulk_int_tx->transfer_flags & USBD_TRANSFER_DIRECTION_IN)
+			pipe = usb_rcvintpipe(udev,
+					      pipe_handle->bEndpointAddress);
+		else
+			pipe = usb_sndintpipe(udev,
+					      pipe_handle->bEndpointAddress);
 	}
 
 	DUMP_IRP(irp);
@@ -679,7 +664,7 @@ static USBD_STATUS wrap_bulk_or_intr_trans(struct irp *irp)
 				 pipe_handle->bInterval);
 		USBTRACE("submitting interrupt urb %p on pipe 0x%x (ep 0x%x), "
 			 "intvl: %d", urb, urb->pipe,
-			 pipe_handle->bEndpointAddress, pipe_handle->bInterval);
+			 pipe_handle->bEndpointAddress, interval);
 	}
 	status = wrap_submit_urb(irp);
 	USBTRACE("status: %08X", status);
@@ -688,7 +673,7 @@ static USBD_STATUS wrap_bulk_or_intr_trans(struct irp *irp)
 
 static USBD_STATUS wrap_vendor_or_class_req(struct irp *irp)
 {
-	__u8 req_type;
+	u8 req_type;
 	unsigned int pipe;
 	struct usbd_vendor_or_class_request *vc_req;
 	struct usb_device *udev;
@@ -790,44 +775,32 @@ static USBD_STATUS wrap_reset_pipe(struct usb_device *udev, struct irp *irp)
 	int ret;
 	union nt_urb *nt_urb;
 	usbd_pipe_handle pipe_handle;
-	unsigned int ep;
 	unsigned int pipe1, pipe2;
-	enum pipe_type pipe_type;
 
 	nt_urb = URB_FROM_IRP(irp);
 	pipe_handle = nt_urb->pipe_req.pipe_handle;
-	ep = pipe_handle->bEndpointAddress;
-	USBTRACE("irp: %p, ep: 0x%x", irp, ep);
 	/* TODO: not clear if both directions should be cleared? */
-	pipe_type = pipe_handle->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK;
-	if (pipe_type == USB_ENDPOINT_XFER_BULK) {
+	if (USBD_IS_BULK_PIPE(pipe_handle)) {
 		pipe1 = usb_rcvbulkpipe(udev, pipe_handle->bEndpointAddress);
 		pipe2 = usb_sndbulkpipe(udev, pipe_handle->bEndpointAddress);
-	} else if (pipe_type == USB_ENDPOINT_XFER_INT) {
+	} else if (USBD_IS_INT_PIPE(pipe_handle)) {
 		pipe1 = usb_rcvintpipe(udev, pipe_handle->bEndpointAddress);
-		/* no outgoing interrupt pipes */
 		pipe2 = pipe1;
-#if 0
-	} else if (pipe_type == USB_ENDPOINT_XFER_CONTROL) {
-		pipe1 = usb_rcvctrlpipe(udev, pipe_handle->bEndpointAddress);
-		pipe2 = usb_sndctrlpipe(udev, pipe_handle->bEndpointAddress);
-#endif
 	} else {
-		WARNING("invalid pipe type %d", pipe_type);
+		WARNING("invalid pipe %d", pipe_handle->bEndpointAddress);
 		return USBD_STATUS_INVALID_PIPE_HANDLE;
 	}
+	USBTRACE("ep: %d, pipe: 0x%x", pipe_handle->bEndpointAddress, pipe1);
 	ret = usb_clear_halt(udev, pipe1);
-	if (ret) {
-		USBTRACE("resetting pipe %d failed: %d", pipe_type, ret);
-		return wrap_urb_status(ret);
-	}
+	if (ret)
+		USBTRACE("resetting pipe %d failed: %d", pipe1, ret);
 	if (pipe2 != pipe1) {
 		ret = usb_clear_halt(udev, pipe2);
 		if (ret)
-			USBTRACE("resetting pipe %d failed: %d",
-				 pipe_type, ret);
+			USBTRACE("resetting pipe %d failed: %d", pipe2, ret);
 	}
-	return wrap_urb_status(ret);
+//	return wrap_urb_status(ret);
+	return USBD_STATUS_SUCCESS;
 }
 
 static USBD_STATUS wrap_abort_pipe(struct usb_device *udev, struct irp *irp)
@@ -865,17 +838,15 @@ static void set_intf_pipe_info(struct wrap_device *wd,
 			       struct usb_interface *usb_intf,
 			       struct usbd_interface_information *intf)
 {
-	int i, pipe_num;
+	int i;
 	struct usb_endpoint_descriptor *ep;
 	struct usbd_pipe_information *pipe;
 
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
-	for (i = 0; i < CUR_ALT_SETTING(usb_intf)->desc.bNumEndpoints;
-	     i++, pipe_num++) {
+	for (i = 0; i < CUR_ALT_SETTING(usb_intf)->desc.bNumEndpoints; i++) {
 		ep = &(CUR_ALT_SETTING(usb_intf)->endpoint[i]).desc;
 #else
-	for (i = 0; i < CUR_ALT_SETTING(usb_intf).bNumEndpoints;
-	     i++, pipe_num++) {
+	for (i = 0; i < CUR_ALT_SETTING(usb_intf).bNumEndpoints; i++) {
 		ep = &((CUR_ALT_SETTING(usb_intf)).endpoint[i]);
 #endif
 		if (i >= intf->bNumEndpoints) {
@@ -894,16 +865,16 @@ static void set_intf_pipe_info(struct wrap_device *wd,
 
 		pipe->wMaxPacketSize = ep->wMaxPacketSize;
 		pipe->bEndpointAddress = ep->bEndpointAddress;
-		if (pipe->type == USB_ENDPOINT_XFER_INT) {
-			if (wd->usb.udev->speed == USB_SPEED_HIGH) {
-				u8 i, interval = ep->bInterval;
-				for (i = 0; interval; i++)
-					interval /= 2;
-				pipe->bInterval = i;
-			} else
+		pipe->type = ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK;
+		if (pipe->type == UsbdPipeTypeInterrupt) {
+			/* for low speed devices, Linux sets bInterval
+			 * as frames per second, whereas Windows
+			 * interprets it as milliseconds */
+			if (wd->usb.udev->speed == USB_SPEED_LOW)
+				pipe->bInterval = 1 << (ep->bInterval - 1);
+			else
 				pipe->bInterval = ep->bInterval;
 		}
-		pipe->type = ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK;
 		pipe->handle = ep;
 		USBTRACE("%d: ep 0x%x, type %d, pkt_sz %d, intv %d (%d),"
 			 "type: %d, handle %p", i, ep->bEndpointAddress,
@@ -916,7 +887,7 @@ static USBD_STATUS wrap_select_configuration(struct wrap_device *wd,
 					     union nt_urb *nt_urb,
 					     struct irp *irp)
 {
-	int n, ret;
+	int i, ret;
 	struct usbd_select_configuration *sel_conf;
 	struct usb_device *udev;
 	struct usbd_interface_information *intf;
@@ -926,34 +897,21 @@ static USBD_STATUS wrap_select_configuration(struct wrap_device *wd,
 	udev = wd->usb.udev;
 	sel_conf = &nt_urb->select_conf;
 	config = sel_conf->config;
+	USBTRACE("%p", config);
 	if (config == NULL) {
-		/* TODO: set to unconfigured state (configuration 0):
-		 * is this correctt? */
-#if 0
-		ret = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
-				      USB_REQ_SET_CONFIGURATION, 0,
-				      0, 0, NULL, 0, USB_CTRL_SET_TIMEOUT);
-		ret = usb_reset_configuration(udev);
-		usb_disable_device(udev, 0);
-#endif
 		kill_all_urbs(wd, 1);
-		ret = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
-				      USB_REQ_SET_CONFIGURATION, 0,
-				      0, 0, NULL, 0, USB_CTRL_SET_TIMEOUT);
-		USBTRACE("ret: %d", ret);
-#if 0
-		if (ret < 0)
-			return wrap_urb_status(ret);
-		else
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
+		ret = usb_reset_configuration(udev);
+#else
+		ret = 0;
 #endif
-			return USBD_STATUS_SUCCESS;
+		return wrap_urb_status(ret);
 	}
 
 	USBTRACE("conf: %d, type: %d, length: %d, numif: %d, attr: %08x",
 		 config->bConfigurationValue, config->bDescriptorType,
 		 config->wTotalLength, config->bNumInterfaces,
 		 config->bmAttributes);
-
 	ret = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
 			      USB_REQ_SET_CONFIGURATION, 0,
 			      config->bConfigurationValue, 0,
@@ -962,28 +920,25 @@ static USBD_STATUS wrap_select_configuration(struct wrap_device *wd,
 		ERROR("ret: %d", ret);
 		return wrap_urb_status(ret);
 	}
-
+	sel_conf->handle = udev->actconfig;
 	intf = &sel_conf->intf;
-	for (n = 0; n < config->bNumInterfaces && intf->bLength > 0;
-	     n++, intf = (((void *)intf) + intf->bLength)) {
+	for (i = 0; i < config->bNumInterfaces && intf->bLength > 0;
+	     i++, intf = (((void *)intf) + intf->bLength)) {
 
 		USBTRACE("intf: %d, alt setting: %d",
 			 intf->bInterfaceNumber, intf->bAlternateSetting);
-#if 0
 		ret = usb_set_interface(udev, intf->bInterfaceNumber,
 					intf->bAlternateSetting);
 		if (ret < 0) {
 			ERROR("failed with %d", ret);
 			return wrap_urb_status(ret);
 		}
-#endif
 		usb_intf = usb_ifnum_to_if(udev, intf->bInterfaceNumber);
 		if (!usb_intf) {
 			ERROR("couldn't obtain ifnum");
 			return USBD_STATUS_REQUEST_FAILED;
 		}
 		USBTRACE("intf: %p, num ep: %d", intf, intf->bNumEndpoints);
-
 		set_intf_pipe_info(wd, usb_intf, intf);
 	}
 	return USBD_STATUS_SUCCESS;
@@ -1014,9 +969,7 @@ static USBD_STATUS wrap_select_interface(struct wrap_device *wd,
 		ERROR("couldn't get interface information");
 		return USBD_STATUS_REQUEST_FAILED;
 	}
-	wd->usb.intf = usb_intf;
-	USBTRACE("intf: %p, num ep: %d", intf, intf->bNumEndpoints);
-
+	USBTRACE("intf: %p, num ep: %d", usb_intf, intf->bNumEndpoints);
 	set_intf_pipe_info(wd, usb_intf, intf);
 	return USBD_STATUS_SUCCESS;
 }
@@ -1041,36 +994,36 @@ static int wrap_usb_get_string(struct usb_device *udev, unsigned short langid,
 static USBD_STATUS wrap_get_descriptor(struct wrap_device *wd,
 				       union nt_urb *nt_urb, struct irp *irp)
 {
-	struct usbd_control_descriptor_request *ctrl_req;
+	struct usbd_control_descriptor_request *control_desc;
 	int ret = 0;
 	struct usb_device *udev;
 
 	udev = wd->usb.udev;
-	ctrl_req = &nt_urb->control_request;
+	control_desc = &nt_urb->control_desc;
 	USBTRACE("desctype = %d, descindex = %d, transfer_buffer = %p,"
-		 "transfer_buffer_length = %d", ctrl_req->desc_type,
-		 ctrl_req->index, ctrl_req->transfer_buffer,
-		 ctrl_req->transfer_buffer_length);
+		 "transfer_buffer_length = %d", control_desc->desc_type,
+		 control_desc->index, control_desc->transfer_buffer,
+		 control_desc->transfer_buffer_length);
 
-	if (ctrl_req->desc_type == USB_DT_STRING) {
-		USBTRACE("langid: %x", ctrl_req->language_id);
-		ret = wrap_usb_get_string(udev, ctrl_req->language_id,
-					  ctrl_req->index,
-					  ctrl_req->transfer_buffer,
-					  ctrl_req->transfer_buffer_length);
+	if (control_desc->desc_type == USB_DT_STRING) {
+		USBTRACE("langid: %x", control_desc->language_id);
+		ret = wrap_usb_get_string(udev, control_desc->language_id,
+					  control_desc->index,
+					  control_desc->transfer_buffer,
+					  control_desc->transfer_buffer_length);
 	} else {
-		ret = usb_get_descriptor(udev, ctrl_req->desc_type,
-					 ctrl_req->index,
-					 ctrl_req->transfer_buffer,
-					 ctrl_req->transfer_buffer_length);
+		ret = usb_get_descriptor(udev, control_desc->desc_type,
+					 control_desc->index,
+					 control_desc->transfer_buffer,
+					 control_desc->transfer_buffer_length);
 	}
 	if (ret < 0) {
-		USBTRACE("request %d failed: %d", ctrl_req->desc_type, ret);
-		ctrl_req->transfer_buffer_length = 0;
+		USBTRACE("request %d failed: %d", control_desc->desc_type, ret);
+		control_desc->transfer_buffer_length = 0;
 		return wrap_urb_status(ret);
 	} else {
 		USBTRACE("ret: %08x", ret);
-		ctrl_req->transfer_buffer_length = ret;
+		control_desc->transfer_buffer_length = ret;
 		irp->io_status.info = ret;
 		return USBD_STATUS_SUCCESS;
 	}
@@ -1207,8 +1160,8 @@ NTSTATUS wrap_submit_irp(struct device_object *pdo, struct irp *irp)
 
 	USBTRACE("%p, %p", pdo, irp);
 	wd = pdo->reserved;
-	if (wd->usb.intf == NULL)
-		USBEXIT(return STATUS_DEVICE_REMOVED);
+//	if (wd->usb.intf == NULL)
+//		USBEXIT(return STATUS_DEVICE_REMOVED);
 	irp->wd = wd;
 	irp_sl = IoGetCurrentIrpStackLocation(irp);
 	switch (irp_sl->params.dev_ioctl.code) {
@@ -1251,7 +1204,7 @@ NTSTATUS wrap_submit_irp(struct device_object *pdo, struct irp *irp)
  * description and examples elsewhere suggest that it should be
  * usbd_interface_list_entry structre. Which is correct? */
 
-wstdcall union nt_urb *WRAP_EXPORT(USBD_CreateConfigurationRequestEx)
+wstdcall union nt_urb *WIN_FUNC(USBD_CreateConfigurationRequestEx,2)
 	(struct usb_config_descriptor *config,
 	 struct usbd_interface_list_entry *intf_list)
 {
@@ -1280,8 +1233,7 @@ wstdcall union nt_urb *WRAP_EXPORT(USBD_CreateConfigurationRequestEx)
 	}
 	memset(select_conf, 0, size);
 	intf = &select_conf->intf;
-	/* handle points to beginning of interface information */
-	select_conf->handle = intf;
+	select_conf->handle = config;
 	for (n = 0; n < config->bNumInterfaces && intf_list[n].intf_desc; n++) {
 		/* initialize 'intf' fields in intf_list so they point
 		 * to appropriate entry; these may be read/written by
@@ -1305,6 +1257,7 @@ wstdcall union nt_urb *WRAP_EXPORT(USBD_CreateConfigurationRequestEx)
 			pipe[i].max_tx_size =
 				USBD_DEFAULT_MAXIMUM_TRANSFER_SIZE;
 		}
+		intf->handle = intf_desc;
 		intf = (((void *)intf) + intf->bLength);
 	}
 	select_conf->header.function = URB_FUNCTION_SELECT_CONFIGURATION;
@@ -1313,10 +1266,10 @@ wstdcall union nt_urb *WRAP_EXPORT(USBD_CreateConfigurationRequestEx)
 	USBEXIT(return (union nt_urb *)select_conf);
 }
 
-WRAP_EXPORT_MAP("_USBD_CreateConfigurationRequestEx@8",	USBD_CreateConfigurationRequestEx);
+WIN_SYMBOL_MAP("_USBD_CreateConfigurationRequestEx@8", USBD_CreateConfigurationRequestEx)
 
 wstdcall struct usb_interface_descriptor *
-WRAP_EXPORT(USBD_ParseConfigurationDescriptorEx)
+WIN_FUNC(USBD_ParseConfigurationDescriptorEx,7)
 	(struct usb_config_descriptor *config, void *start,
 	 LONG bInterfaceNumber, LONG bAlternateSetting, LONG bInterfaceClass,
 	 LONG bInterfaceSubClass, LONG bInterfaceProtocol)
@@ -1352,9 +1305,9 @@ WRAP_EXPORT(USBD_ParseConfigurationDescriptorEx)
 	USBEXIT(return NULL);
 }
 
-WRAP_EXPORT_MAP("_USBD_ParseConfigurationDescriptorEx@28", USBD_ParseConfigurationDescriptorEx);
+WIN_SYMBOL_MAP("_USBD_ParseConfigurationDescriptorEx@28", USBD_ParseConfigurationDescriptorEx)
 
-wstdcall union nt_urb *WRAP_EXPORT(USBD_CreateConfigurationRequest)
+wstdcall union nt_urb *WIN_FUNC(USBD_CreateConfigurationRequest,2)
 	(struct usb_config_descriptor *config, USHORT *size)
 {
 	union nt_urb *nt_urb;
@@ -1378,7 +1331,7 @@ wstdcall union nt_urb *WRAP_EXPORT(USBD_CreateConfigurationRequest)
 }
 
 wstdcall struct usb_interface_descriptor *
-WRAP_EXPORT(USBD_ParseConfigurationDescriptor)
+WIN_FUNC(USBD_ParseConfigurationDescriptor,3)
 	(struct usb_config_descriptor *config, UCHAR bInterfaceNumber,
 	 UCHAR bAlternateSetting)
 {
@@ -1388,7 +1341,7 @@ WRAP_EXPORT(USBD_ParseConfigurationDescriptor)
 						   -1, -1, -1);
 }
 
-wstdcall void WRAP_EXPORT(USBD_GetUSBDIVersion)
+wstdcall void WIN_FUNC(USBD_GetUSBDIVersion,1)
 	(struct usbd_version_info *version_info)
 {
 	/* this function is obsolete in Windows XP */
@@ -1432,13 +1385,13 @@ wstdcall BOOLEAN USBD_InterfaceIsDeviceHighSpeed(void *context)
 wstdcall void USBD_InterfaceReference(void *context)
 {
 	USBTRACE("%p", context);
-	UNIMPL();
+	TODO();
 }
 
 wstdcall void USBD_InterfaceDereference(void *context)
 {
 	USBTRACE("%p", context);
-	UNIMPL();
+	TODO();
 }
 
 wstdcall NTSTATUS USBD_InterfaceQueryBusTime(void *context, ULONG *frame)
@@ -1453,7 +1406,7 @@ wstdcall NTSTATUS USBD_InterfaceSubmitIsoOutUrb(void *context,
 					       union nt_urb *nt_urb)
 {
 	/* TODO: implement this */
-	UNIMPL();
+	TODO();
 	USBEXIT(return STATUS_NOT_IMPLEMENTED);
 }
 
@@ -1467,7 +1420,7 @@ USBD_InterfaceQueryBusInformation(void *context, ULONG level, void *buf,
 
 	bus = wd->usb.udev->bus;
 	bus_info = buf;
-	UNIMPL();
+	TODO();
 	USBEXIT(return STATUS_NOT_IMPLEMENTED);
 }
 
