@@ -35,6 +35,7 @@
 #define STATUS_MORE_PROCESSING_REQUIRED 0xC0000016
 #define STATUS_ACCESS_DENIED            0xC0000022
 #define STATUS_BUFFER_TOO_SMALL         0xC0000023
+#define STATUS_OBJECT_NAME_INVALID      0xC0000023
 #define STATUS_MUTANT_NOT_OWNED		0xC0000046
 #define STATUS_RESOURCES                0xC000009A
 #define STATUS_DELETE_PENDING		0xC0000056
@@ -112,11 +113,13 @@
 #ifdef CONFIG_X86_64
 #define wstdcall
 #define wfastcall
+#define noregparm
 
 #define KI_USER_SHARED_DATA 0xfffff78000000000
 
 #else
 
+#define noregparm __attribute__((regparm(0)))
 #define wstdcall __attribute__((__stdcall__, regparm(0)))
 #if defined(__GNUC__) && ((__GNUC__ == 3 && __GNUC_MINOR__ > 3) || __GNUC__ > 3)
 #undef fastcall
@@ -129,7 +132,6 @@
 
 #endif
 
-#define noregparm __attribute__((regparm(0)))
 #define packed __attribute__((packed))
 #define no_warn_unused __attribute((unused))
 
@@ -194,9 +196,18 @@ struct nt_slist {
 };
 
 #ifdef CONFIG_X86_64
+/* it is not clear how nt_slist_head is used to store pointer to
+ * slists and depth; here we assume 'align' field is used to store
+ * depth and 'region' field is used to store slist pointers */
 struct nt_slist_head {
-	ULONGLONG align;
-	ULONGLONG region;
+	union {
+		USHORT depth;
+		ULONGLONG align;
+	};
+	union {
+		ULONGLONG region;
+		struct nt_slist *next;
+	};
 } __attribute__((aligned(16)));
 typedef struct nt_slist_head nt_slist_header;
 #else
@@ -219,18 +230,13 @@ struct nt_list {
 typedef ULONG_PTR NT_SPIN_LOCK;
 
 struct kdpc;
-typedef wstdcall void (*DPC)(struct kdpc *kdpc, void *ctx, void *arg1,
-			    void *arg2);
-
+typedef void (*DPC)(struct kdpc *kdpc, void *ctx, void *arg1,
+		    void *arg2) wstdcall;
 struct kdpc {
 	SHORT type;
-	/* we don't implement kdpc executing on a specific processor;
-	 * until then number is used to represent if the dpc is queued
-	 * or not */
 	UCHAR nr_cpu;
 	UCHAR importance;
 	struct nt_list list;
-
 	DPC func;
 	void *ctx;
 	void *arg1;
@@ -294,6 +300,8 @@ struct mdl {
 #define MDL_NETWORK_HEADER		0x1000
 #define MDL_MAPPING_CAN_FAIL		0x2000
 #define MDL_ALLOCATED_MUST_SUCCEED	0x4000
+
+#define MDL_POOL_ALLOCATED		0x0400
 #define MDL_CACHE_ALLOCATED		0x8000
 
 #define page_start(ptr) ((void *)((ULONG_PTR)(ptr) & ~(PAGE_SIZE - 1)))
@@ -346,7 +354,7 @@ struct wait_block {
 	struct nt_list list;
 	struct task_struct *thread;
 	void *object;
-	void *thread_wait;
+	void *thread_waitq;
 	USHORT wait_key;
 	USHORT wait_type;
 };
@@ -510,8 +518,8 @@ struct io_status_block32 {
 
 struct driver_extension;
 
-typedef NTSTATUS (driver_dispatch_t)(struct device_object *dev_obj,
-				     struct irp *irp) wstdcall;
+typedef NTSTATUS driver_dispatch_t(struct device_object *dev_obj,
+				   struct irp *irp) wstdcall;
 
 struct driver_object {
 	CSHORT type;
@@ -545,6 +553,8 @@ struct custom_ext {
 	void *client_id;
 };
 
+struct wrap_bin_file;
+
 struct file_object {
 	CSHORT type;
 	CSHORT size;
@@ -555,7 +565,10 @@ struct file_object {
 	void *section_object_pointer;
 	void *private_cache_map;
 	NTSTATUS final_status;
-	struct file_object *related_file_object;
+	union {
+		struct file_object *related_file_object;
+		struct wrap_bin_file *wrap_bin_file;
+	};
 	BOOLEAN lock_operation;
 	BOOLEAN delete_pending;
 	BOOLEAN read_access;
@@ -565,7 +578,7 @@ struct file_object {
 	BOOLEAN shared_write;
 	BOOLEAN shared_delete;
 	ULONG flags;
-	struct unicode_string file_name;
+	struct unicode_string _name_;
 	LARGE_INTEGER current_byte_offset;
 	ULONG waiters;
 	ULONG busy;
@@ -786,6 +799,16 @@ enum device_usage_notification_type {
 	DeviceUsageTypeHibernation, DevbiceUsageTypeDumpFile,
 };
 
+#define METHOD_BUFFERED		0
+#define METHOD_IN_DIRECT	1
+#define METHOD_OUT_DIRECT	2
+#define METHOD_NEITHER		3
+
+#define CTL_CODE(dev_type, func, method, access)	\
+	(((dev_type) << 16) | ((access) << 14) | ((func) << 2) | (method))
+
+#define IO_METHOD_FROM_CTL_CODE(code) (code & 0x3)
+
 #ifndef CONFIG_X86_64
 #pragma pack(push,4)
 #endif
@@ -1130,10 +1153,7 @@ struct object_attr {
 	void *root_dir;
 	struct unicode_string *name;
 	ULONG attr;
-	union {
-		void *security_descriptor;
-		void *file;
-	};
+	void *security_descriptor;
 	void *security_qos;
 };
 
@@ -1164,10 +1184,10 @@ struct common_object_header {
 	struct nt_list list;
 	enum common_object_type type;
 	UINT size;
-	char *name;
 	UINT ref_count;
 	BOOLEAN close_in_process;
 	BOOLEAN permanent;
+	struct unicode_string name;
 };
 
 #define OBJECT_TO_HEADER(object)					\
@@ -1185,7 +1205,7 @@ enum work_queue_type {
 	MaximumWorkQueue
 };
 
-typedef wstdcall void (*NTOS_WORK_FUNC)(void *arg1, void *arg2);
+typedef void (*NTOS_WORK_FUNC)(void *arg1, void *arg2) wstdcall;
 
 struct io_workitem {
 	enum work_queue_type type;
@@ -1218,9 +1238,9 @@ enum ntos_wait_reason {
 
 typedef enum ntos_wait_reason KWAIT_REASON;
 
-typedef wstdcall void *LOOKASIDE_ALLOC_FUNC(enum pool_type pool_type,
-					   SIZE_T size, ULONG tag);
-typedef wstdcall void LOOKASIDE_FREE_FUNC(void *);
+typedef void *LOOKASIDE_ALLOC_FUNC(enum pool_type pool_type,
+				   SIZE_T size, ULONG tag) wstdcall;
+typedef void LOOKASIDE_FREE_FUNC(void *) wstdcall;
 
 struct npaged_lookaside_list {
 	nt_slist_header head;
@@ -1318,7 +1338,8 @@ struct object_attributes {
 	void *security_qos;
 };
 
-typedef void (*PCALLBACK_FUNCTION)(void *context, void *arg1, void *arg2);
+typedef void (*PCALLBACK_FUNCTION)(void *context, void *arg1,
+				   void *arg2) wstdcall;
 
 struct callback_object;
 struct callback_func {
@@ -1437,6 +1458,22 @@ struct io_remove_lock {
 	struct nt_event remove_event;
 };
 
+struct io_error_log_packet {
+	UCHAR major_fn_code;
+	UCHAR retry_count;
+	USHORT dump_data_size;
+	USHORT nr_of_strings;
+	USHORT string_offset;
+	USHORT event_category;
+	NTSTATUS error_code;
+	ULONG unique_error_value;
+	NTSTATUS final_status;
+	ULONG sequence_number;
+	ULONG io_control_code;
+	LARGE_INTEGER device_offset;
+	ULONG dump_data[1];
+};
+
 /* some of the functions below are slightly different from DDK's
  * implementation; e.g., Insert functions return appropriate
  * pointer */
