@@ -22,6 +22,7 @@
 extern char *if_name;
 extern int hangcheck_interval;
 extern struct iw_handler_def ndis_handler_def;
+extern NT_SPIN_LOCK timer_lock;
 
 static int set_packet_filter(struct wrap_ndis_device *wnd,
 			     ULONG packet_filter);
@@ -57,10 +58,11 @@ NDIS_STATUS miniport_reset(struct wrap_ndis_device *wnd)
 	miniport = &wnd->wd->driver->ndis_driver->miniport;
 	cur_lookahead = wnd->nmb->cur_lookahead;
 	max_lookahead = wnd->nmb->max_lookahead;
-	irql = raise_irql(DISPATCH_LEVEL);
 	wnd->ndis_comm_done = 0;
+	WARNING("%s is being reset", wnd->net_dev->name);
+	irql = serialize_lock_irql(wnd);
 	res = LIN2WIN2(miniport->reset, &reset_address, wnd->nmb->adapter_ctx);
-	lower_irql(irql);
+	serialize_unlock_irql(wnd, irql);
 
 	DBGTRACE2("%08X, %08X", res, reset_address);
 	if (res == NDIS_STATUS_PENDING) {
@@ -100,11 +102,11 @@ NDIS_STATUS miniport_query_info_needed(struct wrap_ndis_device *wnd,
 		TRACEEXIT3(return NDIS_STATUS_FAILURE);
 	miniport = &wnd->wd->driver->ndis_driver->miniport;
 	DBGTRACE2("%p, %08X", miniport->query, oid);
-	irql = raise_irql(DISPATCH_LEVEL);
 	wnd->ndis_comm_done = 0;
+	irql = serialize_lock_irql(wnd);
 	res = LIN2WIN6(miniport->query, wnd->nmb->adapter_ctx, oid, buf,
 		       bufsize, &written, needed);
-	lower_irql(irql);
+	serialize_unlock_irql(wnd, irql);
 
 	DBGTRACE2("%08X, %08X", res, oid);
 	if (res == NDIS_STATUS_PENDING) {
@@ -149,11 +151,11 @@ NDIS_STATUS miniport_set_info(struct wrap_ndis_device *wnd, ndis_oid oid,
 		TRACEEXIT3(return NDIS_STATUS_FAILURE);
 	miniport = &wnd->wd->driver->ndis_driver->miniport;
 	DBGTRACE2("%p, %08X", miniport->query, oid);
-	irql = raise_irql(DISPATCH_LEVEL);
 	wnd->ndis_comm_done = 0;
+	irql = serialize_lock_irql(wnd);
 	res = LIN2WIN6(miniport->setinfo, wnd->nmb->adapter_ctx, oid,
 		       buf, bufsize, &written, &needed);
-	lower_irql(irql);
+	serialize_unlock_irql(wnd, irql);
 
 	DBGTRACE2("%08X, %08X", res, oid);
 	if (res == NDIS_STATUS_PENDING) {
@@ -190,20 +192,26 @@ static NDIS_STATUS miniport_pnp_event(struct wrap_ndis_device *wnd,
 				      enum ndis_device_pnp_event event)
 {
 	struct miniport_char *miniport;
-	enum ndis_power_profile power_profile;
+	ULONG power_profile;
 
 	TRACEENTER1("%p, %d", wnd, event);
+	miniport = &wnd->wd->driver->ndis_driver->miniport;
+	if (!miniport->pnp_event_notify) {
+		DBGTRACE1("Windows driver %s doesn't support "
+			  "MiniportPnpEventNotify", wnd->wd->driver->name);
+		return NDIS_STATUS_FAILURE;
+	}
 	/* RNDIS driver doesn't like to be notified if device is
 	 * already halted */
 	if (!test_bit(HW_INITIALIZED, &wnd->hw_status))
 		TRACEEXIT1(return NDIS_STATUS_SUCCESS);
-	miniport = &wnd->wd->driver->ndis_driver->miniport;
 	switch (event) {
 	case NdisDevicePnPEventSurpriseRemoved:
 		DBGTRACE1("%u, %p",
 			  (wnd->attributes & NDIS_ATTRIBUTE_SURPRISE_REMOVE_OK),
 			  miniport->pnp_event_notify);
 		if ((wnd->attributes & NDIS_ATTRIBUTE_SURPRISE_REMOVE_OK) &&
+		    wnd->wd->surprise_removed == TRUE &&
 		    miniport->pnp_event_notify) {
 			DBGTRACE1("calling surprise_removed");
 			LIN2WIN4(miniport->pnp_event_notify,
@@ -215,12 +223,6 @@ static NDIS_STATUS miniport_pnp_event(struct wrap_ndis_device *wnd,
 				  wnd->wd->driver->name);
 		return NDIS_STATUS_SUCCESS;
 	case NdisDevicePnPEventPowerProfileChanged:
-		if (!miniport->pnp_event_notify) {
-			DBGTRACE1("Windows driver %s doesn't support "
-				  "MiniportPnpEventNotify",
-				  wnd->wd->driver->name);
-			return NDIS_STATUS_FAILURE;
-		}
 		power_profile = NdisPowerProfileAcOnLine;
 		LIN2WIN4(miniport->pnp_event_notify, wnd->nmb->adapter_ctx,
 			 NdisDevicePnPEventPowerProfileChanged,
@@ -267,9 +269,9 @@ static NDIS_STATUS miniport_init(struct wrap_ndis_device *wnd)
 	 * fail after boot */
 	sleep_hz(HZ / 2);
 	set_bit(HW_INITIALIZED, &wnd->hw_status);
-	hangcheck_add(wnd);
 	up(&wnd->ndis_comm_mutex);
 	up(&wnd->tx_ring_mutex);
+	hangcheck_add(wnd);
 	/* the description about NDIS_ATTRIBUTE_NO_HALT_ON_SUSPEND is
 	 * misleading/confusing; we just ignore it */
 	status = miniport_query_info(wnd, OID_PNP_CAPABILITIES,
@@ -299,6 +301,28 @@ static void miniport_halt(struct wrap_ndis_device *wnd)
 		DBGTRACE1("halt: %p", miniport->miniport_halt);
 		LIN2WIN1(miniport->miniport_halt, wnd->nmb->adapter_ctx);
 		clear_bit(HW_INITIALIZED, &wnd->hw_status);
+		/* cancel any timers left by bugyy windows driver; also free
+		 * the memory for timers */
+		while (1) {
+			KIRQL irql;
+			struct nt_list *ent;
+			struct wrap_timer *wrap_timer;
+
+			irql = nt_spin_lock_irql(&timer_lock, DISPATCH_LEVEL);
+			ent = RemoveHeadList(&wnd->timer_list);
+			nt_spin_unlock_irql(&timer_lock, irql);
+			if (!ent)
+				break;
+			wrap_timer = container_of(ent, struct wrap_timer, list);
+			/* ktimer that this wrap_timer is associated to can't
+			 * be touched, as it may have been freed by the driver
+			 * already */
+			if (del_timer_sync(&wrap_timer->timer))
+				WARNING("Buggy Windows driver left timer %p "
+					"running", &wrap_timer->timer);
+			memset(wrap_timer, 0, sizeof(*wrap_timer));
+			slack_kfree(wrap_timer);
+		}
 	} else
 		WARNING("device %p is not initialized - not halting", wnd);
 	TRACEEXIT1(return);
@@ -396,7 +420,7 @@ static NDIS_STATUS miniport_set_power_state(struct wrap_ndis_device *wnd,
 	}
 }
 
-static int ndis_set_mac_addr(struct net_device *dev, void *p)
+static int ndis_set_mac_address(struct net_device *dev, void *p)
 {
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
 	struct sockaddr *addr = p;
@@ -407,17 +431,11 @@ static int ndis_set_mac_addr(struct net_device *dev, void *p)
 	unsigned char mac_string[3 * ETH_ALEN];
 	mac_address mac;
 
-	/* string <-> ansi <-> unicode conversion is driving me nuts */
-
 	memcpy(mac, addr->sa_data, sizeof(mac));
 	memset(mac_string, 0, sizeof(mac_string));
-	res = snprintf(mac_string, sizeof(mac_string), MACSTR,
-		       MAC2STR(mac));
-	if (res != (sizeof(mac_string) - 1))
-		TRACEEXIT1(return -EINVAL);
-
-	RtlInitAnsiString(&ansi, "mac_address");
-	if (RtlAnsiStringToUnicodeString(&key, &ansi, TRUE))
+	res = snprintf(mac_string, sizeof(mac_string), MACSTR, MAC2STR(mac));
+	DBGTRACE1("%d", res);
+	if (res != (2 * sizeof(mac)))
 		TRACEEXIT1(return -EINVAL);
 
 	RtlInitAnsiString(&ansi, mac_string);
@@ -426,17 +444,34 @@ static int ndis_set_mac_addr(struct net_device *dev, void *p)
 		TRACEEXIT1(return -EINVAL);
 	}
 	param.type = NdisParameterString;
+	RtlInitAnsiString(&ansi, "mac_address");
+	if (RtlAnsiStringToUnicodeString(&key, &ansi, TRUE))
+		TRACEEXIT1(return -EINVAL);
 	NdisWriteConfiguration(&res, wnd->nmb, &key, &param);
+	RtlFreeUnicodeString(&key);
 	if (res != NDIS_STATUS_SUCCESS) {
-		RtlFreeUnicodeString(&key);
 		RtlFreeUnicodeString(&param.data.string);
 		TRACEEXIT1(return -EINVAL);
 	}
-	if (pnp_stop_device(wnd->wd) == STATUS_SUCCESS &&
-	    pnp_start_device(wnd->wd) == STATUS_SUCCESS)
-		memcpy(dev->dev_addr, addr->sa_data, dev->addr_len);
+
+	RtlInitAnsiString(&ansi, "NetworkAddress");
+	if (RtlAnsiStringToUnicodeString(&key, &ansi, TRUE))
+		TRACEEXIT1(return -EINVAL);
+	NdisWriteConfiguration(&res, wnd->nmb, &key, &param);
 	RtlFreeUnicodeString(&key);
 	RtlFreeUnicodeString(&param.data.string);
+
+	if (res != NDIS_STATUS_SUCCESS)
+		TRACEEXIT1(return -EINVAL);
+	if (ndis_reinit(wnd) == NDIS_STATUS_SUCCESS) {
+		res = miniport_query_info(wnd, OID_802_3_CURRENT_ADDRESS,
+					  mac, sizeof(mac));
+		if (res == NDIS_STATUS_SUCCESS) {
+			DBGTRACE1("mac:" MACSTRSEP, MAC2STR(mac));
+			memcpy(dev->dev_addr, mac, sizeof(mac));
+		} else
+			ERROR("couldn't get mac address: %08X", res);
+	}
 	TRACEEXIT1(return 0);
 }
 
@@ -480,7 +515,7 @@ static struct ndis_packet *alloc_tx_packet(struct wrap_ndis_device *wnd,
 	DBG_BLOCK(4) {
 		dump_bytes(__FUNCTION__, skb->data, skb->len);
 	}
-	DBGTRACE4("packet: %p, buffer: %p, skb: %p", packet, buffer, skb);
+	DBGTRACE3("packet: %p, buffer: %p, skb: %p", packet, buffer, skb);
 	return packet;
 }
 
@@ -489,18 +524,15 @@ void free_tx_packet(struct wrap_ndis_device *wnd, struct ndis_packet *packet,
 {
 	ndis_buffer *buffer;
 	struct ndis_packet_oob_data *oob_data;
-	KIRQL irql;
 
 	TRACEENTER3("%p, %08X", packet, status);
-	irql = nt_spin_lock_irql(&wnd->tx_stats_lock, DISPATCH_LEVEL);
 	if (status == NDIS_STATUS_SUCCESS) {
-		wnd->stats.tx_bytes += packet->private.len;
-		wnd->stats.tx_packets++;
+		pre_atomic_add(wnd->stats.tx_bytes, packet->private.len);
+		atomic_inc_var(wnd->stats.tx_packets);
 	} else {
 		DBGTRACE1("packet dropped: %08X", status);
-		wnd->stats.tx_dropped++;
+		atomic_inc_var(wnd->stats.tx_dropped);
 	}
-	nt_spin_unlock_irql(&wnd->tx_stats_lock, irql);
 	oob_data = NDIS_PACKET_OOB_DATA(packet);
 	if (wnd->use_sg_dma)
 		PCI_DMA_UNMAP_SINGLE(wnd->wd->pci.pdev,
@@ -510,62 +542,46 @@ void free_tx_packet(struct wrap_ndis_device *wnd, struct ndis_packet *packet,
 	buffer = packet->private.buffer_head;
 	DBGTRACE3("freeing buffer %p", buffer);
 	NdisFreeBuffer(buffer);
-	if (oob_data->skb)
-		dev_kfree_skb_any(oob_data->skb);
+	dev_kfree_skb_any(oob_data->skb);
 	DBGTRACE3("freeing packet %p", packet);
 	NdisFreePacket(packet);
 	TRACEEXIT3(return);
 }
 
 /* MiniportSend and MiniportSendPackets */
-/* this function is called holding tx_ring_mutex, so safe to read
- * tx_ring_start (tx_ring_end is not updated in tx_worker or here, so
- * safe to read tx_ring_end, too) without lock */
-static int miniport_tx_packets(struct wrap_ndis_device *wnd)
+/* this function is called holding tx_ring_mutex. start and n are such
+ * that start + n < TX_RING_SIZE; i.e., packets don't wrap around
+ * ring */
+static int miniport_tx_packets(struct wrap_ndis_device *wnd, int start, int n)
 {
 	NDIS_STATUS res;
 	struct miniport_char *miniport;
-	int n, sent, start, end;
 	struct ndis_packet *packet;
+	int sent;
 	KIRQL irql;
 
+	DBGTRACE3("%d, %d", start, n);
 	miniport = &wnd->wd->driver->ndis_driver->miniport;
-	start = wnd->tx_ring_start;
-	end = wnd->tx_ring_end;
-	/* end == start when ring is full: (TX_RING_SIZE - 1) number
-	 * of packets are pending */
-	n = end - start;
-	if (n < 0)
-		n += TX_RING_SIZE;
-	else if (n == 0) {
-		assert(wnd->is_tx_ring_full == 1);
-		n = TX_RING_SIZE - 1;
-	}
-	if (unlikely(n > wnd->max_tx_packets))
-		n = wnd->max_tx_packets;
-	DBGTRACE3("%d, %d, %d", n, start, end);
 	if (miniport->send_packets) {
-		int i;
-		/* copy packets from tx ring to linear tx array */
-		for (i = 0; i < n; i++) {
-			int j = (start + i) % TX_RING_SIZE;
-			wnd->tx_array[i] = wnd->tx_ring[j];
-		}
-		if (wnd->attributes & NDIS_ATTRIBUTE_DESERIALIZE) {
+		if (deserialized_driver(wnd)) {
 			LIN2WIN3(miniport->send_packets, wnd->nmb->adapter_ctx,
-				 wnd->tx_array, n);
+				 &wnd->tx_ring[start], n);
 			sent = n;
 		} else {
 			struct ndis_packet_oob_data *oob_data;
 			irql = raise_irql(DISPATCH_LEVEL);
+			serialize_lock(wnd);
 			LIN2WIN3(miniport->send_packets, wnd->nmb->adapter_ctx,
-				 wnd->tx_array, n);
+				 &wnd->tx_ring[start], n);
+			serialize_unlock(wnd);
 			lower_irql(irql);
 			for (sent = 0; sent < n && wnd->tx_ok; sent++) {
-				packet = wnd->tx_array[sent];
+				NDIS_STATUS pkt_status;
+				packet = wnd->tx_ring[start + sent];
 				oob_data = NDIS_PACKET_OOB_DATA(packet);
-				switch (xchg(&oob_data->status,
-					     NDIS_STATUS_NOT_RECOGNIZED)) {
+				switch ((pkt_status =
+					 xchg(&oob_data->status,
+					      NDIS_STATUS_NOT_RECOGNIZED))) {
 				case NDIS_STATUS_SUCCESS:
 					free_tx_packet(wnd, packet,
 						       NDIS_STATUS_SUCCESS);
@@ -573,7 +589,7 @@ static int miniport_tx_packets(struct wrap_ndis_device *wnd)
 				case NDIS_STATUS_PENDING:
 					break;
 				case NDIS_STATUS_RESOURCES:
-					wnd->tx_ok = 0;
+					atomic_dec_var(wnd->tx_ok);
 					/* resubmit this packet and
 					 * the rest when resources
 					 * become available */
@@ -584,29 +600,27 @@ static int miniport_tx_packets(struct wrap_ndis_device *wnd)
 						       NDIS_STATUS_FAILURE);
 					break;
 				default:
-					ERROR("packet %p: invalid status",
-					      packet);
+					ERROR("%p: invalid status: %08X",
+					      packet, pkt_status);
 					free_tx_packet(wnd, packet,
 						       oob_data->status);
 					break;
 				}
+				DBGTRACE3("%p, %d", packet, pkt_status);
 			}
 		}
 		DBGTRACE3("sent: %d(%d)", sent, n);
 	} else {
-		int i;
 		irql = PASSIVE_LEVEL;
-		for (i = 0; i < n && wnd->tx_ok; i++) {
+		for (sent = 0; sent < n && wnd->tx_ok; sent++) {
 			struct ndis_packet_oob_data *oob_data;
-			packet = wnd->tx_ring[(start + i) % TX_RING_SIZE];
+			packet = wnd->tx_ring[start + sent];
 			oob_data = NDIS_PACKET_OOB_DATA(packet);
 			oob_data->status = NDIS_STATUS_NOT_RECOGNIZED;
-			if (!(wnd->attributes & NDIS_ATTRIBUTE_DESERIALIZE))
-				irql = raise_irql(DISPATCH_LEVEL);
+			irql = serialize_lock_irql(wnd);
 			res = LIN2WIN3(miniport->send, wnd->nmb->adapter_ctx,
 				       packet, packet->private.flags);
-			if (!(wnd->attributes & NDIS_ATTRIBUTE_DESERIALIZE))
-				lower_irql(irql);
+			serialize_unlock_irql(wnd, irql);
 			switch (res) {
 			case NDIS_STATUS_SUCCESS:
 				free_tx_packet(wnd, packet, res);
@@ -614,10 +628,10 @@ static int miniport_tx_packets(struct wrap_ndis_device *wnd)
 			case NDIS_STATUS_PENDING:
 				break;
 			case NDIS_STATUS_RESOURCES:
-				wnd->tx_ok = 0;
+				atomic_dec_var(wnd->tx_ok);
 				/* resend this packet when resources
 				 * become available */
-				i--;
+				sent--;
 				break;
 			case NDIS_STATUS_FAILURE:
 				free_tx_packet(wnd, packet, res);
@@ -628,7 +642,6 @@ static int miniport_tx_packets(struct wrap_ndis_device *wnd)
 				break;
 			}
 		}
-		sent = i + 1;
 	}
 	TRACEEXIT3(return sent);
 }
@@ -644,12 +657,20 @@ static void tx_worker(void *param)
 			break;
 		/* end == start if either ring is empty or full; in
 		 * the latter case is_tx_ring_full is set */
-		if (wnd->tx_ring_end == wnd->tx_ring_start &&
-		    !wnd->is_tx_ring_full) {
-			up(&wnd->tx_ring_mutex);
-			break;
-		}
-		n = miniport_tx_packets(wnd);
+		n = wnd->tx_ring_end - wnd->tx_ring_start;
+		DBGTRACE3("%d, %d, %d", wnd->tx_ring_start, wnd->tx_ring_end, n);
+		if (n == 0) {
+			if (wnd->is_tx_ring_full)
+				n = TX_RING_SIZE - wnd->tx_ring_start;
+			else {
+				up(&wnd->tx_ring_mutex);
+				break;
+			}
+		} else if (n < 0)
+			n = TX_RING_SIZE - wnd->tx_ring_start;
+		if (unlikely(n > wnd->max_tx_packets))
+			n = wnd->max_tx_packets;
+		n = miniport_tx_packets(wnd, wnd->tx_ring_start, n);
 		if (n > 0) {
 			wnd->tx_ring_start =
 				(wnd->tx_ring_start + n) % TX_RING_SIZE;
@@ -658,8 +679,7 @@ static void tx_worker(void *param)
 				netif_wake_queue(wnd->net_dev);
 		}
 		up(&wnd->tx_ring_mutex);
-		DBGTRACE3("%d, %d, %d", n, wnd->tx_ring_start,
-			  wnd->tx_ring_end);
+		DBGTRACE3("%d, %d, %d", wnd->tx_ring_start, wnd->tx_ring_end, n);
 	}
 	TRACEEXIT3(return);
 }
@@ -684,7 +704,7 @@ static int tx_skbuff(struct sk_buff *skb, struct net_device *dev)
 		wnd->is_tx_ring_full = 1;
 		netif_stop_queue(wnd->net_dev);
 	}
-	DBGTRACE3("%d, %d", wnd->tx_ring_start, wnd->tx_ring_end);
+	DBGTRACE3("ring: %d, %d", wnd->tx_ring_start, wnd->tx_ring_end);
 	schedule_wrap_work(&wnd->tx_work);
 	return NETDEV_TX_OK;
 }
@@ -770,10 +790,7 @@ static void ndis_poll_controller(struct net_device *dev)
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
 
 	disable_irq(dev->irq);
-	if (wnd->ndis_irq->req_isr)
-		ndis_isr_shared(dev->irq, wnd, NULL);
-	else
-		ndis_isr_dynamic(dev->irq, wnd, NULL);
+	ndis_isr(dev->irq, wnd, NULL);
 	enable_irq(dev->irq);
 }
 #endif
@@ -823,7 +840,7 @@ static void ndis_get_drvinfo(struct net_device *dev,
 static u32 ndis_get_link(struct net_device *dev)
 {
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
-	return wnd->link_status;
+	return netif_carrier_ok(wnd->net_dev);
 }
 
 static void ndis_get_wol(struct net_device *dev, struct ethtool_wolinfo *wol)
@@ -867,6 +884,30 @@ static struct ethtool_ops ndis_ethtool_ops = {
 };
 #endif
 
+static int notifier_event(struct notifier_block *notifier, unsigned long event,
+			  void *ptr)
+{
+	struct net_device *net_dev = (struct net_device *)ptr;
+	struct wrap_ndis_device *wnd = netdev_priv(net_dev);
+
+	/* called with rtnl lock held, so no need to lock */
+	switch (event) {
+	case NETDEV_CHANGENAME:
+		wrap_procfs_remove_ndis_device(wnd);
+		printk(KERN_INFO "%s: changing interface name from '%s' to "
+		       "'%s'\n", DRIVER_NAME, wnd->netdev_name, net_dev->name);
+		memcpy(wnd->netdev_name, net_dev->name,
+		       sizeof(wnd->netdev_name));
+		wrap_procfs_add_ndis_device(wnd);
+		break;
+	}
+	return NOTIFY_DONE;
+}
+
+static struct notifier_block netdev_notifier = {
+	.notifier_call = notifier_event,
+};
+
 static void update_wireless_stats(struct wrap_ndis_device *wnd)
 {
 	struct iw_statistics *iw_stats = &wnd->wireless_stats;
@@ -876,7 +917,7 @@ static void update_wireless_stats(struct wrap_ndis_device *wnd)
 	unsigned long frag;
 
 	TRACEENTER2("%p", wnd);
-	if (wnd->stats_enabled == FALSE || wnd->link_status == 0) {
+	if (wnd->stats_enabled == FALSE || !netif_carrier_ok(wnd->net_dev)) {
 		memset(iw_stats, 0, sizeof(*iw_stats));
 		TRACEEXIT2(return);
 	}
@@ -941,7 +982,7 @@ static void set_multicast_list(struct wrap_ndis_device *wnd)
 			if (mclist->dmi_addrlen != ETH_ALEN)
 				continue;
 			memcpy(buf + i * ETH_ALEN, mclist->dmi_addr, ETH_ALEN);
-			DBGTRACE2(MACSTR, MAC2STR(mclist->dmi_addr));
+			DBGTRACE2(MACSTRSEP, MAC2STR(mclist->dmi_addr));
 			i++;
 		}
 		res = miniport_set_info(wnd, OID_802_3_MULTICAST_LIST,
@@ -971,25 +1012,25 @@ static void link_status_handler(struct wrap_ndis_device *wnd)
 	int i;
 #endif
 
-	TRACEENTER2("link: %d", wnd->link_status);
+	TRACEENTER2("link: %d", netif_carrier_ok(wnd->net_dev));
 	if (wnd->physical_medium != NdisPhysicalMediumWirelessLan)
 		TRACEEXIT2(return);
-	if (wnd->link_status == 0) {
-		wnd->tx_ok = 0;
+#ifndef CONFIG_NET_RADIO
+	TRACEEXIT2(return);
+#endif
+	if (!netif_carrier_ok(wnd->net_dev)) {
 		memset(&wrqu, 0, sizeof(wrqu));
 		wrqu.ap_addr.sa_family = ARPHRD_ETHER;
 		wireless_send_event(wnd->net_dev, SIOCGIWAP, &wrqu, NULL);
 		TRACEEXIT2(return);
 	}
 
-	wnd->tx_ok = 1;
 	ndis_assoc_info = kmalloc(assoc_size, GFP_KERNEL);
 	if (!ndis_assoc_info) {
 		ERROR("couldn't allocate memory");
 		TRACEEXIT2(return);
 	}
 	memset(ndis_assoc_info, 0, assoc_size);
-
 	res = miniport_query_info(wnd, OID_802_11_ASSOCIATION_INFORMATION,
 				  ndis_assoc_info, assoc_size);
 	if (res) {
@@ -1050,7 +1091,7 @@ static void link_status_handler(struct wrap_ndis_device *wnd)
 	get_ap_address(wnd, (char *)&wrqu.ap_addr.sa_data);
 	wrqu.ap_addr.sa_family = ARPHRD_ETHER;
 	wireless_send_event(wnd->net_dev, SIOCGIWAP, &wrqu, NULL);
-	DBGTRACE2(MACSTR, MAC2STR(wrqu.ap_addr.sa_data));
+	DBGTRACE2(MACSTRSEP, MAC2STR(wrqu.ap_addr.sa_data));
 	TRACEEXIT2(return);
 }
 
@@ -1070,7 +1111,7 @@ static void add_stats_timer(struct wrap_ndis_device *wnd)
 	if (wnd->physical_medium != NdisPhysicalMediumWirelessLan)
 		return;
 	wnd->stats_timer.data = (unsigned long)wnd;
-	wnd->stats_timer.function = &stats_timer_proc;
+	wnd->stats_timer.function = stats_timer_proc;
 	wnd->stats_timer.expires = jiffies + 10 * HZ;
 	add_timer(&wnd->stats_timer);
 }
@@ -1083,14 +1124,20 @@ static void del_stats_timer(struct wrap_ndis_device *wnd)
 static void hangcheck_proc(unsigned long data)
 {
 	struct wrap_ndis_device *wnd = (struct wrap_ndis_device *)data;
+	BOOLEAN reset;
+	struct miniport_char *miniport;
 
 	TRACEENTER3("");
-	set_bit(HANGCHECK, &wnd->wrap_ndis_pending_work);
-	schedule_wrap_work(&wnd->wrap_ndis_work);
-
+	miniport = &wnd->wd->driver->ndis_driver->miniport;
+	if_serialize_lock(wnd);
+	reset = LIN2WIN1(miniport->hangcheck, wnd->nmb->adapter_ctx);
+	if_serialize_unlock(wnd);
+	if (reset) {
+		set_bit(MINIPORT_RESET, &wnd->wrap_ndis_pending_work);
+		schedule_wrap_work(&wnd->wrap_ndis_work);
+	}
 	wnd->hangcheck_timer.expires += wnd->hangcheck_interval;
 	add_timer(&wnd->hangcheck_timer);
-
 	TRACEEXIT3(return);
 }
 
@@ -1103,7 +1150,7 @@ void hangcheck_add(struct wrap_ndis_device *wnd)
 		wnd->hangcheck_interval = hangcheck_interval * HZ;
 	init_timer(&wnd->hangcheck_timer);
 	wnd->hangcheck_timer.data = (unsigned long)wnd;
-	wnd->hangcheck_timer.function = &hangcheck_proc;
+	wnd->hangcheck_timer.function = hangcheck_proc;
 	wnd->hangcheck_timer.expires = jiffies + wnd->hangcheck_interval;
 	add_timer(&wnd->hangcheck_timer);
 	return;
@@ -1134,28 +1181,26 @@ static void wrap_ndis_worker(void *param)
 			       &wnd->wrap_ndis_pending_work))
 		link_status_handler(wnd);
 
-	if (test_and_clear_bit(HANGCHECK, &wnd->wrap_ndis_pending_work)) {
-		NDIS_STATUS res;
-		struct miniport_char *miniport;
-		KIRQL irql;
-
-		miniport = &wnd->wd->driver->ndis_driver->miniport;
-		irql = raise_irql(DISPATCH_LEVEL);
-		res = LIN2WIN1(miniport->hangcheck, wnd->nmb->adapter_ctx);
-		lower_irql(irql);
-		if (res) {
-			WARNING("%s is being reset", wnd->net_dev->name);
-			res = miniport_reset(wnd);
-			DBGTRACE3("%08X", res);
-		}
-	}
+	if (test_and_clear_bit(MINIPORT_RESET, &wnd->wrap_ndis_pending_work))
+		miniport_reset(wnd);
 	TRACEEXIT3(return);
 }
 
 NDIS_STATUS ndis_reinit(struct wrap_ndis_device *wnd)
 {
-	pnp_stop_device(wnd->wd);
-	return pnp_start_device(wnd->wd);
+	NDIS_STATUS status;
+	wnd->pm_capa = FALSE;
+	status = miniport_set_power_state(wnd, NdisDeviceStateD3);
+	if (status != NDIS_STATUS_SUCCESS) {
+		ERROR("halting device %s failed: %08X", wnd->net_dev->name,
+		      status);
+		return status;
+	}
+	status = miniport_set_power_state(wnd, NdisDeviceStateD0);
+	if (status != NDIS_STATUS_SUCCESS)
+		ERROR("starting device %s failed: %08X", wnd->net_dev->name,
+		      status);
+	return status;
 }
 
 void get_encryption_capa(struct wrap_ndis_device *wnd)
@@ -1283,24 +1328,19 @@ void get_encryption_capa(struct wrap_ndis_device *wnd)
 	TRACEEXIT1(return);
 }
 
-/* called as Windows function, so call WIN2LIN2 before accessing
- * arguments */
 wstdcall NTSTATUS NdisDispatchDeviceControl(struct device_object *fdo,
 					    struct irp *irp)
 {
 	struct wrap_ndis_device *wnd;
 
-	WIN2LIN2(fdo, irp);
-
 	DBGTRACE3("fdo: %p", fdo);
 	/* for now, we don't have anything intresting here, so pass it
 	 * down to bus driver */
 	wnd = fdo->reserved;
-	return LIN2WIN2(IoPassIrpDown, wnd->nmb->pdo, irp);
+	return IoPassIrpDown(wnd->nmb->pdo, irp);
 }
+WIN_FUNC_DECL(NdisDispatchDeviceControl,2)
 
-/* called as Windows function, so call WIN2LIN2 before accessing
- * arguments */
 wstdcall NTSTATUS NdisDispatchPower(struct device_object *fdo, struct irp *irp)
 {
 	struct io_stack_location *irp_sl;
@@ -1309,8 +1349,6 @@ wstdcall NTSTATUS NdisDispatchPower(struct device_object *fdo, struct irp *irp)
 	NTSTATUS status;
 	NDIS_STATUS ndis_status;
 
-	WIN2LIN2(fdo, irp);
-
 	irp_sl = IoGetCurrentIrpStackLocation(irp);
 	wnd = fdo->reserved;
 	IOTRACE("fdo: %p, fn: %d:%d, wnd: %p", fdo, irp_sl->major_fn,
@@ -1325,7 +1363,7 @@ wstdcall NTSTATUS NdisDispatchPower(struct device_object *fdo, struct irp *irp)
 	switch (irp_sl->minor_fn) {
 	case IRP_MN_SET_POWER:
 		if (state == NdisDeviceStateD0) {
-			status = LIN2WIN2(IoSyncForwardIrp, wnd->nmb->pdo, irp);
+			status = IoSyncForwardIrp(wnd->nmb->pdo, irp);
 			if (status != STATUS_SUCCESS)
 				break;
 			ndis_status = miniport_set_power_state(wnd, state);
@@ -1342,14 +1380,14 @@ wstdcall NTSTATUS NdisDispatchPower(struct device_object *fdo, struct irp *irp)
 			if (ndis_status != NDIS_STATUS_SUCCESS)
 				WARNING("setting power to %d failed: %08X",
 					state, ndis_status);
-			status = LIN2WIN2(IoAsyncForwardIrp, wnd->nmb->pdo, irp);
+			status = IoAsyncForwardIrp(wnd->nmb->pdo, irp);
 		}
 		break;
 	case IRP_MN_QUERY_POWER:
 		if (wnd->pm_capa) {
-			ndis_status = miniport_query_info(wnd,
-							  OID_PNP_QUERY_POWER,
-							  &state, sizeof(state));
+			ndis_status =
+				miniport_query_info(wnd, OID_PNP_QUERY_POWER,
+						    &state, sizeof(state));
 			DBGTRACE2("%d, %08X", state, ndis_status);
 			/* this OID must always succeed */
 			if (ndis_status != NDIS_STATUS_SUCCESS)
@@ -1358,22 +1396,21 @@ wstdcall NTSTATUS NdisDispatchPower(struct device_object *fdo, struct irp *irp)
 			irp->io_status.status = STATUS_SUCCESS;
 		} else
 			irp->io_status.status = STATUS_SUCCESS;
-		status = LIN2WIN2(IoPassIrpDown, wnd->nmb->pdo, irp);
+		status = IoPassIrpDown(wnd->nmb->pdo, irp);
 		break;
 	case IRP_MN_WAIT_WAKE:
 	case IRP_MN_POWER_SEQUENCE:
 		/* TODO: implement WAIT_WAKE */
-		status = LIN2WIN2(IoPassIrpDown, wnd->nmb->pdo, irp);
+		status = IoPassIrpDown(wnd->nmb->pdo, irp);
 		break;
 	default:
-		status = LIN2WIN2(IoPassIrpDown, wnd->nmb->pdo, irp);
+		status = IoPassIrpDown(wnd->nmb->pdo, irp);
 		break;
 	}
 	IOEXIT(return status);
 }
+WIN_FUNC_DECL(NdisDispatchPower,2)
 
-/* called as Windows function, so call WIN2LIN2 before accessing
- * arguments */
 wstdcall NTSTATUS NdisDispatchPnp(struct device_object *fdo, struct irp *irp)
 {
 	struct io_stack_location *irp_sl;
@@ -1381,15 +1418,13 @@ wstdcall NTSTATUS NdisDispatchPnp(struct device_object *fdo, struct irp *irp)
 	struct device_object *pdo;
 	NTSTATUS status;
 
-	WIN2LIN2(fdo, irp);
-
 	IOTRACE("fdo: %p, irp: %p", fdo, irp);
 	irp_sl = IoGetCurrentIrpStackLocation(irp);
 	wnd = fdo->reserved;
 	pdo = wnd->nmb->pdo;
 	switch (irp_sl->minor_fn) {
 	case IRP_MN_START_DEVICE:
-		status = LIN2WIN2(IoSyncForwardIrp, pdo, irp);
+		status = IoSyncForwardIrp(pdo, irp);
 		if (status != STATUS_SUCCESS)
 			break;
 		if (ndis_start_device(wnd) == NDIS_STATUS_SUCCESS)
@@ -1401,34 +1436,33 @@ wstdcall NTSTATUS NdisDispatchPnp(struct device_object *fdo, struct irp *irp)
 		break;
 	case IRP_MN_QUERY_STOP_DEVICE:
 		/* TODO: implement in NDIS */
-		status = LIN2WIN2(IoPassIrpDown, wnd->nmb->pdo, irp);
+		status = IoPassIrpDown(wnd->nmb->pdo, irp);
 		break;
 	case IRP_MN_STOP_DEVICE:
 		miniport_halt(wnd);
 		irp->io_status.status = STATUS_SUCCESS;
-		status = LIN2WIN2(IoAsyncForwardIrp, pdo, irp);
+		status = IoAsyncForwardIrp(pdo, irp);
 		break;
 	case IRP_MN_REMOVE_DEVICE:
 		DBGTRACE1("%s", wnd->net_dev->name);
-		if (wnd->wd->surprise_removed == TRUE)
-			miniport_pnp_event(wnd,
-					   NdisDevicePnPEventSurpriseRemoved);
+		miniport_pnp_event(wnd, NdisDevicePnPEventSurpriseRemoved);
 		if (ndis_remove_device(wnd)) {
 			status = STATUS_FAILURE;
 			break;
 		}
 		/* wnd is already freed */
-		status = LIN2WIN2(IoAsyncForwardIrp, pdo, irp);
+		status = IoAsyncForwardIrp(pdo, irp);
 		IoDetachDevice(fdo);
 		IoDeleteDevice(fdo);
 		break;
 	default:
-		status = LIN2WIN2(IoAsyncForwardIrp, pdo, irp);
+		status = IoAsyncForwardIrp(pdo, irp);
 		break;
 	}
 	IOTRACE("status: %08X", status);
 	IOEXIT(return status);
 }
+WIN_FUNC_DECL(NdisDispatchPnp,2)
 
 static int set_task_offload(struct wrap_ndis_device *wnd, void *buf,
 			    const int buf_size)
@@ -1559,10 +1593,6 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 		TRACEEXIT1(return NDIS_STATUS_SUCCESS);
 	if (ndis_status != NDIS_STATUS_SUCCESS)
 		TRACEEXIT1(return ndis_status);
-	/* NB: tx_array is used to recognize if device is being
-	 * started for the first time or being re-started */
-	if (wnd->tx_array)
-		TRACEEXIT2(return NDIS_STATUS_SUCCESS);
 	wd = wnd->wd;
 	net_dev = wnd->net_dev;
 
@@ -1581,7 +1611,7 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 		ERROR("couldn't get mac address: %08X", ndis_status);
 		return ndis_status;
 	}
-	DBGTRACE1("mac:" MACSTR, MAC2STR(mac));
+	DBGTRACE1("mac:" MACSTRSEP, MAC2STR(mac));
 	memcpy(&net_dev->dev_addr, mac, ETH_ALEN);
 
 	net_dev->open = ndis_open;
@@ -1594,10 +1624,9 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 		net_dev->get_wireless_stats = get_wireless_stats;
 #endif
 		net_dev->wireless_handlers = &ndis_handler_def;
-	} else
-		wnd->tx_ok = 1;
+	}
 	net_dev->set_multicast_list = ndis_set_multicast_list;
-	net_dev->set_mac_address = ndis_set_mac_addr;
+	net_dev->set_mac_address = ndis_set_mac_address;
 #if defined(HAVE_ETHTOOL)
 	net_dev->ethtool_ops = &ndis_ethtool_ops;
 #endif
@@ -1624,45 +1653,39 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 	}
 
 	set_task_offload(wnd, buf, buf_size);
+
 	if (register_netdev(net_dev)) {
 		ERROR("cannot register net device %s", net_dev->name);
 		goto err_register;
 	}
-
+	register_netdevice_notifier(&netdev_notifier);
+	memcpy(wnd->netdev_name, net_dev->name, sizeof(wnd->netdev_name));
+	wnd->tx_ok = 1;
 	memset(buf, 0, buf_size);
 	ndis_status = miniport_query_info(wnd, OID_GEN_VENDOR_DESCRIPTION,
 					  buf, buf_size);
 	if (ndis_status == NDIS_STATUS_SUCCESS)
 		printk(KERN_INFO "%s: vendor: '%s'\n", net_dev->name, buf);
 
-	printk(KERN_INFO "%s: %s ethernet device " MACSTR " using %sdriver %s,"
-	       " %s\n", net_dev->name, DRIVER_NAME, MAC2STR(net_dev->dev_addr),
-	       wnd->attributes & NDIS_ATTRIBUTE_DESERIALIZE ? "" : "serialized ",
+	printk(KERN_INFO "%s: ethernet device " MACSTRSEP " using %sNDIS "
+	       "driver %s, %s\n", net_dev->name, MAC2STR(net_dev->dev_addr),
+	       deserialized_driver(wnd) ? "" : "serialized ",
 	       wnd->wd->driver->name, wnd->wd->conf_file_name);
 
-	if (wnd->attributes & NDIS_ATTRIBUTE_DESERIALIZE) {
+	if (deserialized_driver(wnd)) {
 		/* deserialized drivers don't have a limit, but we
-		 * keep max at TX_RING_SIZE to allocate tx_array
-		 * below */
+		 * keep max at TX_RING_SIZE */
 		wnd->max_tx_packets = TX_RING_SIZE;
 	} else {
 		ndis_status =
 			miniport_query_int(wnd, OID_GEN_MAXIMUM_SEND_PACKETS,
 					   &wnd->max_tx_packets);
-		if (ndis_status == NDIS_STATUS_NOT_SUPPORTED)
+		if (ndis_status != NDIS_STATUS_SUCCESS)
 			wnd->max_tx_packets = 1;
 		if (wnd->max_tx_packets > TX_RING_SIZE)
 			wnd->max_tx_packets = TX_RING_SIZE;
 	}
-	DBGTRACE1("maximum send packets: %d", wnd->max_tx_packets);
-	wnd->tx_array =
-		kmalloc(sizeof(struct ndis_packet *) * wnd->max_tx_packets,
-			GFP_KERNEL);
-	if (!wnd->tx_array) {
-		ERROR("couldn't allocate memory for tx_packets");
-		goto err_start;
-
-	}
+	DBGTRACE2("maximum send packets: %d", wnd->max_tx_packets);
 	/* we need at least one extra packet for
 	 * EthRxIndicateHandler */
 	NdisAllocatePacketPoolEx(&ndis_status, &wnd->tx_packet_pool,
@@ -1723,8 +1746,6 @@ buffer_pool_err:
 		wnd->tx_packet_pool = NULL;
 	}
 packet_pool_err:
-	kfree(wnd->tx_array);
-	wnd->tx_array = NULL;
 err_register:
 	kfree(buf);
 err_start:
@@ -1737,6 +1758,7 @@ static int ndis_remove_device(struct wrap_ndis_device *wnd)
 	int tx_pending;
 
 	set_bit(SHUTDOWN, &wnd->wrap_ndis_pending_work);
+	unregister_netdevice_notifier(&netdev_notifier);
 	wnd->tx_ok = 0;
 	ndis_close(wnd->net_dev);
 	netif_carrier_off(wnd->net_dev);
@@ -1778,8 +1800,6 @@ static int ndis_remove_device(struct wrap_ndis_device *wnd)
 		NdisFreeBufferPool(wnd->tx_buffer_pool);
 		wnd->tx_buffer_pool = NULL;
 	}
-	if (wnd->tx_array)
-		kfree(wnd->tx_array);
 	printk(KERN_INFO "%s: device %s removed\n", DRIVER_NAME,
 	       wnd->net_dev->name);
 	unregister_netdev(wnd->net_dev);
@@ -1816,8 +1836,8 @@ static wstdcall NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 	if (wrap_is_usb_bus(wd->dev_bus_type))
 		SET_NETDEV_DEV(net_dev, &wd->usb.intf->dev);
 #endif
-	status = IoCreateDevice(drv_obj, 0, NULL, FILE_DEVICE_UNKNOWN, 0,
-				FALSE, &fdo);
+	status = IoCreateDevice(drv_obj, 0, NULL,
+				FILE_DEVICE_UNKNOWN, 0, FALSE, &fdo);
 	if (status != STATUS_SUCCESS) {
 		free_netdev(net_dev);
 		TRACEEXIT2(return status);
@@ -1851,12 +1871,10 @@ static wstdcall NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 	init_MUTEX_LOCKED(&wnd->ndis_comm_mutex);
 	init_waitqueue_head(&wnd->ndis_comm_wq);
 	wnd->ndis_comm_done = 0;
-	wnd->tx_ok = 0;
 	INIT_WORK(&wnd->tx_work, tx_worker, wnd);
 	wnd->tx_ring_start = 0;
 	wnd->tx_ring_end = 0;
 	wnd->is_tx_ring_full = 0;
-	nt_spin_lock_init(&wnd->tx_stats_lock);
 	wnd->encr_mode = Ndis802_11EncryptionDisabled;
 	wnd->auth_mode = Ndis802_11AuthModeOpen;
 	wnd->capa.encr = 0;
@@ -1878,6 +1896,7 @@ static wstdcall NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 		wd->driver->ndis_driver->miniport.shutdown = NULL;
 	wnd->stats_enabled = TRUE;
 	wnd->rx_csum.value = 0;
+	InitializeListHead(&wnd->timer_list);
 
 	fdo->reserved = wnd;
 	nmb->fdo = fdo;
@@ -1885,13 +1904,18 @@ static wstdcall NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 	DBGTRACE1("nmb: %p, pdo: %p, fdo: %p, attached: %p, next: %p",
 		  nmb, pdo, fdo, fdo->attached, nmb->next_device);
 
+	/* dispatch routines are called as Windows functions */
 	for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++)
-		drv_obj->major_func[i] = IoPassIrpDown;
-	drv_obj->major_func[IRP_MJ_PNP] = NdisDispatchPnp;
-	drv_obj->major_func[IRP_MJ_POWER] = NdisDispatchPower;
+		drv_obj->major_func[i] =
+			WIN_FUNC_PTR(IoPassIrpDown,2);
+	drv_obj->major_func[IRP_MJ_PNP] =
+		WIN_FUNC_PTR(NdisDispatchPnp,2);
+	drv_obj->major_func[IRP_MJ_POWER] =
+		WIN_FUNC_PTR(NdisDispatchPower,2);
 	drv_obj->major_func[IRP_MJ_INTERNAL_DEVICE_CONTROL] =
-		NdisDispatchDeviceControl;
-	drv_obj->major_func[IRP_MJ_DEVICE_CONTROL] = NdisDispatchDeviceControl;
+		WIN_FUNC_PTR(NdisDispatchDeviceControl,2);
+//	drv_obj->major_func[IRP_MJ_DEVICE_CONTROL] =
+//		WIN_FUNC_PTR(NdisDispatchDeviceControl,2);
 	TRACEEXIT2(return STATUS_SUCCESS);
 }
 
